use std::time::Duration;

use log::*;
use pokerust::{FromName, Pokemon, PokemonForm, PokemonSpecies, Type, Move};
use serenity::{client::Context, framework::standard::{Args, CommandResult, macros::command}, model::{channel::Message, id::UserId}, prelude::Mentionable, utils::MessageBuilder};

use crate::{
    commands::checks::*,
    data::{DraftPokemon, PokemonFormExt, TradePerformance},
    EMBED_COLOR,
    func::*
};

#[command]
#[description("Join an open draft with its ID.")]
#[usage("<draft ID>")]
#[checks(DraftID)]
#[only_in(guilds)]
async fn join(ctx: &Context, msg: &Message, args: Args) -> CommandResult {
    info!("Draft Join Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            m
        },
        Err(y) => {
            match y {
                DraftParseError::Missing => {
                    match msg.guild_id {
                        Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                            Some(d) => d,
                            None => {
                                let _ = msg.reply(ctx, "You must specify a draft ID to join!").await;
                                return Ok(());
                            }
                        },
                        None => {
                            let _ = msg.reply(ctx, "You must specify a draft ID to join!").await;
                            return Ok(());
                        }
                    }
                },
                DraftParseError::Invalid => {
                    let _ = msg.reply(ctx, "That doesn't seem to be a valid draft ID. Check it for misspellings.").await;
                    return Ok(());
                }
            }
        }
    };

    if ctx.check_draft_open(&draft_id).await {
        match ctx.add_coach_to_draft(&draft_id, msg.author.id.0, msg.author.name.clone(), 0).await {
            Ok((dname, pos, p)) => {
                if p.is_some() {
                    let r = p.unwrap();
                    let g = msg.guild(ctx).unwrap();
                    let mut m = g.member(ctx, msg.author.id).await.expect("Couldn't get Member from User");
                    let _ = m.add_role(ctx, r).await;
                }
                let _ = msg.reply(ctx, format!("You've joined {}. You are position #{} in the drafting order. You will be notified when your tournament organizers start the drafting.", dname, pos)).await;
            },
            Err(why) => {
                let _ = msg.reply(ctx, why).await;
            }
        }
    }
    else {
        let _ = msg.reply(ctx, "That draft is not open for public signups!").await;
    }

    Ok(())
}

#[command]
#[description("Check a draft's status.")]
#[usage("<draft ID>")]
#[checks(DraftID)]
async fn status(ctx: &Context, msg: &Message, args: Args) -> CommandResult {
    info!("Draft Status Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            m
        },
        Err(y) => {
            match y {
                DraftParseError::Missing => {
                    match msg.guild_id {
                        Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                            Some(d) => d,
                            None => {
                                let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                                return Ok(());
                            }
                        },
                        None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                            Some(d) => d,
                            None => {
                                let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                                return Ok(());
                            }
                        }
                    }
                },
                DraftParseError::Invalid => {
                    let _ = msg.reply(ctx, "That doesn't seem to be a valid draft ID. Check it for misspellings.").await;
                    return Ok(());
                }
            }
        }
    };

    let mut embed = ctx.get_draft_embed(draft_id.clone()).await?;
    let joined = ctx.check_draft_joined(draft_id, msg.author.id.as_u64()).await;

    let _ = msg.channel_id.send_message(ctx, |m| {
        embed.title("Draft Information");
        if joined {
            embed.field("Are you in this draft?", "Yes", false);
        }
        else {
            embed.field("Are you in this draft?", "No", false);
        }
        m.set_embed(embed);
        m
    }).await?;

    Ok(())
}

#[command]
#[description("Show information about a Pokémon in a given draft.")]
#[usage("<draft ID> <pokemon>")]
#[checks(DraftID2)]
async fn mon(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Mon Info Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                }
            }
        }
    };
    let mon = match args.single::<String>() {
        Ok(m) => {
            let mut o = m.to_lowercase();
            if o.ends_with("-g") {
                o.push_str("alar");
            }
            if o.ends_with("-a") {
                o.push_str("lola");
            }
            if o.ends_with("-p") {
                o.push_str("aldea");
            }
            if o.as_str() == "darmanitan-galar" {
                o = "darmanitan-standard-galar".to_string();
            }
            let o = if o.as_str() == "gastrodon" { "gastrodon-west".to_owned() } 
            else if ["tauros-paldea", "tauros-combat", "tauros-paldea-combat", "tauros-combat-paldea", "tauros-combat-breed"].contains(&o.as_str()) { "tauros-paldea-combat-breed".to_owned() }
            else if ["tauros-blaze", "tauros-paldea-blaze", "tauros-blaze-paldea", "tauros-blaze-breed"].contains(&o.as_str()) { "tauros-paldea-blaze-breed".to_owned() }
            else if ["tauros-aqua", "tauros-paldea-aqua", "tauros-aqua-paldea", "tauros-aqua-breed"].contains(&o.as_str()) { "tauros-paldea-aqua-breed".to_owned() }
            else { o };

            o
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must mention a Pokémon species or form!").await;
            return Ok(());
        }
    };
    let m_id = match PokemonForm::from_name(mon.as_str()) {
        Ok(f) => f.id,
        Err(_) => {
            match Pokemon::from_name(mon.as_str()) {
                Ok(p) => {
                    p.forms[0].get().unwrap().id
                },
                Err(_) => {
                    match PokemonSpecies::from_name(mon.as_str()) {
                        Ok(s) => {
                            let p = s.varieties.iter().filter(|v| v.is_default).next().unwrap().pokemon.get()?;
                            p.forms[0].get().unwrap().id
                        },
                        Err(_) => {
                            let _ = msg.reply(ctx, "That doesn't seem to be a valid Pokémon species. Check for misspellings.").await;
                            return Ok(());
                        }
                    }
                }
            }
        }
    };

    let s_id = msg.guild_id.map(|g| g.0);
    match ctx.get_mon_embed(draft_id.clone(), m_id, s_id).await {
        Ok(embed) => {
            let _ = msg.channel_id.send_message(ctx, |m| {
                m.set_embed(embed);
                m
            }).await?;
        },
        Err(why) => {
            let _ = msg.reply(ctx, why).await;
        }
    }

    Ok(())
}

#[command]
#[description("When it's your turn in the draft, use this command to make your pick.")]
#[usage("<draft ID> <pokemon>")]
#[checks(Joined)]
async fn pick(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Pick Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                }
            }
        }
    };
    let mon = match args.single::<String>() {
        Ok(m) => {
            let mut o = m.to_lowercase();
            if o.ends_with("-g") {
                o.push_str("alar");
            }
            if o.ends_with("-a") {
                o.push_str("lola");
            }
            if o.ends_with("-p") {
                o.push_str("aldea");
            }
            if o.as_str() == "darmanitan-galar" {
                o = "darmanitan-standard-galar".to_string();
            }
            let o = if o.as_str() == "gastrodon" { "gastrodon-west".to_owned() } 
            else if ["tauros-paldea", "tauros-combat", "tauros-paldea-combat", "tauros-combat-paldea", "tauros-combat-breed"].contains(&o.as_str()) { "tauros-paldea-combat-breed".to_owned() }
            else if ["tauros-blaze", "tauros-paldea-blaze", "tauros-blaze-paldea", "tauros-blaze-breed"].contains(&o.as_str()) { "tauros-paldea-blaze-breed".to_owned() }
            else if ["tauros-aqua", "tauros-paldea-aqua", "tauros-aqua-paldea", "tauros-aqua-breed"].contains(&o.as_str()) { "tauros-paldea-aqua-breed".to_owned() }
            else { o };

            o
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must mention a Pokémon species or form!").await;
            return Ok(());
        }
    };
    let (m_id, mega_id) = {
        fn mon_and_mega_id(f: PokemonForm) -> (i16, Option<i16>) {
            if f.name.contains("-mega") {
                (f.id, Some(f.id))
            }
            else {
                (f.id, None)
            }
        }

        match PokemonForm::from_name(mon.as_str()) {
            Ok(f) => {
                mon_and_mega_id(f)
            },
            Err(_) => {
                match Pokemon::from_name(mon.as_str()) {
                    Ok(p) => {
                        mon_and_mega_id(p.forms[0].get().unwrap())
                    },
                    Err(_) => {
                        match PokemonSpecies::from_name(mon.as_str()) {
                            Ok(s) => {
                                let p = s.varieties.iter().filter(|v| v.is_default).next().unwrap().pokemon.get()?;
                                mon_and_mega_id(p.forms[0].get().unwrap())
                            },
                            Err(_) => {
                                let _ = msg.reply(ctx, "That doesn't seem to be a valid Pokémon species. Check for misspellings.").await;
                                return Ok(());
                            }
                        }
                    }
                }
            }
        }
    };

    match ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
        Ok(a_c) => {
            let s_id = a_c.clone().guild().map(|gc| gc.guild_id.0);
            match ctx.pick_mon(draft_id.clone(), msg.author.id.as_u64(), m_id, false, s_id, mega_id).await {
                Ok((d, n, e, p, complete)) => {
                    let a_nick = msg.author_nick(ctx).await.unwrap_or(msg.author.name.clone());
                    let _ = a_c.id().send_message(ctx, |m| {
                        m.content(format!("**Round #{}:** {} has drafted **{}** in {}!", p, a_nick, n, d));
                        m.set_embed(e);
                        m
                    }).await;

                    if complete {
                        let role = ctx.get_player_role(&draft_id).await;
                        let mut complete_message = MessageBuilder::new();
                        if role.is_some() {
                            let g = ctx.http.get_guild(s_id.unwrap()).await.unwrap();
                            let r = g.roles.iter().filter(|(r, _)| r.0 == role.unwrap()).map(|(_, r)| r).next().unwrap();
                            complete_message.mention(r);
                        }
                        complete_message.push_bold(" Drafting in");
                        complete_message.push(" ");
                        complete_message.push_bold_safe(d);
                        complete_message.push(" ");
                        complete_message.push_bold_line("has concluded!");
                        complete_message.push("Please await instructions from your draft organizers to begin playing matches.");

                        let _ = a_c.id().say(ctx, complete_message.build()).await;
                    }
                },
                Err(y) => {
                    let _ = msg.reply(ctx, y).await;
                }
            }
        },
        Err(y) => {
            error!("Draft Pick: Failed getting draft announcement channel: {}", y);
        }
    }

    Ok(())
}

#[command]
#[description("Change your coach information, such as your team name, Showdown username, and team logo.")]
#[usage("<draft ID> <key> <value>")]
#[checks(Joined)]
async fn setmy(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("User Setting Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                }
            }
        }
    };
    let key = match args.single::<String>() {
        Ok(m) => {
            m.to_lowercase()
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must mention a setting to change!").await;
            return Ok(());
        }
    };
    let value = args.rest().to_owned();
    
    match ctx.set_coach_value(draft_id, msg.author.id.as_u64(), key.clone(), value.clone()).await {
        Ok(old) => {
            let (title, res) = match key.as_str() {
                "showdown" => {
                    match old.as_str() {
                        "" => ("Showdown Username Updated".to_owned(), format!("None => {}", value)),
                        _ => ("Showdown Username Updated".to_owned(), format!("{} => {}", old, value)),
                    }
                },
                "team" | "name" => {
                    match old.as_str() {
                        "" => ("Team Name Updated".to_owned(), format!("None => {}", value)),
                        _ => ("Team Name Updated".to_owned(), format!("{} => {}", old, value)),
                    }
                },
                "logo" => {
                    ("Team Logo Updated".to_owned(), "Verify that the logo embeds correctly below. If an image does not appear, try again with a new image link ensuring it is a direct link, such as one that ends in '.png'.".to_owned())
                },
                _ => ("Error".to_owned(), "That is not a coach setting!".to_owned()),
            };
            let _ = msg.channel_id.send_message(ctx, |m| {
                m.embed(|e| {
                    e.title(title);
                    e.description(res);
                    if key.as_str() == "logo" {
                        e.image(value);
                    }

                    e.color(EMBED_COLOR);

                    e
                });
                m
            }).await;
        },
        Err(y) => {
            let _ = msg.reply(ctx, y).await;
        }
    }

    Ok(())
}

#[command]
#[description("Show the full draft bank, or filter it according to type or tier.")]
#[aliases("mons")]
#[usage("<draft ID> [type:?] AND/OR [tier:?]")]
#[checks(DraftID2)]
async fn bank(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Bank Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                }
            }
        }
    };
    let filter_1 = args.single::<String>().ok();
    let filter_2 = args.single::<String>().ok();

    match ctx.get_bank(draft_id).await {
        Ok(b) => {
            if filter_1.is_some() {
                let filter_1 = filter_1.unwrap();
                let f1_split: Vec<&str> = filter_1.split(":").collect();
                let filtered = if f1_split.len() == 2 {
                    let key = f1_split[0].to_lowercase();
                    let key = key.as_str();
                    let value = f1_split[1];
                    match key {
                        "tier" => {
                            let tiers: Vec<String> = value.split(",").map(|s| s.to_owned()).collect();
                            let f: Vec<DraftPokemon> = b.iter().filter(|p| tiers.contains(&p.tier)).map(|p| p.to_owned()).collect();
                            f
                        },
                        "type" => {
                            let types: Vec<u8> = value.split(",").map(|s| {
                                let t = Type::from_name(s).map_or(200, |t| t.id) as u8;
                                t
                            }).collect();
                            let f: Vec<DraftPokemon> = b.iter().filter(|p| {
                                types.contains(&p.type1) ||
                                p.type2.is_some() && types.contains(&p.type2.unwrap())
                            }).map(|p| p.to_owned()).collect();
                            f
                        },
                        "move" => {
                            let moves: Vec<i16> = value.split(",").map(|s| {
                                let s = s.to_lowercase();
                                //debug!("{}", s);
                                let m = Move::from_name(s.replace(" ", "-").as_str());
                                //debug!("{:?}", m);
                                let m = m.map_or(-1, |m| m.id) as i16;
                                //debug!("move id {}", m);
                                m
                            }).collect();
                            let reg = ctx.get_moves(21).await;
                            let f: Vec<DraftPokemon> = b.iter().filter(|p| {
                                if reg.contains_key(&p.id) {
                                    for m in &moves {
                                        //debug!("checking move id {} on pokemon {}", m, p.id);
                                        if reg[&p.id].contains(m) {
                                            return true;
                                        }
                                    }
                                    false
                                } else { false }
                            }).map(|p| p.to_owned()).collect();
                            f
                        },
                        _ => {
                            b
                        }
                    }
                }
                else {
                    b
                };
                if filter_2.is_some() {
                    let filter_2 = filter_2.unwrap();
                    let f2_split: Vec<&str> = filter_2.split(":").collect();
                    let filtered = if f2_split.len() == 2 {
                        let key = f2_split[0].to_lowercase();
                        let key = key.as_str();
                        let value = f2_split[1];
                        match key {
                            "tier" => {
                                let tiers: Vec<String> = value.split(",").map(|s| s.to_owned()).collect();
                                let f: Vec<DraftPokemon> = filtered.iter().filter(|p| tiers.contains(&p.tier)).map(|p| p.to_owned()).collect();
                                f
                            },
                            "type" => {
                                let types: Vec<u8> = value.split(",").map(|s| {
                                    let t = Type::from_name(s).map_or(200, |t| t.id) as u8;
                                    t
                                }).collect();
                                let f: Vec<DraftPokemon> = filtered.iter().filter(|p| {
                                    types.contains(&p.type1) ||
                                    p.type2.is_some() && types.contains(&p.type2.unwrap())
                                }).map(|p| p.to_owned()).collect();
                                f
                            },
                            "move" => {
                                let moves: Vec<i16> = value.split(",").map(|s| {
                                    let s = s.to_lowercase();
                                    let m = Move::from_name(s.replace(" ", "-").as_str()).map_or(-1, |m| m.id) as i16;
                                    //debug!("move id {}", m);
                                    m
                                }).collect();
                                let reg = ctx.get_moves(21).await;
                                let f: Vec<DraftPokemon> = filtered.iter().filter(|p| {
                                    if reg.contains_key(&p.id) {
                                        for m in &moves {
                                            //debug!("checking move id {} on pokemon {}", m, p.id);
                                            if reg[&p.id].contains(m) {
                                                return true;
                                            }
                                        }
                                        false
                                    } else { false }
                                }).map(|p| p.to_owned()).collect();
                                f
                            },
                            _ => {
                                filtered
                            }
                        }
                    }
                    else {
                        filtered
                    };

                    let _ = page(ctx, msg, filtered, "Draft Bank (Filtered)".to_owned(), None).await;
                }
                else {
                    let _ = page(ctx, msg, filtered, "Draft Bank (Filtered)".to_owned(), None).await;
                }
            }
            else {
                let _ = page(ctx, msg, b, "Draft Bank".to_owned(), None).await;
            }
        },
        Err(y) => {
            let _ = msg.reply(ctx, y).await;
        }
    }

    Ok(())
}

#[command]
#[description("Show your (or another coach's) current drafted team.")]
#[usage("<draft ID> [@user]")]
#[checks(Joined)]
async fn team(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Team Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                }
            }
        }
    };
    let target = args.single::<UserId>().ok();

    if target.is_some() {
        let target = target.unwrap();

        match ctx.get_team(draft_id.clone(), target.as_u64()).await {
            Ok(b) => {
                let s_id = msg.guild_id.map(|g| g.0);
                let n = ctx.get_coach_name(&draft_id, target.as_u64(), s_id).await;
                let _ = page(ctx, msg, b, format!("{}'s Draft Team", n), ctx.get_coach_logo(&draft_id, target.as_u64()).await).await;
            },
            Err(y) => {
                let _ = msg.reply(ctx, y).await;
            }
        }
    }
    else {
        match ctx.get_team(draft_id.clone(), msg.author.id.as_u64()).await {
            Ok(b) => {
                let _ = page(ctx, msg, b, "Your Draft Team".to_owned(), ctx.get_coach_logo(&draft_id, msg.author.id.as_u64()).await).await;
            },
            Err(y) => {
                let _ = msg.reply(ctx, y).await;
            }
        }
    }

    Ok(())
}

#[command]
#[description("Show a draft's played matches, or a specific coach's matches in that draft.")]
#[usage("<draft ID> [@user]")]
#[checks(DraftID2)]
async fn matches(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Match List Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                }
            }
        }
    };
    let target = args.single::<UserId>().ok().map(|i| i.as_u64().to_owned());

    match ctx.get_matches(draft_id.clone(), target.clone()).await {
        Ok(b) => {
            debug!("got matchvec");
            let dname = ctx.get_draft_name(draft_id.clone()).await;
            let ptitle = if target.is_some() {
                let t = target.unwrap();
                let s_id = msg.guild_id.map(|g| g.0);
                let cname = ctx.get_coach_name(&draft_id, &t, s_id).await;
                format!("{} Matches Involving {}", dname, cname)
            }
            else {
                format!("{} Matches", dname)
            };

            let _ = page(ctx, msg, b, ptitle, None).await;
        },
        Err(y) => {
            let _ = msg.reply(ctx, y).await;
        }
    }

    Ok(())
}

#[command]
#[description("Show a league's ongoing and completed playoff matches, or those where a specific coach is involved.")]
#[usage("<draft ID> [@user]")]
#[checks(DraftID2)]
async fn playoffs(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Playoff List Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                }
            }
        }
    };
    let target = args.single::<UserId>().ok().map(|i| i.as_u64().to_owned());

    match ctx.get_playoffs(draft_id.clone(), target.clone()).await {
        Ok(b) => {
            debug!("got matchvec");
            let dname = ctx.get_draft_name(draft_id.clone()).await;
            let ptitle = if target.is_some() {
                let t = target.unwrap();
                let s_id = msg.guild_id.map(|g| g.0);
                let cname = ctx.get_coach_name(&draft_id, &t, s_id).await;
                format!("{} Playoff Matches Involving {}", dname, cname)
            }
            else {
                format!("{} Playoff Matches", dname)
            };

            let _ = page(ctx, msg, b, ptitle, None).await;
        },
        Err(y) => {
            let _ = msg.reply(ctx, y).await;
        }
    }

    Ok(())
}

#[command]
#[description("Show a draft's enrolled coaches.")]
#[usage("<draft ID> [division #]")]
#[checks(DraftID2)]
async fn coaches(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Coach List Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                }
            }
        }
    };
    let division = args.single::<u8>().ok();

    match ctx.get_coaches(&draft_id, division.clone()).await {
        Ok(b) => {
            let dname = ctx.get_draft_name(draft_id.clone()).await;
            let ptitle = if division.is_some() {
                format!("{} Coaches in Division {}", dname, division.unwrap())
            }
            else {
                format!("{} Coaches", dname)
            };

            let _ = page(ctx, msg, b, ptitle, None).await;
        },
        Err(y) => {
            let _ = msg.reply(ctx, y).await;
        }
    }

    Ok(())
}

#[command]
#[description("Show all of a draft's currently drafted teams, or just those for a certain division.")]
#[usage("<draft ID> [division #]")]
#[checks(DraftID2)]
async fn teams(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Teams List Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                }
            }
        }
    };
    let division = args.single::<u8>().ok();

    let s_id = msg.guild_id.map(|g| g.0);
    match ctx.teams_embed(&draft_id, division.clone(), s_id).await {
        Ok(mut e) => {
            let dname = ctx.get_draft_name(draft_id.clone()).await;
            let ptitle = if division.is_some() {
                format!("{} Teams in Division {}", dname, division.unwrap())
            }
            else {
                format!("{} Teams", dname)
            };
            e.title(ptitle);

            let _ = msg.channel_id.send_message(ctx, |m| {
                m.set_embed(e);
                m
            }).await;
        },
        Err(y) => {
            let _ = msg.reply(ctx, y).await;
        }
    }

    Ok(())
}

#[command]
#[description("Show a match result. (Get the match numbers from g~matches.)")]
#[usage("<draft ID> <match #>")]
#[aliases("match")]
#[checks(DraftID2)]
async fn _match(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Match Info Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                }
            }
        }
    };
    let match_id = match args.single::<u16>() {
        Ok(m) => {
            m
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must specify a match number!").await;
            return Ok(());
        }
    };

    let s_id = msg.guild_id.map(|g| g.0);
    match ctx.match_embed_from_id(&draft_id, match_id, s_id).await {
        Ok(e) => {
            let _ = msg.channel_id.send_message(ctx, |m| {
                m.set_embed(e);
                m
            }).await;
        },
        Err(y) => {
            let _ = msg.reply(ctx, y).await;
        }
    }

    Ok(())
}

#[command]
#[description("Show how many picks you have left in a draft.")]
#[usage("<draft ID>")]
#[checks(Joined)]
async fn picks(ctx: &Context, msg: &Message, args: Args) -> CommandResult {
    info!("Draft Picks Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            m
        },
        Err(y) => {
            match y {
                DraftParseError::Missing => {
                    match msg.guild_id {
                        Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                            Some(d) => d,
                            None => {
                                let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                                return Ok(());
                            }
                        },
                        None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                            Some(d) => d,
                            None => {
                                let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                                return Ok(());
                            }
                        }
                    }
                },
                DraftParseError::Invalid => {
                    let _ = msg.reply(ctx, "That doesn't seem to be a valid draft ID. Check it for misspellings.").await;
                    return Ok(());
                }
            }
        }
    };

    let e = ctx.coach_remaining_picks_embed(&draft_id, msg.author.id.as_u64()).await;
    let _ = msg.channel_id.send_message(ctx, |m| {
        m.set_embed(e);
        m
    }).await;

    Ok(())
}

#[command]
#[description("Trade away one of your drafted Pokémon, either back to the bank or with another coach. (Depends on draft rules)")]
#[usage("<draft ID> bank <your pokemon> <bank pokemon> OR <draft ID> player <@user>")]
#[checks(Joined)]
async fn trade(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Trade Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                }
            }
        }
    };
    let (o_mon, n_mon, bank, target_user) = if let Ok(_type) = args.single::<String>() {
        let _type = _type.as_str();
        match _type {
            "bank" => {
                let u = None;
                let oldmon = match args.single::<String>() {
                    Ok(m) => {
                        m.to_lowercase()
                    },
                    Err(_) => {
                        let _ = msg.reply(ctx, "You must mention a Pokémon species or form to trade away!").await;
                        return Ok(());
                    }
                };
                let newmon = match args.single::<String>() {
                    Ok(m) => {
                        m.to_lowercase()
                    },
                    Err(_) => {
                        let _ = msg.reply(ctx, "You must mention a Pokémon species or form to receive!").await;
                        return Ok(());
                    }
                };
                let o_id = match PokemonForm::from_name_exhaustive(oldmon.as_str()) {
                    Ok(f) => f.id,
                    Err(_) => {
                        let _ = msg.reply(ctx, "That doesn't seem to be a valid Pokémon species. Check for misspellings.").await;
                        return Ok(());
                    }
                };
                let n_id = match PokemonForm::from_name_exhaustive(newmon.as_str()) {
                    Ok(f) => f.id,
                    Err(_) => {
                        let _ = msg.reply(ctx, "That doesn't seem to be a valid Pokémon species. Check for misspellings.").await;
                        return Ok(());
                    }
                };

                (o_id, n_id, true, u)
            },
            "player" => {
                if let Ok(target) = args.single::<UserId>() {
                    let t = Some(target.0);
                    match player_trade(ctx, msg, &draft_id, target).await {
                        Ok((o, n)) => (o, n, false, t),
                        Err(y) => {
                            let _ = msg.reply(ctx, y).await;
                            return Ok(());
                        }
                    }
                }
                else {
                    let _ = msg.reply(ctx, "To make a player trade, you must mention the other player.").await;
                    return Ok(());
                }
            },
            _ => {
                let _ = msg.reply(ctx, "You must specify either a 'bank' or a 'player' trade.").await;
                return Ok(());
            }
        }
    }
    else {
        let _ = msg.reply(ctx, "You must specify either a 'bank' or a 'player' trade.").await;
        return Ok(());
    };

    let trade_preview = {
        let s_id = msg.guild_id.map(|g| g.0);
        let uname1 = ctx.get_coach_name(&draft_id, msg.author.id.as_u64(), s_id.clone()).await;
        let uname2 = if bank { "Bank".into() } else { ctx.get_coach_name(&draft_id, &target_user.clone().unwrap(), s_id).await };
        let mon1 = ctx.get_mon_info(&draft_id, &o_mon).await;
        let mon2 = ctx.get_mon_info(&draft_id, &n_mon).await;

        msg.channel_id.send_message(&ctx, |m| {
            m.embed(|e| {
                e.title("Trade Preview");
                e.description("Look over the details of the trade before confirming or denying.");
                e.field(format!("{}'s Offer", uname1), mon1, false); 
                e.field(format!("{}'s Offer", uname2), mon2, false);

                e.footer(|f| f.text("Confirm or deny the trade using the reactions below."));
                
                e.colour(EMBED_COLOR);

                e
            });
            m
        }).await?
    };

    let _ = trade_preview.react(&ctx, '🆗').await;
    let _ = trade_preview.react(&ctx, '⛔').await;

    let users = if target_user.is_some() {
        [target_user.clone().unwrap(), msg.author.id.0]
    }
    else {
        [msg.author.id.0, 0]
    };

    let mut confirms = 0;
    let ceiling = if bank { 1 } else { 2 };
    let mut canceled = false;

    while confirms < ceiling && !canceled {
        if let Some(reaction) = &trade_preview.await_reaction(&ctx).timeout(Duration::from_secs(300))
        .filter(move |r| {
            let id = r.user_id.expect("no user id on reac");
            let emoji = &r.emoji;
                
            let correct = match emoji.to_string().as_str() {
                "🆗" => {
                    true
                },
                "⛔" => {
                    true
                },
                _ => false
            };
    
            users.contains(id.as_u64()) && correct
        }).await {
            let emoji = &reaction.as_inner_ref().emoji;
            
            match emoji.to_string().as_str() {
                "🆗" => {
                    confirms += 1
                },
                "⛔" => {
                    canceled = true
                },
                _ => (),
            }
        }
    }

    if canceled {
        let _ = msg.reply(ctx, "Your trade was canceled.").await;
        return Ok(());
    }

    if let Ok(a_c) = ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
        let s_id = a_c.clone().guild().map(|gc| gc.guild_id.0);
        match ctx.trade(draft_id.clone(), msg.author.id.as_u64(), o_mon, n_mon, false, target_user, s_id).await {
            Ok((d, n, o, e, _, r)) => {
                let l_name = ctx.get_coach_name(&draft_id, msg.author.id.as_u64(), s_id).await;
                let content = match r {
                    TradePerformance::CoachBank => {
                        format!("**Trade Completed:** {} has exchanged **{}** for **{}** from the bank!", l_name, o, n)
                    },
                    TradePerformance::CoachPlayer => {
                        let r_name = ctx.get_coach_name(&draft_id, &target_user.unwrap(), s_id).await;
                        format!("**Trade Completed:** {} has traded **{}** to {} in return for **{}**!", l_name, o, r_name, n)
                    },
                    TradePerformance::ModBank => {
                        format!("**Trade Completed:** A moderator has pulled **{}** out of the bank in {} for {}!", n, d, l_name)
                    },
                    _ => {
                        format!("")
                    }
                };
                let _ = a_c.id().send_message(ctx, |m| {
                    m.content(content);
                    m.set_embed(e);
                    m
                }).await;
            },
            Err(y) => {
                let _ = msg.reply(ctx, y).await;
            }
        }
    }

    Ok(())
}

async fn player_trade(ctx: &Context, msg: &Message, draft_id: &String, target: UserId) -> Result<(i16, i16), String> {
    if !ctx.check_draft_joined(draft_id.clone(), &target.0).await {
        return Err("Your chosen trade partner is not part of that draft!".into());
    }

    let g = match msg.guild(ctx) {
        Some(g) => g,
        None => return Err("You must make player trades **in servers**!".into())
    };

    let l_name = ctx.get_coach_name(draft_id, msg.author.id.as_u64(), Some(g.id.0)).await;
    //let r_name = ctx.get_coach_name(draft_id, target.as_u64(), Some(g.id.0)).await;

    match msg.channel_id.send_message(ctx, |m| {
            m.content(format!("{}:", target.mention()));
            m.embed(|e| {
                e.title(format!("{} wants to trade with you!", l_name));
                e.description("React with 🆗 or ⛔ to initiate or cancel the trade.");

                e.color(EMBED_COLOR);
                e
            });
            m
        }).await {
        Ok(r_prompt) => {
            let _ = r_prompt.react(&ctx, '🆗').await;
            let _ = r_prompt.react(&ctx, '⛔').await;

            let proceed = if let Some(reaction) = &r_prompt.await_reaction(&ctx).timeout(Duration::from_secs(180)).author_id(target.0).filter(|r| {
                let emoji = &r.emoji;
            
                match emoji.to_string().as_str() {
                    "🆗" => {
                        true
                    },
                    "⛔" => {
                        true
                    },
                    _ => false
                }
            }).await {
                let emoji = &reaction.as_inner_ref().emoji;
            
                match emoji.to_string().as_str() {
                    "🆗" => {
                        true
                    },
                    "⛔" => {
                        false
                    },
                    _ => false
                }
            } else {
                false
            };

            if !proceed {
                return Err("The other player canceled the trade request.".into());
            }
        },
        Err(_) => {
            panic!("Failed to message");
        }
    }
    
    let (mut l_team_embed, l_team_ids) = ctx.team_embed(draft_id, msg.author.id.as_u64(), Some(g.id.0)).await.expect("Need coaches bro");

    l_team_embed.description("Select a Pokémon from your team to offer for trade. **Reply with one of the numbers on the left.**");

    let o_id = match msg.channel_id.send_message(ctx, |m| {
            m.content(format!("{}:", msg.author.mention()));
            m.set_embed(l_team_embed);
            m
        }).await {
        Ok(_l_select) => {
            let index = match msg.author.await_reply(ctx)
                .channel_id(msg.channel_id)
                .timeout(Duration::from_secs(300))
                .filter(|m| {
                    m.content.parse::<u8>().is_ok()
                })
                .await {
                Some(answer) => {
                    let resp = answer.content.parse::<u8>().unwrap();
                    resp - 1
                },
                None => {
                    return Err("Trade offer selection timed out.".into());
                }
            };

            l_team_ids[index as usize]
        },
        Err(_) => {
            panic!("Failed to message");
        }
    };

    let (mut r_team_embed, r_team_ids) = ctx.team_embed(draft_id, target.as_u64(), Some(g.id.0)).await.expect("Need coaches bro");

    r_team_embed.description("Select a Pokémon from your team to offer for trade. **Reply with one of the numbers on the left.**");

    let n_id = match msg.channel_id.send_message(ctx, |m| {
            m.content(format!("{}:", target.mention()));
            m.set_embed(r_team_embed);
            m
        }).await {
        Ok(_r_select) => {
            let t_u = target.to_user(ctx).await.expect("couldn't get target user");
            let index = match t_u.await_reply(ctx)
                .channel_id(msg.channel_id)
                .timeout(Duration::from_secs(300))
                .filter(|m| {
                    m.content.parse::<u8>().is_ok()
                })
                .await {
                Some(answer) => {
                    let resp = answer.content.parse::<u8>().unwrap();
                    resp - 1
                },
                None => {
                    return Err("Trade offer selection timed out.".into());
                }
            };

            r_team_ids[index as usize]
        },
        Err(_) => {
            panic!("Failed to message");
        }
    };
    
    Ok((o_id, n_id))
}
