use std::{collections::{HashMap, HashSet}, fs::File, io::{Read, Write}, path::Path, sync::{Arc, Mutex}, thread, time::Duration};

use chrono::{DateTime, NaiveDateTime, Utc};
use log::*;
use pokerust::{FromName, Pokemon, PokemonForm, PokemonSpecies};
use rayon::prelude::*;
use serenity::{client::Context, framework::standard::{Args, CommandResult, Delimiter, macros::command}, model::{Permissions, channel::{Message, PermissionOverwrite, PermissionOverwriteType}, id::{RoleId, UserId}}, utils::MessageBuilder};

use crate::{DoublesTierList, EMBED_COLOR, PPATierList, SmogonTierList, commands::checks::*, data::{AwardableKey, Draft, DraftPokemon, DraftStatus, DraftType, Format, PokemonFormExt, TierSource, TradePolicy}, func::*};

#[command]
#[description("Start a new draft.")]
#[usage("[key:value]*")]
#[only_in(guilds)]
async fn new(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("New Draft Initiated");
    let mut draft = Draft {
        draft_id: "???".to_owned(),
        draft_name: "Unset".to_owned(),
        logo: None,
        owner_id: msg.author.id.as_u64().clone(),
        moderator_ids: None,
        moderator_role_id: None,
        player_role_id: None,
        announcements_channel: Some(msg.channel_id.as_u64().clone()),
        porygon_channel: None,
        server_id: msg.guild_id.unwrap().as_u64().clone(),
        challonge: None,
        format: Format::Singles,
        tier_source: TierSource::Custom,
        tiers_allowed: Vec::new(),
        league: false,
        league_season: 0,
        total_picks: 6,
        tier_picks: HashMap::new(),
        trade_policy: TradePolicy::Disabled,
        trades_max: 0,
        coaches: HashMap::new(),
        status: DraftStatus::Init,
        draft_type: DraftType::Linear,
        draft_order: Vec::new(),
        draft_round: Vec::new(),
        matches: Vec::new(),
        past_seasons: None,
        mons: HashMap::new(),
        closed: false,
        last_round_drafted: Vec::new(),
        playoff_matches: Vec::new(),
        playoff_best_of: 1,
        division_names: None,
        team_divisions: false,
        next_playoff_letter: "A".to_owned(),
    };
    let mut modify: HashSet<String> = HashSet::new();
    async fn process_kv(ctx: &Context, msg: &Message, key: &str, value: &str, draft: &mut Draft, modify: &mut HashSet<String>) {
        match key {
            "name" | "title" => {
                modify.insert("name".to_owned());
                draft.draft_name = value.to_owned();
            },
            "logo" => {
                if value.starts_with("https:") {
                    modify.insert("logo".to_owned());
                    draft.logo = Some(value.to_owned());
                }
            },
            "id" | "abbreviation" | "acronym" | "abbr" => {
                let v = value.replace(" ", "-");
                let u = v.to_uppercase();
                if !ctx.check_draft_id(&u).await {
                    modify.insert("id".to_owned());
                    draft.draft_id = u;
                }
                else {
                    let _ = msg.reply(ctx, "That Draft ID is already in use. Please choose another.").await;
                }
            },
            "announcements" | "channel" => {
                let v = value.replace("#", "")
                    .replace("<", "")
                    .replace(">", "");
                match v.parse::<u64>() {
                    Ok(c) => {
                        modify.insert("channel".to_owned());
                        draft.announcements_channel = Some(c);
                    },
                    Err(_) => ()
                }
            },/*
            "porygon" => {
                let v = value.replace("#", "")
                    .replace("<", "")
                    .replace(">", "");
                match v.parse::<u64>() {
                    Ok(c) => {
                        let (isc, _) = ctx.is_porygon_channel(msg.guild_id.unwrap().as_u64(), &c).await;
                        if !isc {
                            modify.insert("porygon".to_owned());
                            draft.porygon_channel = Some(c);
                        }
                    },
                    Err(_) => ()
                }
            },
            "mods" | "moderators" | "mod" | "moderator" | "modrole" => {
                let v = value.replace("@", "")
                    .replace("&", "")
                    .replace("<", "")
                    .replace(">", "");
                match v.parse::<u64>() {
                    Ok(c) => {
                        modify.insert("modrole".to_owned());
                        draft.moderator_role_id = Some(c);
                    },
                    Err(_) => ()
                }
            },*/
            "format" => {
                let l = value.to_lowercase();
                match l.as_str() {
                    "single" | "singles" => {
                        modify.insert("format".to_owned());
                        draft.format = Format::Singles;
                    },
                    "double" | "doubles" => {
                        modify.insert("format".to_owned());
                        draft.format = Format::Doubles;
                    },
                    "triple" | "triples" => (),
                    "rotation" => (),
                    _ => (),
                }
            },
            "tiers" | "tier" | "tiersource" => {
                let l = value.to_lowercase();
                match l.as_str() {
                    "smogon" => {
                        modify.insert("tiersource".to_owned());
                        draft.tier_source = TierSource::Smogon;
                    },
                    "custom" => {
                        modify.insert("tiersource".to_owned());
                        draft.tier_source = TierSource::Custom;
                    },
                    "wbe" => {
                        modify.insert("tiersource".to_owned());
                        draft.tier_source = TierSource::WBE;
                    },
                    "ppa" => {
                        modify.insert("tiersource".to_owned());
                        draft.tier_source = TierSource::PPA;
                    },
                    _ => (),
                }
            },
            "legaltiers" | "legal" | "allowed" | "allowedtiers" => {
                if draft.tier_source == TierSource::Smogon || draft.tier_source == TierSource::PPA {
                    let despaced = value.replace(" ", "");
                    let split: Vec<String> = despaced.split(',').map(|s| s.to_owned()).collect();
                    let tiers = ctx.get_tiers(&draft.format, &draft.tier_source).await;
                    let allowed: Vec<String> = tiers.iter().filter(|x| {
                        split.contains(x)
                    }).map(|x| x.to_owned()).collect();
                    draft.tiers_allowed = allowed;
                }
            },
            "bannedtiers" | "banned" => {
                if draft.tier_source == TierSource::Smogon || draft.tier_source == TierSource::PPA {
                    let despaced = value.replace(" ", "");
                    let banned: Vec<String> = despaced.split(',').map(|s| s.to_owned()).collect();
                    let tiers = ctx.get_tiers(&draft.format, &draft.tier_source).await;
                    let allowed: Vec<String> = tiers.iter().filter(|x| {
                        !banned.contains(x)
                    }).map(|x| x.to_owned()).collect();
                    draft.tiers_allowed = allowed;
                }
            },
            "league" => {
                let l = value.to_lowercase();
                match l.as_str() {
                    "true" | "t" | "yes" | "y" => {
                        modify.insert("league".to_owned());
                        draft.league = true;
                    },
                    "false" | "f" | "no" | "n" => {
                        modify.insert("league".to_owned());
                        draft.league = false;
                    },
                    _ => (),
                }
            },
            "tournament" => {
                let l = value.to_lowercase();
                match l.as_str() {
                    "true" | "t" | "yes" | "y" => {
                        modify.insert("league".to_owned());
                        draft.league = false;
                    },
                    "false" | "f" | "no" | "n" => {
                        modify.insert("league".to_owned());
                        draft.league = true;
                    },
                    _ => (),
                }
            },
            "season" => {
                if let Ok(s) = value.parse::<u16>() {
                    modify.insert("season".to_owned());
                    draft.league_season = s;
                }
            },
            "picks" => {
                if let Ok(s) = value.parse::<u8>() {
                    modify.insert("picks".to_owned());
                    draft.total_picks = s;
                }
            },
            "trading" | "tradepolicy" => {
                let l = value.to_lowercase();
                match l.as_str() {
                    "true" | "t" | "yes" | "y" | "free" => {
                        modify.insert("trading".to_owned());
                        draft.trade_policy = TradePolicy::Free;
                    },
                    "bank" => {
                        modify.insert("trading".to_owned());
                        draft.trade_policy = TradePolicy::Bank;
                    },
                    "player" => {
                        modify.insert("trading".to_owned());
                        draft.trade_policy = TradePolicy::Player;
                    },
                    "false" | "f" | "no" | "n" => {
                        modify.insert("trading".to_owned());
                        draft.trade_policy = TradePolicy::Disabled;
                    },
                    _ => (),
                }
            },
            "trades" | "trademax" | "maxtrades" => {
                if let Ok(s) = value.parse::<u8>() {
                    modify.insert("trades".to_owned());
                    draft.trades_max = s;
                }
            },
            "direction" | "order" => {
                let l = value.to_lowercase();
                match l.as_str() {
                    "linear" | "l" => {
                        modify.insert("direction".to_owned());
                        draft.draft_type = DraftType::Linear;
                    },
                    "snake" | "s" => {
                        modify.insert("direction".to_owned());
                        draft.draft_type = DraftType::Snake;
                    },
                    _ => (),
                }
            },
            "closed" => {
                draft.closed = !draft.closed;
            },
            _ => (),
        }
    }
    while let Ok(arg) = args.single_quoted::<String>() {
        let split: Vec<String> = arg.split(':').map(|s| s.to_owned()).collect();
        if split.len() == 2 && !vec!["", " "].contains(&split[1].as_str()) {
            let l = split[0].to_lowercase();
            let key = l.as_str();
            let value = split[1].as_str();
            process_kv(ctx, msg, key, value, &mut draft, &mut modify).await;
        }
    }
    let mut done = false;
    
    while !done {
        let ready = {
            modify.contains("name") &&
            modify.contains("id") &&
            (
                draft.tier_source == TierSource::Custom ||
                draft.tier_source == TierSource::WBE || 
                (
                    draft.tier_source == TierSource::Smogon && !draft.tiers_allowed.is_empty()
                ) || 
                (
                    draft.tier_source == TierSource::PPA && !draft.tiers_allowed.is_empty()
                )
            )
        };
    
        let (desc, foot) = if ready {
            (
                "Your draft is **ready** to be created, but **review the information below *carefully*** before confirming.",
                "Respond to this message with more draft settings in key:value format, or respond \"done\"."
            )
        }
        else {
            (
                "Your draft is still missing some essential information. **Review the information below *carefully***. __Underlined__ fields *must* be filled before the draft is ready.",
                "Respond to this message with more draft settings in key:value format."
            )
        };

        let mut embed = draft.embed(&ctx).await?;

        let message = msg.channel_id.send_message(ctx, |m| {
            embed.title("Draft Creation Process");
            embed.description(desc);
            embed.footer(|f| {
                f.text(foot);
                f
            });
            m.set_embed(embed);
            m
        }).await?;

        // wait for replies
        let r = match msg.author.await_reply(&ctx)
            .channel_id(msg.channel_id)
            .timeout(Duration::from_secs(6000))
            .filter(|m| {
                let c_u = m.content.to_lowercase();
                m.content.contains(":") || c_u == "done".to_owned()
            })
            .await {
            Some(answer) => {
                let resp = answer.content.clone();
                let _ = answer.delete(ctx).await;
                resp
            },
            None => {
                "timeout".to_owned()
            }
        };

        let mut args = Args::new(r.as_str(), &[Delimiter::Single(' ')]);
        
        while let Ok(arg) = args.single_quoted::<String>() {
            let split: Vec<String> = arg.split(':').map(|s| s.to_owned()).collect();
            if split.len() == 2 && !vec!["", " "].contains(&split[1].as_str()) {
                let l = split[0].to_lowercase();
                let key = l.as_str();
                let value = split[1].as_str();
                process_kv(ctx, msg, key, value, &mut draft, &mut modify).await;
            }
            else if split.len() == 1 && split[0].to_lowercase().as_str() == "done" && ready {
                done = true;
            }
        }

        let _ = message.delete(ctx).await?;
    }

    if draft.tier_source == TierSource::Smogon {
        let tiered = {
            let path = Path::new("data/smogon_cache.ron");
            match (File::open(path), draft.format) {
                (Err(_), _) | (_, Format::Doubles) => {
                    let mut prog = msg.channel_id.send_message(ctx, |m| {
                        m.embed(|e| {
                            e.title("Loading Tiered Mons");
                            e.description("Auto-populating draft bank from Smogon tier lists. This may take a while.");
                            e.color(EMBED_COLOR);

                            e
                        });

                        m
                    }).await?;
                    let sm_filter = {
                        let data = ctx.data.read().await;
                        let reg = if draft.format == Format::Singles {
                            data.get::<SmogonTierList>().unwrap()
                        }
                        else {
                            data.get::<DoublesTierList>().unwrap()
                        };
                        let f: HashMap<i16, String> = reg.iter().filter(|(_, y)| {
                            draft.tiers_allowed.contains(y)
                        }).map(|(x, y)| (x.clone(), y.clone())).collect();
                        f
                    };
                    let mut tiered: HashMap<i16, DraftPokemon> = HashMap::new();
                    let mut i = 1;
                    let len = sm_filter.len();
                    debug!("Draft New Tiers:");
                    for (x, y) in sm_filter {
                        debug!("    {}, {}", x, &y);
                        let d = DraftPokemon::new_from_id(x.to_owned(), y.to_owned(), None)?;
                        let _ = prog.edit(ctx, |m| {
                            m.embed(|e| {
                                e.title("Loading Tiered Mons");
                                e.description("Auto-populating draft bank from Smogon tier lists. This may take a while.");

                                e.field("Progress", format!("{} / {}", i, len), false);
                                e.color(EMBED_COLOR);

                                e
                            });

                            m
                        }).await;
                        tiered.insert(d.id.clone(), d);

                        i += 1;
                    }
                    tiered
                },
                (Ok(mut f), Format::Singles) => {
                    let mut ron = String::new();
                    match f.read_to_string(&mut ron) {
                        Err(_) => panic!("failed reading {:?}", path),
                        Ok(_) => {
                            let r: HashMap<i16, DraftPokemon> = ron::from_str(&ron).unwrap_or_default();
                            r.iter().filter(|(_, m)| {
                                draft.tiers_allowed.contains(&m.tier)
                            }).map(|(x, y)| (x.clone(), y.clone())).collect()
                        },
                    }
                }
            }
        };

        draft.mons = tiered;
    }
    else if draft.tier_source == TierSource::PPA {
        let tiered = {
            let path = Path::new("data/ppa_cache.ron");
            match File::open(path) {
                Err(_) => {
                    let mut prog = msg.channel_id.send_message(ctx, |m| {
                        m.embed(|e| {
                            e.title("Loading Tiered Mons");
                            e.description("Auto-populating draft bank from PPA tier lists. This may take a while.");
                            e.color(EMBED_COLOR);
            
                            e
                        });
            
                        m
                    }).await?;
                    let sm_filter = {
                        let data = ctx.data.read().await;
                        let reg = data.get::<PPATierList>().unwrap();
                        let f: HashMap<i16, (String, Option<u8>)> = reg.iter().filter(|(_, (y, _))| {
                            draft.tiers_allowed.contains(y)
                        }).map(|(x, y)| (x.clone(), y.clone())).collect();
                        f
                    };
                    let mut tiered: HashMap<i16, DraftPokemon> = HashMap::new();
                    let mut i = 1;
                    let len = sm_filter.len();
                    debug!("Draft New Tiers:");
                    for (x, (y, b)) in sm_filter {
                        debug!("    {}, {}", x, &y);
                        let d = DraftPokemon::new_from_id(x.to_owned(), y.to_owned(), b)?;
                        let _ = prog.edit(ctx, |m| {
                            m.embed(|e| {
                                e.title("Loading Tiered Mons");
                                e.description("Auto-populating draft bank from PPA tier lists. This may take a while.");
            
                                e.field("Progress", format!("{} / {}", i, len), false);
                                e.color(EMBED_COLOR);
            
                                e
                            });
            
                            m
                        }).await;
                        tiered.insert(d.id.clone(), d);
            
                        i += 1;
                    }
                    tiered
                },
                Ok(mut f) => {
                    let mut ron = String::new();
                    match f.read_to_string(&mut ron) {
                        Err(_) => panic!("failed reading {:?}", path),
                        Ok(_) => {
                            let r: HashMap<i16, DraftPokemon> = ron::from_str(&ron).unwrap_or_default();
                            r.iter().filter(|(_, m)| {
                                draft.tiers_allowed.contains(&m.tier)
                            }).map(|(x, y)| (x.clone(), y.clone())).collect()
                        },
                    }
                }
            }
        };

        draft.mons = tiered;
    }

    let g = msg.guild(ctx).unwrap();

    let (moderator_role, player_role) = {
        let m = g.create_role(ctx, |r| {
            r.mentionable(true).name(format!("{} Moderator", draft.draft_name))
        }).await;
        let m = m.ok();
        if m.is_some() {
            let m_role = m.clone().unwrap();
            let _ = g.member(ctx, msg.author.id).await.expect("Couldn't get Member from User").add_role(ctx, m_role).await;
        }
        let p = g.create_role(ctx, |r| {
            r.mentionable(true).name(format!("{} Player", draft.draft_name))
        }).await;
        (m.map(|r| r.id.0), p.ok().map(|r| r.id.0))
    };
    //let roles_success = moderator_role.is_some() && player_role.is_some();

    draft.moderator_role_id = moderator_role.clone();
    draft.player_role_id = player_role;

    if let Ok(porygon) = g.member(ctx, 692091256477581423).await {
        if let Ok(live_links) = g.create_channel(ctx, |c| {
            c.name(format!("{}-live-links", &draft.draft_id));
            c.topic(format!("Drop live battle links for {} in this channel.", &draft.draft_name));
            c
        }).await {
            let everyone = g.role_by_name("@everyone").unwrap();
            let self_user = ctx.http.get_current_user().await.expect("bot probably broken");

            let mut permissions = vec![
                PermissionOverwrite {
                    allow: Permissions::from_bits(0).unwrap(),
                    deny: Permissions::READ_MESSAGE_HISTORY,
                    kind: PermissionOverwriteType::Role(everyone.id),
                },
                PermissionOverwrite {
                    allow: Permissions::from_bits(76800).unwrap(),
                    deny: Permissions::from_bits(0).unwrap(),
                    kind: PermissionOverwriteType::Member(porygon.user.id),
                },
                PermissionOverwrite {
                    allow: Permissions::from_bits(76800).unwrap(),
                    deny: Permissions::from_bits(0).unwrap(),
                    kind: PermissionOverwriteType::Member(self_user.id),
                }
            ];
            if moderator_role.is_some() {
                let m_id = moderator_role.unwrap();
                permissions.push(
                    PermissionOverwrite {
                        allow: Permissions::from_bits(76800).unwrap(),
                        deny: Permissions::from_bits(0).unwrap(),
                        kind: PermissionOverwriteType::Role(RoleId(m_id)),
                    }
                );
            }
            match g.create_channel(ctx, |c| {
                c.name(format!("{}-ipc", &draft.draft_id));
                c.topic("Porygon puts battle results in here and Genesect reads them. Please do not manually send messages here.");
                c.permissions(permissions);
                c
            }).await {
                Ok(ipc) => {
                    let mb = MessageBuilder::new()
                    .push("porygon, use mode -c ")
                    .mention(&ipc)
                    .build();
    
                    let _ = live_links.say(ctx, &mb).await;
                    thread::sleep(Duration::from_secs(3));
                    let _ = live_links.say(ctx, &draft.draft_name).await;
                    thread::sleep(Duration::from_secs(3));
                    let _ = live_links.say(ctx, "porygon, use rule -spoiler false -format csv -tb true").await;
                    let _ = live_links.say(ctx, "porygon, use rule -spoiler false -format csv -tb true").await;
                    let _ = live_links.say(ctx, "porygon, use rule -spoiler false -format csv -tb true").await;
                    let _ = live_links.say(ctx, &mb).await;
                    let _ = live_links.say(ctx, &mb).await;
    
                    draft.porygon_channel = Some(ipc.id.0);
                },
                Err(why) => {
                    error!("Failed to make IPC channel: {:?}", why);
                    return Ok(());
                }
            }
        }
    }

    {
        draft.status = DraftStatus::Waiting;
        let mut embed = draft.embed(&ctx).await?;

        let _ = ctx.add_draft(draft).await;

        let _ = msg.channel_id.send_message(ctx, |m| {
            embed.title("Draft Creation Finalized");
            embed.description("Your draft has been created. You may manage your draft using the ID you created, and don't forget to **add draftable Pokémon** if you're not using Smogon tiers.");
            embed.footer(|f| {
                f.text("Share the draft ID to your coaches so they can join, or add them manually.");
                f
            });
            m.set_embed(embed);
            m
        }).await?;
    }

    Ok(())
}

#[command]
#[description("Modify draft settings in a draft that you manage.")]
#[usage("<draft ID> [key:value]*")]
#[checks(Manager, DraftID2)]
async fn modify(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Modify Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                }
            }
        }
    };
    
    match ctx.modify_draft(msg, draft_id, args).await {
        Ok(_) => (),
        Err(y) => {
            let _ = msg.reply(ctx, "There was an error in draft modification! Try again later.").await;
            error!("Draft Modify error: {}", y);
        }
    }

    Ok(())
}

#[command]
#[description("Add a draftable Pokémon to a draft that you manage.")]
#[usage("<draft ID> <pokemon> [tier]")]
#[checks(Manager, DraftID2)]
async fn addmon(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Mon Add Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                }
            }
        }
    };
    let mon = match args.single::<String>() {
        Ok(m) => {
            let mut o = m.to_lowercase();
            if !o.contains("unown") {
                if o.ends_with("-g") {
                    o.push_str("alar");
                }
                if o.ends_with("-a") {
                    o.push_str("lola");
                }
                if o.ends_with("-p") {
                    o.push_str("aldea");
                }
                if o.as_str() == "darmanitan-galar" {
                    o = "darmanitan-standard-galar".to_string();
                }
            }
            let o = if o.as_str() == "gastrodon" { "gastrodon-west".to_owned() } 
            else if ["tauros-paldea", "tauros-combat", "tauros-paldea-combat", "tauros-combat-paldea", "tauros-combat-breed"].contains(&o.as_str()) { "tauros-paldea-combat-breed".to_owned() }
            else if ["tauros-blaze", "tauros-paldea-blaze", "tauros-blaze-paldea", "tauros-blaze-breed"].contains(&o.as_str()) { "tauros-paldea-blaze-breed".to_owned() }
            else if ["tauros-aqua", "tauros-paldea-aqua", "tauros-aqua-paldea", "tauros-aqua-breed"].contains(&o.as_str()) { "tauros-paldea-aqua-breed".to_owned() }
            else { o };

            o
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must mention a Pokémon species or form!").await;
            return Ok(());
        }
    };
    let tier = args.single::<String>().unwrap_or("0".to_owned());

    match DraftPokemon::new_from_name(mon, tier.clone(), None) {
        Ok(p) => {
            let dname = ctx.get_draft_name(draft_id.clone()).await;
            match ctx.add_mon_to_draft(draft_id, p.clone()).await {
                Ok(_) => {
                    let _ = msg.reply(ctx, format!("Adding {} in tier {} to draft {}.", p.mon_name(), tier, dname)).await;
                },
                Err(why) => {
                    let _ = msg.reply(ctx, why).await;
                }
            }
        },
        Err(_) => {
            let _ = msg.reply(ctx, "That doesn't seem to be a valid Pokémon. Check for misspellings.").await;
        }
    }

    Ok(())
}

#[command]
#[description("Begin the drafting for a draft that you manage.")]
#[usage("<draft ID>")]
#[checks(Manager, DraftID)]
async fn start(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Start Initiated");
    let draft_id = match args.single::<String>() {
        Ok(m) => {
            let v = m.replace(" ", "-");
            v.to_uppercase()
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must include a draft ID to start it!").await;
            return Ok(());
        }
    };

    match ctx.start_draft(draft_id.clone()).await {
        Ok(dname) => {
            match ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
                Ok(a_c) => {
                    let role = ctx.get_player_role(&draft_id).await;
                    let mut begin_message = MessageBuilder::new();
                    if role.is_some() {
                        let g = msg.guild(ctx).unwrap();
                        let r = g.roles.iter().filter(|(r, _)| r.0 == role.unwrap()).map(|(_, r)| r).next().unwrap();
                        begin_message.mention(r);
                    }
                    begin_message.push_bold(" Drafting in");
                    begin_message.push(" ");
                    begin_message.push_bold_safe(dname);
                    begin_message.push(" ");
                    begin_message.push_bold("has begun!");

                    let _ = a_c.id().say(ctx, begin_message.build()).await;
                },
                Err(why) => {
                    let _ = msg.reply(ctx, why).await;
                },
            }
        },
        Err(why) => {
            let _ = msg.reply(ctx, why).await;
        }
    }

    Ok(())
}

#[command]
#[description("Force a draft pick for the next user.")]
#[only_in(guilds)]
#[usage("<draft ID> <@user> <pokemon>")]
#[checks(Manager, DraftID2)]
async fn modpick(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Modpick Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => {
                    let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                    return Ok(());
                }
            }
        }
    };
    let picking_for = match args.single::<UserId>() {
        Ok(u) => {
            u
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must mention a user to pick for!").await;
            return Ok(());
        }
    };
    let mon = match args.single::<String>() {
        Ok(m) => {
            let mut o = m.to_lowercase();
            if o.ends_with("-g") {
                o.push_str("alar");
            }
            if o.ends_with("-a") {
                o.push_str("lola");
            }
            if o.as_str() == "darmanitan-galar" {
                o = "darmanitan-standard-galar".to_string();
            }

            o
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must mention a Pokémon species or form!").await;
            return Ok(());
        }
    };
    let (m_id, mega_id) = {
        fn mon_and_mega_id(f: PokemonForm) -> (i16, Option<i16>) {
            if f.name.contains("-mega") {
                let s = f.species();
                (s.id, Some(f.id))
            }
            else {
                (f.id, None)
            }
        }

        match PokemonForm::from_name(mon.as_str()) {
            Ok(f) => {
                mon_and_mega_id(f)
            },
            Err(_) => {
                match Pokemon::from_name(mon.as_str()) {
                    Ok(p) => {
                        mon_and_mega_id(p.forms[0].get().unwrap())
                    },
                    Err(_) => {
                        match PokemonSpecies::from_name(mon.as_str()) {
                            Ok(s) => {
                                let p = s.varieties.iter().filter(|v| v.is_default).next().unwrap().pokemon.get()?;
                                mon_and_mega_id(p.forms[0].get().unwrap())
                            },
                            Err(_) => {
                                let _ = msg.reply(ctx, "That doesn't seem to be a valid Pokémon species. Check for misspellings.").await;
                                return Ok(());
                            }
                        }
                    }
                }
            }
        }
    };

    match ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
        Ok(a_c) => {
            let s_id = a_c.clone().guild().map(|gc| gc.guild_id.0);
            match ctx.pick_mon(draft_id.clone(), picking_for.as_u64(), m_id, true, s_id, mega_id).await {
                Ok((d, n, e, p, complete)) => {
                    let u = picking_for.to_user(ctx).await?;
                    let a_nick = u.nick_in(ctx, a_c.clone().guild().unwrap().guild_id).await.unwrap_or(u.name.clone());
                    let _ = a_c.id().send_message(ctx, |m| {
                        m.content(format!("**Round #{}:** A moderator has drafted **{}** for {} in {}!", p, n, a_nick, d.clone()));
                        m.set_embed(e);
                        m
                    }).await;

                    if complete {
                        let role = ctx.get_player_role(&draft_id).await;
                        let mut complete_message = MessageBuilder::new();
                        if role.is_some() {
                            let g = msg.guild(ctx).unwrap();
                            let r = g.roles.iter().filter(|(r, _)| r.0 == role.unwrap()).map(|(_, r)| r).next().unwrap();
                            complete_message.mention(r);
                        }
                        complete_message.push_bold(" Drafting in");
                        complete_message.push(" ");
                        complete_message.push_bold_safe(d);
                        complete_message.push(" ");
                        complete_message.push_bold_line("has concluded!");
                        complete_message.push("Please await instructions from your draft organizers to begin playing matches.");
                        
                        let _ = a_c.id().say(ctx, complete_message.build()).await;
                    }
                },
                Err(y) => {
                    let _ = msg.reply(ctx, y).await;
                }
            }
        },
        Err(y) => {
            error!("Modpick: Failed getting draft announcement channel: {}", y);
        }
    }

    Ok(())
}

#[command]
#[description("Fix a draft pick for a coach, or force a trade (midseason).")]
#[only_in(guilds)]
#[aliases("modtrade")]
#[usage("<draft ID> <@user> <old pokemon> <new pokemon>")]
#[checks(Manager, DraftID2)]
async fn fixpick(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Fixpick Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => {
                    let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                    return Ok(());
                }
            }
        }
    };
    let picking_for = match args.single::<UserId>() {
        Ok(u) => {
            u
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must mention a user to pick for!").await;
            return Ok(());
        }
    };
    let oldmon = match args.single::<String>() {
        Ok(m) => {
            m.to_lowercase()
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must mention a Pokémon species or form to replace!").await;
            return Ok(());
        }
    };
    let newmon = match args.single::<String>() {
        Ok(m) => {
            m.to_lowercase()
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must mention a Pokémon species or form to pick!").await;
            return Ok(());
        }
    };
    let o_id = match PokemonForm::from_name_exhaustive(oldmon.as_str()) {
        Ok(f) => f.id,
        Err(_) => {
            let _ = msg.reply(ctx, "That doesn't seem to be a valid Pokémon species. Check for misspellings.").await;
            return Ok(());
        }
    };
    let n_id = match PokemonForm::from_name_exhaustive(newmon.as_str()) {
        Ok(f) => f.id,
        Err(_) => {
            let _ = msg.reply(ctx, "That doesn't seem to be a valid Pokémon species. Check for misspellings.").await;
            return Ok(());
        }
    };

    if let Ok(a_c) = ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
        let s_id = a_c.clone().guild().map(|gc| gc.guild_id.0);
        match ctx.trade(draft_id.clone(), picking_for.as_u64(), o_id, n_id, true, None, s_id).await {
            Ok((d, n, _, e, _, _)) => {
                
                let u = picking_for.to_user(ctx).await?;
                let a_nick = u.nick_in(ctx, a_c.clone().guild().unwrap().guild_id).await.unwrap_or(u.name.clone());
                let _ = a_c.id().send_message(ctx, |m| {
                    m.content(format!("**Pick Fixed:** A moderator has drafted **{}** for {} in {}!", n, a_nick, d));
                    m.set_embed(e);
                    m
                }).await;
            },
            Err(y) => {
                let _ = msg.reply(ctx, y).await;
            }
        }
    }


    Ok(())
}

#[command]
#[description("Submit a match result (Porygon output, or manually created) to the draft record.")]
#[usage("<draft ID>")]
#[checks(Manager, DraftID)]
async fn submit(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Match Submission Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            m
        },
        Err(y) => {
            match y {
                DraftParseError::Missing => {
                    match msg.guild_id {
                        Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                            Some(d) => d,
                            None => {
                                let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                                return Ok(());
                            }
                        },
                        None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                            Some(d) => d,
                            None => {
                                let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                                return Ok(());
                            }
                        }
                    }
                },
                DraftParseError::Invalid => {
                    let _ = msg.reply(ctx, "That doesn't seem to be a valid draft ID. Check it for misspellings.").await;
                    return Ok(());
                }
            }
        }
    };
    args.advance();
    let repair = if let Ok(r) = args.single::<String>() {
        r.to_lowercase().as_str() == "repair"
    }
    else {
        false
    };
    
    let submit_prompt = msg.reply(ctx, "Please respond with a match result exactly as it was output by Porygon in CSV format and with spoilers off.\n**Include *all* original formatting for best results.**").await?;
    let content = match msg.author.await_reply(&ctx)
        .channel_id(msg.channel_id)
        .timeout(Duration::from_secs(6000))
        .filter(|m| {
            let c_u = m.content.to_lowercase();
            m.content.contains("\n") || c_u == "cancel".to_owned()
        })
        .await {
        Some(answer) => {
            let resp = answer.content.clone();
            let _ = answer.delete(ctx).await;
            let _ = submit_prompt.delete(ctx).await;
            resp
        },
        None => {
            "timeout".to_owned()
        }
    };
    
    let mut the_match = match ctx.create_match(draft_id.clone(), content).await {
        Ok(the_match) => {
            the_match
        },
        Err(y) => {
            let _ = msg.reply(ctx, y).await;
            return Ok(());
        }
    };

    let mut done = false;
    let mut denied = false;
    
    while !done {
        let s_id = msg.guild_id.map(|g| g.0);
        let e = ctx.match_embed(&draft_id, &the_match, s_id).await;
        
        let match_preview = msg.channel_id.send_message(ctx, |m| {
            m.content("The match results were parsed successfully. Review it below before replying with 'confirm' or 'deny', or change the match time with 'time:YYYY/MM/DD@HH:MM'.");
            m.set_embed(e);
            m
        }).await?;

        //let from = NaiveDateTime::parse_from_str(&start_date, "%Y-%m-%dT%H:%M:%S")?;
        //let from_utc = DateTime::<Utc>::from_utc(from, Utc);
        // wait for replies
        let r = match msg.author.await_reply(&ctx)
            .channel_id(msg.channel_id)
            .timeout(Duration::from_secs(6000))
            .filter(|m| {
                let c_u = m.content.to_lowercase();
                m.content.contains(":") || c_u == "confirm".to_owned() || c_u == "deny".to_owned()
            })
            .await {
            Some(answer) => {
                let resp = answer.content.clone();
                let _ = answer.delete(ctx).await;
                resp
            },
            None => {
                "timeout".to_owned()
            }
        };

        let mut args = Args::new(r.as_str(), &[Delimiter::Single(' ')]);
        
        while let Ok(arg) = args.single_quoted::<String>() {
            let split: Vec<String> = arg.split(':').map(|s| s.to_owned()).collect();
            if split.len() == 3 && !vec!["", " "].contains(&split[1].as_str()) {
                let l = split[0].to_lowercase();
                let key = l.as_str();
                let value = format!("{}:{}", split[1], split[2]);
                info!("Match Submission: Trying to set {} to {}", key, value);
                if key == "time" {
                    match NaiveDateTime::parse_from_str(value.as_str(), "%Y/%m/%d@%H:%M") {
                        Ok(d) => {
                            let dt = DateTime::<Utc>::from_utc(d, Utc);
                            debug!("Match: {}, {}", d, dt);
                            the_match.date = dt;
                        },
                        Err(_) => {
                            let _ = msg.reply(ctx, "Error parsing match date and time.").await;
                        }
                    }
                }
            }
            else if split.len() == 1 && split[0].to_lowercase().as_str() == "confirm" {
                done = true;
            }
            else if split.len() == 1 && split[0].to_lowercase().as_str() == "deny" {
                done = true;
                denied = true;
            }
        }

        let _ = match_preview.delete(ctx).await?;
    }

    if !denied {
        match ctx.commit_match(&draft_id, the_match.clone(), repair).await {
            Ok(p) => {
                match ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
                    Ok(a_c) => {
                        if p.is_some() {
                            the_match.playoffs = true;
                        }
                        let s_id = a_c.clone().guild().map(|gc| gc.guild_id.0);
                        let e = ctx.match_embed(&draft_id, &the_match, s_id).await;
                        let _ = a_c.id().send_message(ctx, |m| {
                            m.content("Match submitted successfully.");
                            m.set_embed(e);
                            m
                        }).await;
                        if p.is_some() {
                            let p = p.as_ref().unwrap();
                            let e = ctx.playoff_match_embed(p, &draft_id, s_id).await;
                            let _ = a_c.id().send_message(ctx, |m| {
                                m.set_embed(e);
                                m
                            }).await;
                        }
                    }
                    Err(y) => {
                        error!("Porygon: Failed getting draft announcement channel: {}", y);
                    }
                }
            },
            Err(y) => {
                let _ = msg.reply(ctx, y).await;
            }
        }
    }
    else {
        let _ = msg.reply(ctx, "You canceled the match submission.").await;
    }
    

    Ok(())
}

#[command]
#[description("Add a coach to a draft.")]
#[usage("<draft ID> <@user> [division (NUMBER 1-255)]")]
#[only_in(guilds)]
#[checks(Manager, DraftID2)]
async fn add(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Coach Drag Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => {
                    let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                    return Ok(());
                }
            }
        }
    };
    let dragging = match args.single::<UserId>() {
        Ok(u) => {
            u.to_user(ctx).await?
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must mention a user to pick for!").await;
            return Ok(());
        }
    };
    let division = args.single::<u8>().ok().unwrap_or(0);

    match ctx.add_coach_to_draft(&draft_id, dragging.id.0, dragging.name.clone(), division).await {
        Ok((dname, pos, p)) => {
            if p.is_some() {
                let r = p.unwrap();
                let g = msg.guild(ctx).unwrap();
                let mut m = g.member(ctx, dragging.id).await.expect("Couldn't get Member from User");
                let _ = m.add_role(ctx, r).await;
            }
            let _ = msg.reply(ctx, format!("You've added {} to {}. They are position #{} in the drafting order.", dragging.name.clone(), dname, pos)).await;
            let _ = dragging.dm(ctx, |m| m.content(format!("You've been added to {}. You are position #{} in the drafting order. You will be notified when your tournament organizers start the drafting.", dname, pos))).await;
        },
        Err(why) => {
            let _ = msg.reply(ctx, why).await;
        }
    }

    Ok(())
}

#[command]
#[description("Set the number of picks allowed per tier and/or per division. Settings can be specified in the command or after a prompt.")]
#[usage("<draft ID> [tier:picks] OR [tier:picks|tier:picks|...] OR [div:tier:picks] OR [div|tier:picks|tier:picks|...]")]
#[checks(Manager, DraftID2)]
async fn setpicks(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Set Picks Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                }
            }
        }
    };
    let mut settings = {
        let mut s: Vec<String> = Vec::new();
        while let Ok(arg) = args.single_quoted::<String>() {
            s.push(arg);
        }
        s
    };

    let draft_status = ctx.get_draft_status(&draft_id).await;
    if !vec![DraftStatus::Waiting, DraftStatus::Ready].contains(&draft_status) {
        let _ = msg.reply(ctx, "You can't change a draft's allowed picks after drafting has already started!").await;
        return Ok(());
    }

    let mut tier_picks = ctx.get_picks(&draft_id).await;
    
    async fn process_settings(ctx: &Context, draft_id: &String, settings: &Vec<String>, tier_picks: &mut HashMap<u8, HashMap<std::string::String, u8>>) -> HashMap<u8, HashMap<String, u8>> {
        let old_picks = tier_picks.clone();

        if settings.len() > 0 {
            for s in settings {
                let mut split1: Vec<String> = s.split('|').map(|s| s.to_owned()).collect();
                if split1.len() > 1 {
                    let div = if !split1[0].contains(':') {
                        let div = split1.remove(0).parse::<u8>().ok();
                        div
                    }
                    else {
                        Some(0)
                    };
    
                    if div.is_some() {
                        let div = div.unwrap();
                        let mut new_picks: HashMap<String, u8> = HashMap::new();
    
                        for set in split1 {
                            let split2: Vec<String> = set.split(':').map(|s| s.to_owned()).collect();
                            if split2.len() == 2 {
                                let tier = split2[0].clone();
                                if ctx.check_tier(draft_id, &tier).await {
                                    let picks = split2[1].parse::<u8>().ok();
                                    if picks.is_some() {
                                        let picks = picks.unwrap();
                                        new_picks.insert(tier, picks);
                                    }
                                }
                            }
                        }
    
                        if new_picks.len() > 0 {
                            tier_picks.insert(div, new_picks);
                        }
                    }
                }
                else if split1.len() == 1 {
                    let mut split2: Vec<String> = split1[0].split(':').map(|s| s.to_owned()).collect();
                    let div = if split2.len() == 3 {
                        let div = split2.remove(0).parse::<u8>().ok();
                        div
                    }
                    else if split2.len() == 2 {
                        Some(0)
                    }
                    else {
                        None
                    };
                    
                    if div.is_some() {
                        let div = div.unwrap();
                        let mut new_picks: HashMap<String, u8> = HashMap::new();
    
                        let tier = split2[0].clone();
                        if ctx.check_tier(draft_id, &tier).await {
                            let picks = split2[1].parse::<u8>().ok();
                            if picks.is_some() {
                                let picks = picks.unwrap();
                                new_picks.insert(tier, picks);
                            }
                        }
    
                        if new_picks.len() > 0 {
                            tier_picks.insert(div, new_picks);
                        }
                    }
                }
            }
        }

        old_picks
    }

    let mut done = false;
    let mut denied = false;

    while !done {
        let _old_picks = process_settings(ctx, &draft_id, &settings, &mut tier_picks).await;
    
        let mut picks_embed = ctx.make_picks_embed(&draft_id, &tier_picks).await;
        picks_embed.description("The changes are not committed yet. Review them below before replying with more settings, 'confirm', or 'deny'.");
    
        let prompt = msg.channel_id.send_message(ctx, |m| {
            m.set_embed(picks_embed);
            m
        }).await?;

        let r = match msg.author.await_reply(&ctx)
            .channel_id(msg.channel_id)
            .timeout(Duration::from_secs(6000))
            .filter(|m| {
                let c_u = m.content.to_lowercase();
                m.content.contains(":") || c_u == "confirm".to_owned() || c_u == "deny".to_owned()
            })
            .await {
            Some(answer) => {
                let resp = answer.content.clone();
                let _ = answer.delete(ctx).await;
                resp
            },
            None => {
                "timeout".to_owned()
            }
        };

        let l = r.to_lowercase();
        match l.as_str() {
            "confirm" => {
                done = true;
            },
            "deny" => {
                done = true;
                denied = true;
            },
            _ => ()
        }

        if r.contains(':') {
            settings = r.split(' ').map(|s| s.to_owned()).collect();
        }

        let _ = prompt.delete(ctx).await;
    }

    if !denied {
        //debug!("pre-commit");
        let e = ctx.commit_picks(&draft_id, tier_picks).await;
        //debug!("post-commit");
        
        let _ = msg.channel_id.send_message(ctx, |m| {
            m.content("Picks per tier and division updated successfully.");
            m.set_embed(e);
            m
        }).await;
    }
    else {
        let _ = msg.reply(ctx, "You canceled the pick settings.").await;
    }

    Ok(())
}

#[command]
#[description("Set a default draft ID for this server. Allows your players/mods to run commands on that draft without necessarily including its ID each time.")]
#[usage("<draft ID>")]
#[checks(DMsOrAdmin, DraftID)]
async fn default(ctx: &Context, msg: &Message, args: Args) -> CommandResult {
    info!("Default ID Setting Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            m
        },
        Err(y) => {
            match y {
                DraftParseError::Missing => {
                    match msg.guild_id {
                        Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                            Some(d) => d,
                            None => {
                                let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                                return Ok(());
                            }
                        },
                        None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                            Some(d) => d,
                            None => {
                                let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                                return Ok(());
                            }
                        }
                    }
                },
                DraftParseError::Invalid => {
                    let _ = msg.reply(ctx, "That doesn't seem to be a valid draft ID. Check it for misspellings.").await;
                    return Ok(());
                }
            }
        }
    };

    if msg.is_private() {
        let old = ctx.set_default_draft(msg.author.id.as_u64(), &draft_id).await;
        let _ = msg.channel_id.send_message(ctx, |m| {
            m.embed(|e| {
                e.title(format!("Default Draft for DMs Updated"));
                e.description(format!("{} => {}", old.unwrap_or("None".to_string()), &draft_id));
    
                e.color(EMBED_COLOR);
    
                e
            });
            m
        }).await;
    }
    else {
        let old = ctx.set_default_draft(msg.guild_id.unwrap().as_u64(), &draft_id).await;
        let s_name = msg.guild_id.unwrap().name(ctx).unwrap();
        let _ = msg.channel_id.send_message(ctx, |m| {
            m.embed(|e| {
                e.title(format!("Default Draft for {} Updated", s_name));
                e.description(format!("{} => {}", old.unwrap_or("None".to_string()), &draft_id));
    
                e.color(EMBED_COLOR);
    
                e
            });
            m
        }).await;
    }

    Ok(())
}

#[command]
#[description("Cache draft mons for a given tiersource (Smogon or PPA). Bot owners only.")]
#[owners_only]
async fn cache(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    let source = if let Ok(s) = args.single::<String>() {
        let l = s.to_lowercase();
        if vec!["smogon", "ppa"].contains(&l.as_str()) {
            match l.as_str() {
                "smogon" => TierSource::Smogon,
                "ppa" => TierSource::PPA,
                _ => TierSource::Custom,
            }
        }
        else {
            let _ = msg.reply(ctx, "Caching must be done based on either Smogon or PPA tiers!").await;
            return Ok(());
        }
    }
    else {
        TierSource::PPA
    };

    let cachemap: HashMap<i16, DraftPokemon> = HashMap::new();
    let cachemap = Arc::new(Mutex::new(cachemap));

    match source {
        TierSource::PPA => {
            let ppa_tiers = {
                let data = ctx.data.read().await;
                let reg = data.get::<PPATierList>().unwrap();
                reg.clone()
            };
            let mut prog = msg.channel_id.send_message(ctx, |m| {
                m.embed(|e| {
                    e.title("Caching Tiered Mons");
                    e.description("Caching draft bank from PPA tier lists. This may take a while.");
                    e.color(EMBED_COLOR);

                    e
                });

                m
            }).await?;
            let len = ppa_tiers.len();

            ppa_tiers.par_iter().for_each(|(id, (tier, ability_ban))| {
                match DraftPokemon::new_from_id(*id, tier.clone(), *ability_ban) {
                    Ok(mon) => {
                        cachemap.lock().unwrap().insert(*id, mon);
                    }
                    Err(why) => {
                        warn!("Error creating draft mon: form #{}", id);
                        warn!("    {:?}", why);
                    },
                }
                let cur = {
                    let c = cachemap.lock().unwrap();
                    let c = c.len();
                    c
                };
                debug!("Smogon caching: {} / {}", cur, len);
            });

            let cachemap = Arc::try_unwrap(cachemap).unwrap().into_inner().unwrap();

            let string = ron::ser::to_string_pretty(&cachemap, ron::ser::PrettyConfig::new()).unwrap();

            let path = Path::new("data/ppa_cache.ron");
            let display = path.display();

            let mut file = match File::create(&path) {
                Err(why) => panic!("couldn't create {}: {}", display, why),
                Ok(file) => file,
            };

            match file.write_all(string.as_bytes()) {
                Err(why) => panic!("couldn't write to {}: {}", display, why),
                Ok(_) => debug!("successfully wrote to {}", display),
            }

            let _ = prog.edit(ctx, |m| {
                m.embed(|e| {
                    e.title("Caching Tiered Mons");
                    e.description("Caching complete.");
                    e.color(EMBED_COLOR);

                    e
                });

                m
            }).await?;
        },
        TierSource::Smogon => {
            let smogon_tiers: HashMap<i16, String> = {
                let data = ctx.data.read().await;
                let reg = data.get::<SmogonTierList>().unwrap();
                reg.par_iter().filter(|(_, t)| t.as_str() != "Illegal").map(|(i, t)| (*i, t.clone())).collect()
            };
            let mut prog = msg.channel_id.send_message(ctx, |m| {
                m.embed(|e| {
                    e.title("Caching Tiered Mons");
                    e.description("Caching draft bank from Smogon tier lists. This may take a while.");
                    e.color(EMBED_COLOR);

                    e
                });

                m
            }).await?;
            let len = smogon_tiers.len();

            smogon_tiers.par_iter().for_each(|(id, tier)| {
                match DraftPokemon::new_from_id(*id, tier.clone(), None) {
                    Ok(mon) => {
                        cachemap.lock().unwrap().insert(*id, mon);
                    }
                    Err(why) => {
                        warn!("Error creating draft mon: form #{}", id);
                        warn!("    {:?}", why);
                    },
                }
                let cur = {
                    let c = cachemap.lock().unwrap();
                    let c = c.len();
                    c
                };
                debug!("Smogon caching: {} / {}", cur, len);
            });

            let cachemap = Arc::try_unwrap(cachemap).unwrap().into_inner().unwrap();

            let string = ron::ser::to_string_pretty(&cachemap, ron::ser::PrettyConfig::new()).unwrap();

            let path = Path::new("data/smogon_cache.ron");
            let display = path.display();

            let mut file = match File::create(&path) {
                Err(why) => panic!("couldn't create {}: {}", display, why),
                Ok(file) => file,
            };

            match file.write_all(string.as_bytes()) {
                Err(why) => panic!("couldn't write to {}: {}", display, why),
                Ok(_) => debug!("successfully wrote to {}", display),
            }

            let _ = prog.edit(ctx, |m| {
                m.embed(|e| {
                    e.title("Caching Tiered Mons");
                    e.description("Caching complete.");
                    e.color(EMBED_COLOR);

                    e
                });

                m
            }).await?;
        },
        _ => todo!("")
    }

    Ok(())
}

#[command]
#[description("Give moderator permissions in a draft to a given user. (Useful if you are running a draft in a server where you do not, yourself, have Manage Roles permissions.)")]
#[usage("<draft ID> <@user>*")]
#[only_in(guilds)]
#[checks(Manager, DraftID2)]
async fn moderator(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Moderator Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => {
                    let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                    return Ok(());
                }
            }
        }
    };
    if let Some(mod_role) = ctx.get_mod_role(draft_id.clone()).await {
        let mut result = Ok(());
        let mut total = 0;
        let mut rm_total = 0;
        let g = msg.guild(ctx).expect("Draft Moderator Add: Expected Guild");
        while let Ok(u) = args.single::<UserId>() {
            if let Ok(mut m) = g.member(ctx, u).await {
                if !m.user.has_role(ctx, g.id, mod_role).await.expect("error checking role have") {
                    if let Ok(_) = m.add_role(ctx, mod_role).await {
                        info!("{}: Added mod role to {}", &draft_id, u.0);
                        total += 1;
                    }
                    else {
                        warn!("Draft Moderator Add: failed to add mod role to {}", u.0);
                        result = Err("I couldn't add/remove the moderator role to/from one or more users. This probably means I don't have the **Manage Roles** permission. Please tell the server administrators to give me that permission.");
                    }
                }
                else {
                    if let Ok(_) = m.remove_role(ctx, mod_role).await {
                        info!("{}: Removed mod role from {}", &draft_id, u.0);
                        rm_total += 1;
                    }
                    else {
                        warn!("Draft Moderator Remove: failed to remove mod role from {}", u.0);
                        result = Err("I couldn't add/remove the moderator role to/from one or more users. This probably means I don't have the **Manage Roles** permission. Please tell the server administrators to give me that permission.");
                    }
                }
            }
            else {
                warn!("Draft Moderator Add: failed to get guild member for {}", u.0);
            }
        }

        if let Err(y) = result {
            let _ = msg.reply(ctx, y).await;
        }
        else {
            let _ = msg.reply(ctx, format!("{} moderators successfully added.\n{} moderators successfully removed.", total, rm_total)).await;
        }
    }
    else {
        let _ = msg.reply(ctx, "There is no moderator role associated with this draft. This means only the draft owner may manage it. The draft owner may manually create and associate a moderator role with the draft by using the `dm modify` command.").await;
    }

    Ok(())
}

#[command]
#[description("Give extra trades to a coach in a draft.")]
#[usage("<draft ID> <@user> [# extra trades (default 1)]")]
#[only_in(guilds)]
#[checks(Manager, DraftID2)]
async fn bonustrade(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Bonus Trades Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => {
                    let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                    return Ok(());
                }
            }
        }
    };
    let target = match args.single::<UserId>() {
        Ok(u) => u,
        Err(_) => {
            let _ = msg.reply(ctx, "You must mention a user to give them extra trades!").await;
            return Ok(());
        }
    };
    let bonus = match args.single::<i8>() {
        Ok(x) => x,
        Err(_) => {
            1
        }
    };

    let new_limit = ctx.add_coach_trade(&draft_id, target.as_u64(), bonus).await;

    let _ = msg.channel_id.say(ctx, format!("Coach trade limit updated to {}", new_limit)).await;


    Ok(())
}

#[command]
#[description("Give extra picks to a coach in a draft. Particularly good if you want to give certain coaches more picks in a tier without involving division numbers.")]
#[usage("<draft ID> <@user> <tier name>[:# of bonus picks (default 1)]")]
#[only_in(guilds)]
#[checks(Manager, DraftID2)]
async fn bonuspick(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Bonus Picks Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => {
                    let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                    return Ok(());
                }
            }
        }
    };
    let target = match args.single::<UserId>() {
        Ok(u) => u,
        Err(_) => {
            let _ = msg.reply(ctx, "You must mention a user to give them extra picks!").await;
            return Ok(());
        }
    };
    let bonus = match args.single::<String>() {
        Ok(x) => x,
        Err(_) => {
            let _ = msg.reply(ctx, "You must include 'tier[:picks]' to give to the user!").await;
            return Ok(());
        }
    };

    let parse_result = {
        let s = bonus.split(':').map(|s| s.to_owned()).collect::<Vec<String>>();
        if s.len() >= 1 {
            if ctx.check_tier(&draft_id, &s[0]).await {
                let t = s[0].clone();
                let i = if s.len() >= 2 {
                    if let Ok(i) = s[1].parse::<i8>() {
                        Ok(i)
                    }
                    else {
                        Err("Failed to parse number of bonus picks")
                    }
                }
                else { Ok(1) };
                match i {
                    Ok(i) => {
                        let new_pick = ctx.add_coach_pick(&draft_id, target.as_u64(), &t, i).await;

                        Ok((t, new_pick))
                    },
                    Err(s) => {
                        Err(s)
                    }
                }
            }
            else {
                Err("That tier name is invalid or banned in this draft")
            }
        }
        else {
            Err("Something went wrong")
        }
    };

    match parse_result {
        Ok((t, l)) => {
            let _ = msg.channel_id.say(ctx, format!("Coach now has {} picks left in Tier {}", l, t)).await;
        },
        Err(y) => {
            let _ = msg.channel_id.say(ctx, format!("{}", y)).await;
        }
    }


    Ok(())
}

#[command]
#[description("Allow a coach to pick (more) Megas.")]
#[usage("<draft ID> <@user> [# of Megas (default 1)]")]
#[only_in(guilds)]
#[checks(Manager, DraftID2)]
async fn bonusmega(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Bonus Megas Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => {
                    let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                    return Ok(());
                }
            }
        }
    };
    let target = match args.single::<UserId>() {
        Ok(u) => u,
        Err(_) => {
            let _ = msg.reply(ctx, "You must mention a user to give them extra picks!").await;
            return Ok(());
        }
    };
    let bonus = match args.single::<i8>() {
        Ok(x) => x,
        Err(_) => {
            1
        }
    };
    
    let new_limit = ctx.add_coach_mega(&draft_id, target.as_u64(), bonus).await;

    let _ = msg.channel_id.say(ctx, format!("Coach Mega limit updated to {}", new_limit)).await;


    Ok(())
}

#[command]
#[description("Give an award to a coach or a Pokémon.")]
#[usage("<draft ID> <@user OR mon name> <award title>")]
#[checks(Manager, DraftID2)]
async fn award(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Award Initiated");
    let draft_id = match ctx.filter_draft_id(args.current()).await {
        Ok(m) => {
            args.advance();
            m
        },
        Err(_) => {
            match msg.guild_id {
                Some(g) => match ctx.get_default_draft(g.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                },
                None => match ctx.get_default_draft(msg.author.id.as_u64()).await {
                    Some(d) => d,
                    None => {
                        let _ = msg.reply(ctx, "You must specify a draft ID!").await;
                        return Ok(());
                    }
                }
            }
        }
    };
    let target = match args.parse::<UserId>() {
        Ok(u) => AwardableKey::Coach(u.0),
        Err(_) => match args.parse::<String>() {
            Ok(m) => {
                let mut o = m.to_lowercase();
                if o.ends_with("-g") {
                    o.push_str("alar");
                }
                if o.ends_with("-a") {
                    o.push_str("lola");
                }
                if o.as_str() == "darmanitan-galar" {
                    o = "darmanitan-standard-galar".to_string();
                }

                match PokemonForm::from_name_exhaustive(&*o) {
                    Ok(f) => AwardableKey::Mon(f.id),
                    Err(_) => {
                        let _ = msg.reply(ctx, "That doesn't seem to be a valid Pokémon species. Check for misspellings.").await;
                        return Ok(());
                    }
                }
            },
            Err(_) => {
                let _ = msg.reply(ctx, "You must mention a coach or a Pokémon in the draft!").await;
                return Ok(());
            }
        }
    };
    args.advance();
    let award: String = args.rest().into();
    if award.len() == 0 {
        let _ = msg.reply(ctx, "You must supply a name for the award you are giving!").await;
        return Ok(());
    }

    match ctx.add_award(&draft_id, target, award.clone()).await {
        Ok(t_name) => {
            let _ = msg.reply(ctx, format!("Award **{}** has been granted to **{}**", award, t_name)).await;
        }
        Err(y) => {
            let _ = msg.reply(ctx, y).await;
            return Ok(());
        }
    }


    Ok(())
}

#[command]
#[description("Move the current league season into the playoffs phase, creating a new Challonge bracket.")]
#[usage("<draft ID> [days to wait until bracket opens (default 7)]")]
#[checks(Manager, DraftID)]
async fn playoff(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Playoffs Initiated");
    let draft_id = match args.single::<String>() {
        Ok(m) => {
            let v = m.replace(" ", "-");
            v.to_uppercase()
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must include a draft ID to finish it!").await;
            return Ok(());
        }
    };
    let days = args.single::<u8>().ok();

    match ctx.start_playoffs(&draft_id, days).await {
        Ok(r) => {
            if let Ok(a_c) = ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
                let role = ctx.get_player_role(&draft_id).await;
                let mut begin_message = MessageBuilder::new();
                if role.is_some() {
                    let g = msg.guild(ctx).unwrap();
                    let r = g.roles.iter().filter(|(r, _)| r.0 == role.unwrap()).map(|(_, r)| r).next().unwrap();
                    begin_message.mention(r);
                }
                begin_message.push_line("");
                begin_message.push(r);
                let _ = a_c.id().say(ctx, begin_message.build()).await;
            }
        },
        Err(y) => {
            let _ = msg.reply(ctx, y).await;
        }
    }

    Ok(())
}

#[command]
#[description("Finalize results for either the current season, in a league draft, or overall results in a tournament draft.")]
#[usage("<draft ID>")]
#[aliases("finalize", "endseason")]
#[checks(Manager, DraftID)]
async fn finish(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Finalize Initiated");
    let draft_id = match args.single::<String>() {
        Ok(m) => {
            let v = m.replace(" ", "-");
            v.to_uppercase()
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must include a draft ID to finish it!").await;
            return Ok(());
        }
    };

    match ctx.end_season(&draft_id).await {
        Ok((r, e)) => {
            if let Ok(a_c) = ctx.http.get_channel(ctx.get_announce_channel_id(draft_id.clone()).await).await {
                let _ = a_c.id().send_message(ctx, |m| {
                    m.content(r);
                    if e.is_some() {
                        m.set_embed(e.unwrap());
                    }
                    m
                }).await;
            }
        },
        Err(y) => {
            let _ = msg.reply(ctx, y).await;
        }
    }

    Ok(())
}

#[command]
#[description("Begin a new season of a league draft. (Requires that a season is not already in progress: use the `finish` command first.)")]
#[usage("<draft ID>")]
#[checks(Manager, DraftID)]
async fn newseason(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft New Season Initiated");
    let draft_id = match args.single::<String>() {
        Ok(m) => {
            let v = m.replace(" ", "-");
            v.to_uppercase()
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must include a draft ID to begin a new season!").await;
            return Ok(());
        }
    };

    match ctx.new_season(&draft_id).await {
        Ok(r) => {
            let mut resp = MessageBuilder::new();
            if let Some(role) = ctx.get_mod_role(draft_id.clone()).await {
                resp.mention(&RoleId(role));
                resp.push_line(":");
            }
            resp.push_line(r);
            resp.push_line("Moderators should review the list of bank Pokémon and their tiers, and then start adding coaches back to the draft.");

            let _ = msg.reply(ctx, resp.build()).await;
        },
        Err(y) => {
            let _ = msg.reply(ctx, y).await;
        }
    }

    Ok(())
}

#[command]
#[description("Mark (or unmark) a coach as a drop-out, excluding them from playoff seeding.")]
#[usage("<draft ID> <@user>")]
#[checks(Manager, DraftID)]
async fn dropout(ctx: &Context, msg: &Message, mut args: Args) -> CommandResult {
    info!("Draft Dropout Initiated");
    let draft_id = match args.single::<String>() {
        Ok(m) => {
            let v = m.replace(" ", "-");
            v.to_uppercase()
        },
        Err(_) => {
            let _ = msg.reply(ctx, "You must include a draft ID to finish it!").await;
            return Ok(());
        }
    };
    let target = match args.single::<UserId>() {
        Ok(u) => u,
        Err(_) => {
            let _ = msg.reply(ctx, "You must mention a user to mark them as a drop-out!").await;
            return Ok(());
        }
    };

    let (d, n) = ctx.set_coach_dropout(&draft_id, target.as_u64()).await;
    let mes = if d {
        format!("{} is now marked as a drop-out.", n)
    }
    else {
        format!("{} is NO LONGER marked as a drop-out.", n)
    };
    let _ = msg.reply(ctx, mes).await;

    Ok(())
}
