use core::{f64, panic};
use std::{collections::{HashMap, HashSet}, env, f64::consts, fmt::{self, Display}, fs::{self, File}, io::Write, path::Path};

use challonge::{Challonge, MatchId, MatchScore, MatchScores, MatchState, MatchUpdate, Participant, ParticipantCreate, ParticipantId, Tournament, TournamentCreate, TournamentId, TournamentIncludes, tournament::RankedBy};
use chrono::{DateTime, Duration, Utc};
use chrono_tz::US::Eastern;
use itertools::Itertools;
use log::*;
use plotters::{prelude::*};
use pokerust::{Ability, FromId, FromName, Pokemon, PokemonForm, PokemonSpecies};
use rayon::prelude::*;
use serde::{Deserialize, Serialize};
use serenity::{Error, async_trait, builder::CreateEmbed, client::Context, model::{id::{ChannelId}}, utils::MessageBuilder};
use phf::phf_map;
//use tokio::{task, time::delay_for};

use crate::{DraftRegistry, EMBED_COLOR, TYPE_ABBR, TYPE_NAMES, emoji_from_type_id};
use crate::func::*;

/*pub(crate) const UNOWN_LETTERS: [&str; 28] = [
    "Alpha", "Bravo", "Charlie", "Delta", "Echo", "Foxtrot", "Golf",
    "Hotel", "India", "Juliet", "Kilo", "Lima", "Mike", "November",
    "Oscar", "Papa", "Quebec", "Romeo", "Sierra", "Tango", "Uniform",
    "Victor", "Whiskey", "X-Ray", "Yankee", "Zulu", "Exclamation", "Question"
];*/

pub trait PokemonFormExt {
    fn from_name_exhaustive(name: &str) -> Result<Self, ()> where Self: Sized;

    fn pokemon(&self) -> Pokemon;
    fn species(&self) -> PokemonSpecies;
    fn stat_hexagon(&self) -> String;
}

impl PokemonFormExt for PokemonForm {
    fn pokemon(&self) -> Pokemon {
        if let Ok(p) = self.pokemon.get() {
            p
        }
        else {
            if let Ok(p) = Pokemon::from_name(self.pokemon.name.as_str()) {
                p
            }
            else {
                panic!("PokeAPI Mon failure: {} (#{})", self.name, self.id)
            }
        }
    }

    fn species(&self) -> PokemonSpecies {
        let p = self.pokemon();
        if let Ok(s) = p.species.get() {
            s
        }
        else {
            if let Ok(s) = PokemonSpecies::from_name(p.species.name.as_str()) {
                s
            }
            else {
                panic!("PokeAPI Species failure: {} (#{})", self.name, self.id)
            }
        }
    }

    fn stat_hexagon(&self) -> String {
        match fs::create_dir("img/") {
            Ok(_) => info!("Created new img dir"),
            Err(_) => debug!("drafts dir already exists")
        }
        
        let path = format!("img/{}.png", self.name);
        let path = Path::new(path.as_str());
        match File::open(path) {
            Err(_) => {
                let p = self.pokemon();
                let stats: HashMap<String, u8> = p.stats.iter().map(|s| (s.stat.name.clone(), s.base_stat)).collect();
                let stat_keys: HashMap<String, f64> = vec![
                    ("hp", consts::TAU/4_f64),
                    ("special-attack", 5_f64*consts::FRAC_PI_6), 
                    ("special-defense", 7_f64*consts::FRAC_PI_6), 
                    ("speed", 3_f64*consts::TAU/4_f64), 
                    ("defense", 11_f64*consts::FRAC_PI_6),
                    ("attack", consts::FRAC_PI_6)].iter().map(|(k, v)| ((*k).to_owned(), *v)).collect();
                let stat_order: HashMap<String, u8> = vec![
                    ("hp", 1),
                    ("special-attack", 2), 
                    ("special-defense", 3), 
                    ("speed", 4), 
                    ("defense", 5),
                    ("attack", 6)].iter().map(|(k, v)| ((*k).to_owned(), *v)).collect();
                let stat_names: HashMap<String, String> = vec![
                    ("hp", "HP"),
                    ("special-attack", "SpA"), 
                    ("special-defense", "SpD"), 
                    ("speed", "Spe"), 
                    ("defense", "Def"),
                    ("attack", "Atk")].iter().map(|(k, v)| ((*k).to_owned(), (*v).to_owned())).collect();
                let max_stat = stats.values().max().unwrap_or(&0).to_owned();
                let lines = (max_stat as f32 / 50_f32).ceil() as u8;
                let bounds = lines as i16 * 70;
                let bounds_float = bounds as f64;
                let statgon_points = stat_keys.iter().map(|(k, v)| {
                    let stat = stats[k];
                    let i = stat_order[k];
                    let x = v.cos() * stat as f64;
                    let y = v.sin() * stat as f64;
                    (i, x, y)
                });
                let statgon_points = statgon_points.sorted_by(|(a, _, _), (b, _, _)| a.cmp(b));
                let statgon_points: Vec<(f64, f64)> = statgon_points.map(|(_, x, y)| (x, y)).collect();
                debug!("Statgon Points:");
                for (x, y) in statgon_points.clone() {
                    debug!("    ({:.2}, {:.2})", x, y);
                }
                let mut grid_lines = Vec::new();
                for i in -(lines as i16)..=-1 {
                    let h = i.abs() * 50;
                    let coords = stat_keys.iter().map(|(k, v)| {
                        let i = stat_order[k];
                        let x = v.cos() * h as f64;
                        let y = v.sin() * h as f64;
                        (i, x, y)
                    });
                    let coords = coords.sorted_by(|(a, _, _), (b, _, _)| a.cmp(b));
                    let coords: Vec<(f64, f64)> = coords.map(|(_, x, y)| (x, y)).collect();
                    // lines = 3; i = 3, 2, 1
                    // 3 % 2 = 1; rem = 0
                    // 3 % 2 = 1; color = darker
                    // 2 % 2 = 0; color = lighter...?
                    // 1 % 2 = 1; color = darker
                    // lines = 4; i = 4, 3, 2, 1
                    // 4 % 2 = 0; rem = 1
                    // 4 % 2 = 0; color = darker
                    // 3 % 2 = 1; color = lighter
                    let lines_even = lines % 2 == 0;
                    let i_even = i % 2 == 0;
                    let color = match (i_even, lines_even) {
                        //darker
                        (true, true) | (false, false) => RGBColor(37, 39, 43),
                        //lighter
                        (true, false) | (false, true) => RGBColor(47, 49, 54),
                    };
                    debug!("Grid Points:");
                    for (x, y) in coords.clone() {
                        debug!("    ({:.2}, {:.2})", x, y);
                    }
                    grid_lines.push(Polygon::new(coords, &color));
                }
                let stat_labels = stat_keys.iter().map(|(k, theta)| {
                    let r = lines as u16 * 62;
                    let mut x = theta.cos() * r as f64 - 1_f64 * lines as f64;
                    if x.floor() != 0.0 && x.floor() != -0.0 {
                        x = x - (4.0 * lines as f64);
                    }
                    let y = theta.sin() * r as f64 + 7_f64 * lines as f64;
                    Text::new(format!("{}", stat_names[k]), (x, y), ("Ubuntu", 13).into_font().color(&RGBColor(242, 255, 255)))
                });
                let stat_nums = stat_keys.iter().map(|(k, theta)| {
                    let stat = stats[k];
                    let r = lines as u16 * 62;
                    let mut x = theta.cos() * r as f64 - 1_f64 * lines as f64;
                    if x.floor() != 0.0 && x.floor() != -0.0 {
                        x = x - (4.0 * lines as f64);
                    }
                    let y = theta.sin() * r as f64;
                    Text::new(format!("{}", stat), (x, y), ("Ubuntu", 15, FontStyle::Bold).into_font().color(&RGBColor(242, 255, 255)))
                });
                //let vline = LineSeries::new((-bounds..bounds).map(|y| (0_f64, y as f64)), &RGBColor(209, 210, 211));
                //let pi6line = LineSeries::new((-bounds..bounds).map(|x| (x as f64, x as f64 / 3_f64.sqrt())), &RGBColor(209, 210, 211));
                //let fpi6line = LineSeries::new((-bounds..bounds).map(|x| (x as f64, -(x as f64 / 3_f64.sqrt()))), &RGBColor(209, 210, 211));
                let drawing_area = BitMapBackend::new(path, (250, 250))
                    .into_drawing_area();
                
                drawing_area.fill(&RGBColor(47, 49, 54)).unwrap();

                let mut chart = ChartBuilder::on(&drawing_area)
                    .build_cartesian_2d(-bounds_float..bounds_float, -bounds_float..bounds_float)
                    .unwrap();
                chart.draw_series(grid_lines).unwrap();
                //chart.draw_series(vec![vline]).unwrap();
                //chart.draw_series(vec![pi6line]).unwrap();
                //chart.draw_series(vec![fpi6line]).unwrap();
                chart.draw_series(vec![Polygon::new(statgon_points,
                    &RGBColor(156, 90, 197).mix(0.66))]).unwrap();
                chart.draw_series(stat_labels).unwrap();
                chart.draw_series(stat_nums).unwrap();
            },
            Ok(_) => (),
        }
        format!("https://xayah.net/static/genesect/{}.png", self.name)
    }

    fn from_name_exhaustive(name: &str) -> Result<Self, ()> {
        let mut o = name.to_owned();
        if !o.contains("unown") {
            if o.ends_with("-g") {
                o.push_str("alar");
            }
            if o.ends_with("-a") {
                o.push_str("lola");
            }
            if o.ends_with("-p") {
                o.push_str("aldea");
            }
            if o.as_str() == "darmanitan-galar" {
                o = "darmanitan-standard-galar".to_string();
            }
        }
        let name = if o.as_str() == "gastrodon" { "gastrodon-west" } 
        else if ["tauros-paldea", "tauros-combat", "tauros-paldea-combat", "tauros-combat-paldea", "tauros-combat-breed"].contains(&o.as_str()) { "tauros-paldea-combat-breed" }
        else if ["tauros-blaze", "tauros-paldea-blaze", "tauros-blaze-paldea", "tauros-blaze-breed"].contains(&o.as_str()) { "tauros-paldea-blaze-breed" }
        else if ["tauros-aqua", "tauros-paldea-aqua", "tauros-aqua-paldea", "tauros-aqua-breed"].contains(&o.as_str()) { "tauros-paldea-aqua-breed" }
        else if ["ogerpon-c", "ogerpon-cornerstone"].contains(&o.as_str()) { "ogerpon-cornerstone-mask" }
        else if ["ogerpon-h", "ogerpon-hearthflame"].contains(&o.as_str()) { "ogerpon-hearthflame-mask" }
        else if ["ogerpon-w", "ogerpon-wellspring"].contains(&o.as_str()) { "ogerpon-wellspring-mask" }
        else { o.as_str() };


        if let Ok(f) = Self::from_name(name) {
            debug!("Form exhaustive succeeded first try");
            Ok(f)
        }
        else {
            if let Ok(p) = Pokemon::from_name(name) {
                if let Ok(f) = p.forms[0].get() {
                    warn!("Form exhaustive failed first try, succeeded second try. input {} form name {}", name, f.name);
                    Ok(f)
                }
                else {
                    if let Ok(f) = Self::from_name(&p.forms[0].name) {
                        warn!("Form exhaustive failed twice, succeeded 3rd time. input {} form name {}", name, f.name);
                        Ok(f)
                    }
                    else {
                        error!("Form exhaustive failed, but pokemon succeeded. input {} mon name {}", name, p.name);
                        Err(())
                    }
                }
            }
            else {
                if let Ok(s) = PokemonSpecies::from_name(name) {
                    if let Some(default) = s.varieties.iter().filter(|p| p.is_default).next() {
                        if let Ok(p) = default.pokemon.get() {
                            if let Ok(f) = p.forms[0].get() {
                                warn!("Form exhaustive failed twice, succeeded with species. input {} form name {}", name, f.name);
                                Ok(f)
                            }
                            else {
                                if let Ok(f) = Self::from_name(&p.forms[0].name) {
                                    warn!("Form exhaustive failed thrice, succeeded with species. input {} form name {}", name, f.name);
                                    Ok(f)
                                }
                                else {
                                    error!("Form exhaustive failed, but pokemon/species succeeded. input {} species {} mon {}", name, s.name, p.name);
                                    Err(())
                                }
                            }
                        }
                        else {
                            if let Ok(p) = Pokemon::from_name(&default.pokemon.name) {
                                if let Ok(f) = p.forms[0].get() {
                                    warn!("Form exhaustive failed thrice, succeeded with species. input {} form name {}", name, f.name);
                                    Ok(f)
                                }
                                else {
                                    if let Ok(f) = Self::from_name(&p.forms[0].name) {
                                        warn!("Form exhaustive failed four times, succeeded with species. input {} form name {}", name, f.name);
                                        Ok(f)
                                    }
                                    else {
                                        error!("Form exhaustive failed, but pokemon/species succeeded. input {} species {} mon {}", name, s.name, p.name);
                                        Err(())
                                    }
                                }
                            }
                            else {
                                error!("Form exhaustive failed, species succeeded but pokemon failed. input {} species {}", name, s.name);
                                Err(())
                            }
                        }
                    }
                    else {
                        error!("Form exhaustive failed, no default variety of species. input {} species {}", name, s.name);
                        Err(())
                    }
                }
                else {
                    error!("Form exhaustive failed, species failed. input {}", name);
                    Err(())
                }
            }
        }
    }
}

#[derive(PartialEq, Eq, Clone, Copy)]
pub enum SortType {
    Default,
    NumID,
    StrID,
    Nickname,
    SpeciesName,
    Tier,
    Date,
    Type,
    CurrentRecord,
    LifetimeRecord,
    BST,
    PlayoffIdentifier,
    Random,
}

#[async_trait]
pub trait Listable {
    fn type_name() -> String;
    fn type_name_plural() -> String {
        Self::type_name()
    }
    fn page_size() -> u8 {
        20
    }

    fn valid_sorts() -> Vec<SortType>;
    fn default_sort() -> SortType;

    async fn field_name(&self, ctx: &Context) -> String;
    async fn field_value(&self, ctx: &Context) -> String;

    fn sort_str(&self, mode: SortType) -> String;
    fn sort_num(&self, mode: SortType) -> isize;
    fn sort_bool(&self, mode: SortType) -> bool;
    fn sort_date(&self) -> DateTime<Utc>;
}

pub trait Awardable {
    fn awards<'draft>(&'draft self) -> &'draft HashMap<String, String>;
    fn awards_mut<'draft>(&'draft mut self) -> &'draft mut HashMap<String, String>;
    fn grant_award(&mut self, award: String) {
        let awards = self.awards_mut();
        let id = award.to_lowercase();
        let id = id.replace(" ", "");
        awards.insert(id, award);
    }
    fn revoke_award(&mut self, award: String) {
        let awards = self.awards_mut();
        let id = award.to_lowercase();
        let id = id.replace(" ", "");
        awards.remove(&id);
    }
    fn has_award(&self, award: String) -> bool {
        let id = award.to_lowercase();
        let id = id.replace(" ", "");
        self.awards().contains_key(&id)
    }
}

pub enum AwardableKey {
    Coach(u64),
    Mon(i16),
}

#[derive(Serialize, Deserialize, Debug, Clone, Copy, PartialEq)]
pub enum Format {
    Singles,
    Doubles,
}

#[derive(Serialize, Deserialize, Debug, Clone, Copy, PartialEq)]
pub enum TierSource {
    Smogon,
    PPA,
    WBE,
    Custom,
}

#[derive(Serialize, Deserialize, Debug, Clone, Copy, PartialEq)]
pub enum DraftStatus {
    Init,
    Waiting,
    Ready,
    Drafting,
    Playing,
    SeasonPlayoffs,
    TournComplete,
    SeasonComplete,
    RepairPlayoffs,
}

impl Display for DraftStatus {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let s = match self {
            DraftStatus::Init => "Not Ready",
            DraftStatus::Waiting => "Waiting for Coaches",
            DraftStatus::Ready => "Ready to Start Drafting",
            DraftStatus::Drafting => "Drafting",
            DraftStatus::Playing => "Playing Matches",
            DraftStatus::TournComplete => "Completed",
            DraftStatus::SeasonComplete => "Off-Season",
            DraftStatus::SeasonPlayoffs => "Playoffs",
            DraftStatus::RepairPlayoffs => "Repairing Playoff Matches",
        };
        write!(f, "{}", s)
    }
}

#[derive(Serialize, Deserialize, Debug, Clone, Copy, PartialEq)]
pub enum DraftType {
    Linear,
    Snake,
}

#[derive(Serialize, Deserialize, Debug, Clone, Copy, PartialEq)]
pub enum TradePolicy {
    Disabled,
    Bank,
    Player,
    Free,
}

impl Display for TradePolicy {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let s = match self {
            TradePolicy::Disabled => "Disabled",
            TradePolicy::Bank => "Bank",
            TradePolicy::Player => "P2P",
            TradePolicy::Free => "Bank or P2P",
        };
        write!(f, "{}", s)
    }
}

#[derive(Debug, Clone, Copy, PartialEq)]
pub enum TradePerformance {
    Fix,
    CoachBank,
    CoachPlayer,
    ModBank,
    //ModPlayer,
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Match {
    pub uuid: u64,
    pub id: u16,
    pub coach1: u64,
    pub coach2: u64,
    pub team1: Vec<MatchPokemon>,
    pub team2: Vec<MatchPokemon>,
    pub winner_is_1: bool,
    pub c1_score: u8,
    pub c2_score: u8,
    pub deaths1: u8,
    pub deaths2: u8,
    pub replay: String,
    pub date: DateTime<Utc>,
    #[serde(default)]
    pub playoffs: bool,
}

#[async_trait]
impl Listable for Match {
    fn type_name() -> String {
        "Match".to_owned()
    }

    fn type_name_plural() -> String {
        "Matches".to_owned()
    }

    fn page_size() -> u8 {
        10
    }

    async fn field_name(&self, ctx: &Context) -> String {
        let draft_id = self.get_draft_id(ctx).await.unwrap();
        if self.winner_is_1 {
            let mut s = MessageBuilder::new();
                s.push(format!("#{}", self.id));
                if self.playoffs {
                    s.push(" (Playoff Game)");
                }
                s.push(": ");
                s.push_bold(ctx.get_coach_name(&draft_id, &self.coach1, None).await);
                s.push(" vs. ");
                s.push(ctx.get_coach_name(&draft_id, &self.coach2, None).await);
                s.push(" (");
                s.push_safe(self.date.with_timezone(&Eastern).format("%d/%m/%Y at %r %Z"));
                s.push(")");
            s.build()
        }
        else {
            let mut s = MessageBuilder::new();
                s.push(format!("#{}", self.id));
                if self.playoffs {
                    s.push(" (Playoff Game)");
                }
                s.push(": ");
                s.push(ctx.get_coach_name(&draft_id, &self.coach1, None).await);
                s.push(" vs. ");
                s.push_bold(ctx.get_coach_name(&draft_id, &self.coach2, None).await);
                s.push(" (");
                s.push_safe(self.date.with_timezone(&Eastern).format("%d/%m/%Y at %r %Z"))
                .push(")");
            s.build()
        }
    }

    async fn field_value(&self, _: &Context) -> String {
        let s = MessageBuilder::new()
            .push_bold("Result: ")
            .push_spoiler_line(format!("{} - {} ({} Deaths - {} Deaths)", self.c1_score, self.c2_score, self.deaths1, self.deaths2))
            .push_bold("Replay: ")
            .push_safe(self.replay.clone())
            .build();
        s
    }

    fn valid_sorts() -> Vec<SortType> {
        vec![SortType::Date, SortType::CurrentRecord]
    }

    fn default_sort() -> SortType {
        SortType::Date
    }

    fn sort_str(&self, _mode: SortType) -> String {
        todo!()
    }

    fn sort_num(&self, _mode: SortType) -> isize {
        self.absolute_differential() as isize
    }

    fn sort_bool(&self, _mode: SortType) -> bool {
        todo!()
    }

    fn sort_date(&self) -> DateTime<Utc> {
        self.date
    }
}

impl Match {
    pub fn absolute_differential(&self) -> u8 {
        if self.winner_is_1 {
            self.c1_score - self.c2_score
        }
        else {
            self.c2_score - self.c1_score
        }
    }

    pub fn get_coach1<'a>(&self, draft: &'a Draft) -> &'a Coach {
        let c = &draft.coaches[&self.coach1];
        c
    }

    pub fn get_coach2<'a>(&self, draft: &'a Draft) -> &'a Coach {
        let c = &draft.coaches[&self.coach2];
        c
    }

    pub fn get_team1<'a>(&self, draft: &'a Draft) -> HashMap<i16, &'a DraftPokemon> {
        let t1: HashMap<i16, &'a DraftPokemon> = self.team1.iter().map(|i| (i.id, &draft.mons[&i.id])).collect();
        t1
    }

    pub fn get_team2<'a>(&self, draft: &'a Draft) -> HashMap<i16, &'a DraftPokemon> {
        let t2: HashMap<i16, &'a DraftPokemon> = self.team2.iter().map(|i| (i.id, &draft.mons[&i.id])).collect();
        t2
    }

    pub async fn get_draft_id(&self, ctx: &Context) -> Option<String> {
        let data = ctx.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("Bot didn't start right");
        let f = reg.par_iter().find_first(|(_, d)| 
            d.matches.par_iter().filter(|m| m.uuid == self.uuid).count() == 1
        )
        .map(|(_, d)| d.draft_id.clone());
        f
    }

    pub async fn embed(&self, ctx: &Context, draft: &Draft, server_id: Option<u64>) -> CreateEmbed {
        let t1 = self.get_team1(draft);
        let t2 = self.get_team2(draft);
        let c1 = self.get_coach1(draft);
        let c2 = self.get_coach2(draft);

        let t1_results = {
            let mut s = String::new();
            s.push_str("||");
            for m in self.team1.clone() {
                let p = t1[&m.id];
                s.push_str(format!("{} had {} active kills, {} passive kills, and {} deaths.\n", p.mon_name(), m.active_kills, m.passive_kills, m.deaths).as_str());
            }
            s.push_str("||");
            s
        };
        let t2_results = {
            let mut s = String::new();
            s.push_str("||");
            for m in self.team2.clone() {
                let p = t2[&m.id];
                s.push_str(format!("{} had {} active kills, {} passive kills, and {} deaths.\n", p.mon_name(), m.active_kills, m.passive_kills, m.deaths).as_str());
            }
            s.push_str("||");
            s
        };

        let e = {
            let mut e = CreateEmbed::default();
            let c1_dn = c1.display_name(ctx, server_id).await;
            let c2_dn = c2.display_name(ctx, server_id).await;

            let desc = if self.playoffs {
                format!("*{}* vs. *{}* (Playoff Game)", &c1_dn, &c2_dn)
            }
            else {
                format!("*{}* vs. *{}*", &c1_dn, &c2_dn)
            };

            if self.playoffs {
                e.title(format!("{} Playoff Game (Match #{}) Results", draft.draft_name, self.id));
            }
            else {
                e.title(format!("{} Match #{} Results", draft.draft_name, self.id));
            }

            e.description(desc);
            
            e.field(&c1_dn, t1_results, false);
            e.field(&c2_dn, t2_results, false);

            if self.winner_is_1 {
                e.field("Result", format!("||**{}** won {}-{}||", &c1_dn, self.c1_score, self.c2_score), false);
            }
            else {
                e.field("Result", format!("||**{}** won {}-{}||", &c2_dn, self.c2_score, self.c1_score), false);
            }

            e.field("Replay", &self.replay, false);

            e.footer(|f| {
                if self.playoffs {
                    f.text(format!("Playoff game played on {}", self.date.format("%e %B %Y at %R %Z")));
                }
                else {
                    f.text(format!("Match played on {}", self.date.format("%e %B %Y at %R %Z")));
                }

                f
            });

            e.color(EMBED_COLOR);

            e
        };

        e
    }
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct PlayoffMatch {
    pub challonge_id: u64,
    pub letter: String,
    pub matches: Vec<Match>,
    pub coach1: u64,
    pub coach2: u64,
    pub score1: u8,
    pub score2: u8,
    pub complete: bool,
    pub winner_is_1: Option<bool>,
}

impl PlayoffMatch {
    pub fn new_from_match(challonge_id: u64, letter: String, the_match: Match) -> PlayoffMatch {
        let mut matches = Vec::new();
        let coach1 = the_match.coach1;
        let coach2 = the_match.coach2;
        let (score1, score2) = if the_match.winner_is_1 { (1, 0) } else { (0, 1) };
        let mut pushmatch = the_match.clone();
        pushmatch.playoffs = true;
        matches.push(pushmatch);

        PlayoffMatch {
            challonge_id,
            letter,
            matches,
            coach1,
            coach2,
            score1,
            score2,
            complete: false,
            winner_is_1: None,
        }
    }

    pub fn add_match(&mut self, the_match: Match) -> Result<(), (String, PlayoffMatch)> {
        let backup = self.clone();
        
        if self.coach1 == the_match.coach1 && self.coach2 == the_match.coach2 {
            if the_match.winner_is_1 {
                self.score1 += 1;
            }
            else {
                self.score2 += 1;
            }
        }
        else if self.coach1 == the_match.coach2 && self.coach2 == the_match.coach1 {
            if the_match.winner_is_1 {
                self.score2 += 1;
            }
            else {
                self.score1 += 1;
            }
        }
        else {
            return Err(("The coaches detected in this game don't match the ones in the playoff matchup.".into(), backup));
        }
        let mut pushmatch = the_match.clone();
        pushmatch.playoffs = true;
        self.matches.push(pushmatch);

        Ok(())
    }

    pub fn finalize_match(&mut self) -> Result<(), (String, PlayoffMatch)> {
        if self.score1 != self.score2 {
            self.winner_is_1 = Some(self.score1 > self.score2);
            self.complete = true;
        }

        Ok(())
    }

    pub fn _letter_identifier(&self, id: &TournamentId) -> Result<String, String> {
        let c = match (env::var("CHALLONGE_USER"), env::var("CHALLONGE_TOKEN")) {
            (Ok(u), Ok(k)) => {
                Challonge::new(u, k)
            },
            _ => {
                return Err("Missing Challonge API keys in environment".into());
            }
        };
        let m = c.get_match(id, &MatchId(self.challonge_id), false).unwrap();
        Ok(m.identifier.clone())
    }

    pub fn get_coach1<'a>(&self, draft: &'a Draft) -> &'a Coach {
        let c = &draft.coaches[&self.coach1];
        c
    }

    pub fn get_coach2<'a>(&self, draft: &'a Draft) -> &'a Coach {
        let c = &draft.coaches[&self.coach2];
        c
    }

    pub async fn embed(&self, ctx: &Context, draft: &Draft, server_id: Option<u64>) -> CreateEmbed {
        let c1 = self.get_coach1(draft);
        let c2 = self.get_coach2(draft);
        let l = self.letter.clone();

        let e = {
            let mut e = CreateEmbed::default();
            let c1_dn = c1.display_name(ctx, server_id).await;
            let c2_dn = c2.display_name(ctx, server_id).await;

            e.title(format!("{} Playoff Match {} Results", draft.draft_name, l));
            e.description(format!("*{}* vs. *{}* (Best of {})", &c1_dn, &c2_dn, draft.playoff_best_of * 2 + 1));

            let game_results = {
                let mut s = String::new();
                let mut g = 1;
                for m in self.matches.iter().sorted_by(|a, b| a.date.cmp(&b.date)) {
                    let mc1 = m.get_coach1(draft);
                    let mc2 = m.get_coach2(draft);
                    let mc1_dn = mc1.display_name(ctx, server_id).await;
                    let mc2_dn = mc2.display_name(ctx, server_id).await;
                    s.push_str(format!("Game {} (Match #{}): ||{} {}-{} {}||\n", g, m.id, mc1_dn, m.c1_score, m.c2_score, mc2_dn).as_str());
                    g += 1;
                }
                s
            };
            e.field("Game Results", game_results, false);

            if self.winner_is_1.is_some() {
                let w1 = self.winner_is_1.as_ref().unwrap();
                if *w1 {
                    e.field("Winner", format!("**{}** {}-{}", c1_dn, self.score1, self.score2), false);
                }
                else {
                    e.field("Winner", format!("**{}** {}-{}", c2_dn, self.score2, self.score1), false);
                }
            }
            else {
                e.footer(|f| f.text("Playoff match not final, has not reached the best-of score threshold"));
            }

            e.color(EMBED_COLOR);

            e
        };

        e
    }

    pub async fn get_draft_id(&self, ctx: &Context) -> Option<String> {
        let data = ctx.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("Bot didn't start right");
        let f = reg.par_iter().find_first(|(_, d)| 
            d.playoff_matches.par_iter().filter(|m| m.challonge_id == self.challonge_id).count() == 1
        )
        .map(|(_, d)| d.draft_id.clone());
        f
    }
}

#[async_trait]
impl Listable for PlayoffMatch {
    fn type_name() -> String {
        "Playoff Match".into()
    }
    fn type_name_plural() -> String {
        "Playoff Matches".into()
    }

    fn valid_sorts() -> Vec<SortType> {
        vec![SortType::PlayoffIdentifier]
    }

    fn default_sort() -> SortType {
        SortType::PlayoffIdentifier
    }

    async fn field_name(&self, ctx: &Context) -> String {
        let draft_id = self.get_draft_id(ctx).await.unwrap();
        let mut s = MessageBuilder::new();
            s.push(format!("Match {}: ", self.letter));
            s.push(ctx.get_coach_name(&draft_id, &self.coach1, None).await);
            s.push(" vs. ");
            s.push(ctx.get_coach_name(&draft_id, &self.coach2, None).await);
        s.build()
    }

    async fn field_value(&self, ctx: &Context) -> String {
        let draft_id = self.get_draft_id(ctx).await.unwrap();
        if self.winner_is_1.is_some() {
            let w1 = self.winner_is_1.as_ref().unwrap();
            if *w1 {
                let c1_dn = ctx.get_coach_name(&draft_id, &self.coach1, None).await;
                format!("Winner: **{}** {}-{}", c1_dn, self.score1, self.score2)
            }
            else {
                let c2_dn = ctx.get_coach_name(&draft_id, &self.coach2, None).await;
                format!("Winner: **{}** {}-{}", c2_dn, self.score2, self.score1)
            }
        }
        else {
            "Ongoing".into()
        }
    }

    fn sort_str(&self, _: SortType) -> String {
        self.letter.clone()
    }

    fn sort_num(&self, _: SortType) -> isize {
        todo!()
    }

    fn sort_bool(&self, _: SortType) -> bool {
        todo!()
    }

    fn sort_date(&self) -> DateTime<Utc> {
        todo!()
    }
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct MatchPokemon {
    pub id: i16,
    pub active_kills: u8,
    pub passive_kills: u8,
    pub deaths: u8,
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct DraftPokemon {
    pub id: i16,
    pub name_id: String,
    pub spid: i16,
    pub species_name: String,
    pub form_name: Option<String>,
    pub bst: u16,
    pub type1: u8,
    pub type2: Option<u8>,
    pub tier: String,
    pub drafted: bool,
    pub coach: Option<u64>,
    pub returned: bool,
    pub active_kills: u16,
    pub passive_kills: u16,
    pub deaths: u16,
    pub lifetime_active: u16,
    pub lifetime_passive: u16,
    pub lifetime_deaths: u16,
    #[serde(default)]
    pub awards: HashMap<String, String>,
    pub banned_ability: Option<u8>,
    #[serde(default)]
    pub mega_id: Option<i16>,
}

impl DraftPokemon {
    pub fn new_from_name(mon: String, tier: String, banned_ability: Option<u8>) -> Result<DraftPokemon, minreq::Error> {
        let form = PokemonForm::from_name(mon.as_str())?;
        debug!("Draft Mon Creation: Form succeeded");
        DraftPokemon::new(form, tier, banned_ability)
    }

    pub fn new_from_id(id: i16, tier: String, banned_ability: Option<u8>) -> Result<DraftPokemon, minreq::Error> {
        let form = PokemonForm::from_id(id)?;
        debug!("Draft Mon Creation: Form succeeded");
        DraftPokemon::new(form, tier, banned_ability)
    }

    pub fn new(form: PokemonForm, tier: String, banned_ability: Option<u8>) -> Result<DraftPokemon, minreq::Error> {
        let id = form.id;
        let pokemon = form.pokemon();
        debug!("Draft Mon Creation: Pokemon succeeded");
        let species = form.species();
        debug!("Draft Mon Creation: Species succeeded");
        let species_name = species.names.iter().filter(|x| x.language.name == "en").next().map_or(&species.name, |v| &v.name).clone();
        let form_name = match &form.form_names.iter().filter(|x| x.language.name == "en").next() {
            Some(n) => Some(n.name.clone()),
            None => None,
        };
        let bhp = pokemon.stats.iter().filter(|x| x.stat.name == "hp").next().map_or(0, |v| v.base_stat);
        let batk = pokemon.stats.iter().filter(|x| x.stat.name == "attack").next().map_or(0, |v| v.base_stat);
        let bdef = pokemon.stats.iter().filter(|x| x.stat.name == "defense").next().map_or(0, |v| v.base_stat);
        let bspa = pokemon.stats.iter().filter(|x| x.stat.name == "special-attack").next().map_or(0, |v| v.base_stat);
        let bspd = pokemon.stats.iter().filter(|x| x.stat.name == "special-defense").next().map_or(0, |v| v.base_stat);
        let bspe = pokemon.stats.iter().filter(|x| x.stat.name == "speed").next().map_or(0, |v| v.base_stat);
        let bst: u16 = bhp as u16 + batk as u16 + bdef as u16 + bspa as u16 + bspd as u16 + bspe as u16;
        let t1 = if pokemon.types[0].type_.name.as_str() == "electric" {
            13
        }
        else {
            pokemon.types[0].type_.get()?.id as u8
        };
        let t2 = {
            if pokemon.types.len() > 1 {
                let i = pokemon.types[1].type_.get()?.id as u8;
                Some(i)
            }
            else { None }
        };
        debug!("Draft Mon Creation: Types succeeded");

        Ok(DraftPokemon {
            id,
            name_id: form.name,
            spid: species.id,
            species_name,
            form_name,
            bst,
            type1: t1,
            type2: t2,
            tier,
            drafted: false,
            coach: None,
            returned: false,
            active_kills: 0,
            passive_kills: 0,
            deaths: 0,
            lifetime_active: 0,
            lifetime_passive: 0,
            lifetime_deaths: 0,
            awards: HashMap::new(),
            banned_ability,
            mega_id: None,
        })
    }

    pub fn correct(&self) -> Option<(i16, DraftPokemon)> {
        let form = self.form();
        let species = form.as_ref().map(|f| f.species());
        if species.is_none() || species.as_ref().unwrap().id != self.spid {
            let name_id = match self.name_id.as_str() {
                "darmanitan-standard-galar" => "darmanitan-galar-standard",
                "darmanitan-zen-galar" => "darmanitan-galar-zen",
                _ => self.name_id.as_str()
            };
            if let Ok(correct_form) = PokemonForm::from_name(&name_id) {
                info!("Correcting Draft Mon form id (#{} -> #{})", self.id, correct_form.id);
                let mut fix = self.clone();
                fix.id = correct_form.id;
                Some((fix.id, fix))
            }
            else {
                panic!("PokeAPI failure, name_id {}", &self.name_id);
            }
        }
        else {
            debug!("Draft Mon (#{}) doesn't require changes", self.id);
            None
        }
    }

    pub fn reset(&mut self, ownership: bool, returned: bool) {
        self.active_kills = 0;
        self.passive_kills = 0;
        self.deaths = 0;
        if ownership {
            self.drafted = false;
            self.coach = None;
            self.returned = returned;
        }
    }

    pub fn get_coach<'a>(&self, draft: &'a Draft) -> Option<&'a Coach> {
        match self.coach {
            Some(c) => {
                let c = &draft.coaches[&c];
                Some(c)
            },
            None => None,
        }
    }
    
    pub fn mon_name(&self) -> String {
        match self.form_name.clone() {
            Some(f) => format!("{} ({})", self.species_name, f),
            None => self.species_name.clone()
        }
    }

    pub fn name_with_text_type(&self) -> String {
        if self.banned_ability.is_some() {
            format!("{} (Abilities Restricted) ({})", self.mon_name(), self.short_type_string())
        }
        else {
            format!("{} ({})", self.mon_name(), self.short_type_string())
        }
    }

    pub fn typemojis(&self) -> String {
        let mut s = String::new();
        let e1 = emoji_from_type_id(self.type1);
        s.push_str(e1);
        if self.type2.is_some() {
            let e2 = emoji_from_type_id(self.type2.unwrap());
            s.push_str(e2);
        }
        s
    }

    pub fn type_string(&self) -> String {
        let mut s = String::new();
        let e1 = emoji_from_type_id(self.type1);
        s.push_str(e1);
        s.push_str(" ");
        s.push_str(TYPE_NAMES[self.type1 as usize]);
        if self.type2.is_some() {
            let t2 = self.type2.unwrap();
            let e2 = emoji_from_type_id(t2);
            s.push_str(" / ");
            s.push_str(e2);
            s.push_str(" ");
            s.push_str(TYPE_NAMES[t2 as usize]);
        }
        s
    }

    pub fn short_type_string(&self) -> String {
        let mut s = String::new();
        s.push_str(TYPE_ABBR[self.type1 as usize]);
        if self.type2.is_some() {
            let t2 = self.type2.unwrap();
            s.push_str("/");
            s.push_str(TYPE_ABBR[t2 as usize]);
        }
        s
    }

    pub fn tier_string(&self) -> String {
        if let Ok(i) = self.tier.parse::<usize>() {
            format!("Tier {}", i)
        }
        else {
            format!("{} Tier", self.tier)
        }
    }

    pub fn sprite(&self) -> Result<String, minreq::Error> {
        let id = if self.mega_id.is_some() {
            self.mega_id.unwrap()
        }
        else {
            self.id
        };
        let f = PokemonForm::from_id(id).expect("form error");
        let p = f.pokemon.get().expect("pokemon error");
        let s = p.species.get().expect("species error");
        let shiny = "normal";
        let sprite = if s.id != p.id {
            format!("https://xayah.net/static/pokemon/gifs/{}/{}.gif", shiny, p.name)
        }
        else {
            if s.id == 201 {
                format!("https://xayah.net/static/pokemon/gifs/{}/unown-alpha.gif", shiny)
            }
            else if [664, 665].contains(&s.id) {
                format!("https://xayah.net/static/pokemon/gifs/{}/{}.gif", shiny, s.name)
            }
            else if s.id != f.id {
                format!("https://xayah.net/static/pokemon/gifs/{}/{}.gif", shiny, f.name)
            }
            else {
                format!("https://xayah.net/static/pokemon/gifs/{}/{}.gif", shiny, s.name)
            }
        };
        debug!("Draft Mon: Sprite url {}", sprite);
        Ok(sprite)
    }

    pub async fn embed(&self, ctx: &Context, draft: &Draft, server_id: Option<u64>) -> Result<CreateEmbed, Error> {
        let id = if self.mega_id.is_some() {
            self.mega_id.unwrap()
        }
        else {
            self.id
        };
        let f = PokemonForm::from_id(id).expect("form error");
        let p = f.pokemon.get().expect("pokemon error");

        let coach_name = self.get_coach(draft).map(|c| {
            c.display_name(ctx, server_id)
        });

        let sprite = self.sprite().expect("sprite error");

        let abilities = {
            let m: Vec<(u8, Ability)> = p.abilities.iter().map(|a| {
                (a.slot, a.ability.get())
            })
            .filter(|(_, a)| {
                a.is_ok()
            })
            .map(|(s, a)| {
                (s, a.unwrap())
            }).collect();
            let mut split: Vec<(u8, String, String)> = m.iter().map(|(s, a)| {
                let n = a.names.iter().filter(|a| a.language.name == "en").next();
                let f = a.flavor_text_entries.iter().filter(|a| a.language.name == "en").next();
                (
                    s.to_owned(),
                    if n.is_some() { n.unwrap().name.clone() } else { a.name.clone() },
                    if f.is_some() { f.unwrap().flavor_text.clone() } else { "Flavor text missing".to_owned() }
                )
            }).collect();
            split.sort_by(|(a, _, _), (b, _, _)| a.cmp(b));
            split
        };

        let statgon = self.get_stat_hexagon();

        let awards = {
            let mut s = String::new();
            for (_, a) in &self.awards {
                s.push_str(a.as_str());
                s.push('\n');
            }
            s
        };

        let e = {
            let mut e = CreateEmbed::default();

            e.title(format!("{} (click for Smogon)", self.mon_name()));
            e.description(self.field_value(ctx).await);
            e.url(format!("https://www.smogon.com/dex/ss/pokemon/{}/", f.name));
            e.thumbnail(sprite);
            e.image(statgon);
            
            e.field("Type", self.type_string(), false);
            
            for (s, k, v) in abilities {
                if self.banned_ability.is_none() || s != self.banned_ability.clone().unwrap() {
                    match s {
                        1 => {
                            e.field(format!("Ability 1: {}", k), v, false);
                        },
                        2 => {
                            e.field(format!("Ability 2: {}", k), v, false);
                        },
                        3 => {
                            e.field(format!("Hidden Ability: {}", k), v, false);
                        },
                        _ => ()
                    }
                }
                else if self.banned_ability.is_some() && s == self.banned_ability.clone().unwrap() {
                    e.field(format!("BANNED Ability: {}", k), format!("~~{}~~", v), false);
                }
            }
            if coach_name.is_some() {
                e.field("Drafted by", coach_name.unwrap().await, false);
            }
            if awards.len() > 0 {
                e.field("Awards", awards, false);
            }

            e.footer(|f| {
                //f.text(format!("{} | Caught on {}", ownership_footer, indiv.caught.format("%e %B %Y at %R %Z")));

                f
            });

            e.color(EMBED_COLOR);

            e
        };

        Ok(e)
    }

    pub fn form(&self) -> Option<PokemonForm> {
        let id = if self.mega_id.is_some() {
            self.mega_id.unwrap()
        }
        else {
            self.id
        };
        if let Ok(f) = PokemonForm::from_id(id) {
            Some(f)
        }
        else {
            None
        }
    }

    pub fn get_stat_hexagon(&self) -> String {
        self.form().unwrap().stat_hexagon()
    }

    pub fn change_tier(&mut self, new_tier: String, banned_ability: Option<u8>) -> String {
        let old_tier = self.tier.clone();
        self.tier = new_tier;
        self.banned_ability = banned_ability;
        old_tier
    }
}

#[async_trait]
impl Listable for DraftPokemon {
    fn type_name() -> String {
        "Pokémon".to_owned()
    }

    async fn field_name(&self, _: &Context) -> String {
        if self.banned_ability.is_some() {
            format!("{} (Abilities Restricted) {}", self.mon_name(), self.typemojis())
        }
        else {
            format!("{} {}", self.mon_name(), self.typemojis())
        }
    }

    async fn field_value(&self, _: &Context) -> String {
        let a = self.lifetime_active > self.active_kills;
        let p = self.lifetime_passive > self.passive_kills;
        let d = self.lifetime_deaths > self.deaths;
        // true only if mon got used at all in previous season matches
        let y = a || p || d;
        // true only if mon has never been used at all in a match
        let z = self.lifetime_active == 0 && self.lifetime_passive == 0 && self.lifetime_deaths == 0;
        // so, y and z can never both be true.
        // IF both are FALSE, that means it has been used at all for the first time in THIS season.
        // SO, we can print only the current season values.
        // Furthermore, y and z AND self.drafted cannot all be false,
        // because if y && z is false it has current stats,
        // and must already be drafted. When a mon is returned through bank trades,
        // self.drafted is false, and current stats are cleared.
        // P2p traded mons also have y = true (if they were used by the first coach)
        match (y, z, self.drafted) {
            // mon has not been used in previous seasons,
            // but now it has stats (can be 0), and is currently drafted
            (false, false, true) | (false, true, true) => {
                format!("{} | {} BST | {}/{} Kills (A/P) | {} Deaths",
                    self.tier_string(), self.bst,
                    self.active_kills, self.passive_kills, self.deaths)
            },
            // mon has been used in previous seasons,
            // and is currently drafted. May or may not have current stats.
            (true, false, true) => {
                format!("{} | {} BST | {}/{} Kills (A/P) | {} Deaths\n{}/{} Lifetime Kills (A/P) | {} Lifetime Deaths",
                    self.tier_string(), self.bst,
                    self.active_kills, self.passive_kills, self.deaths,
                    self.lifetime_active, self.lifetime_passive, self.lifetime_deaths)
            },
            // mon has been used in previous seasons and is not currently drafted.
            (true, false, false) => {
                format!("{} | {} BST | {}/{} Lifetime Kills (A/P) | {} Lifetime Deaths",
                    self.tier_string(), self.bst,
                    self.lifetime_active, self.lifetime_passive, self.lifetime_deaths)
            },
            // mon is completely untouched.
            (false, true, false) => {
                format!("{} | {} BST", self.tier_string(), self.bst)
            },
            _ => {
                // we shouldn't get here, but there's no point in panicking, LOL
                format!("{} | {} BST", self.tier_string(), self.bst)
            }
        }
    }

    fn valid_sorts() -> Vec<SortType> {
        vec![SortType::NumID,
            SortType::SpeciesName, 
            SortType::Type, 
            SortType::CurrentRecord, 
            SortType::LifetimeRecord, 
            SortType::BST,
            SortType::Tier,]
    }

    fn default_sort() -> SortType {
        SortType::Type
    }

    fn sort_str(&self, mode: SortType) -> String {
        match mode {
            SortType::SpeciesName => self.species_name.clone(),
            SortType::Tier => self.tier.clone(),
            _ => panic!("We shouldn't get here")
        }
    }

    fn sort_num(&self, mode: SortType) -> isize {
        match mode {
            SortType::NumID => self.spid as isize,
            SortType::Type => self.type1 as isize,
            SortType::CurrentRecord => {
                (self.active_kills + self.passive_kills - self.deaths) as isize
            },
            SortType::LifetimeRecord => {
                (self.lifetime_active + self.lifetime_passive - self.lifetime_deaths) as isize
            },
            SortType::BST => self.bst as isize,
            _ => panic!("we shouldn't get here")
        }
    }

    fn sort_bool(&self, _mode: SortType) -> bool {
        panic!("we shouldn't get here")
    }

    fn sort_date(&self) -> DateTime<Utc> {
        panic!("we shouldn't get here")
    }
}

impl Awardable for DraftPokemon {
    fn awards<'draft>(&'draft self) -> &'draft HashMap<String, String> {
        &self.awards
    }
    fn awards_mut<'draft>(&'draft mut self) -> &'draft mut HashMap<String, String> {
        &mut self.awards
    }
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Coach {
    pub discord_id: u64,
    pub discord_name: String,
    pub showdown_name: Option<String>,
    pub name: Option<String>,
    pub logo: Option<String>,
    pub division: u8,
    pub seed: u8,
    pub mons: Vec<i16>,
    pub wins: u8,
    pub losses: u8,
    pub differential: i16,
    #[serde(default)]
    pub awards: HashMap<String, String>,
    pub next_pick: Option<i16>,
    #[serde(default)]
    pub used_trades: u8,
    #[serde(default)]
    pub extra_trades: u8,
    #[serde(default)]
    pub playoff_differential: i16,
    pub challonge_participant: Option<u64>,
    #[serde(default)]
    pub dropout: bool,
    #[serde(default)]
    pub extra_picks: HashMap<String, u8>,
    #[serde(default)]
    pub megas: HashSet<i16>,
    #[serde(default)]
    pub megas_allowed: u8,
}

#[async_trait]
impl Listable for Coach {
    fn type_name() -> String {
        "Coach".to_string()
    }

    fn type_name_plural() -> String {
        "Coaches".to_string()
    }

    async fn field_name(&self, ctx: &Context) -> String {
        self.display_name(ctx, None).await
    }

    async fn field_value(&self, _: &Context) -> String {
        match (self.division > 0, self.awards.len() > 0) {
            (true, true) => {
                format!("Division {} | {} Wins | {} Losses ({:+}) | {} Used Trades | {} Awards", self.division, self.wins, self.losses, self.differential, self.used_trades, self.awards.len())
            },
            (true, false) => {
                format!("Division {} | {} Wins | {} Losses ({:+}) | {} Used Trades", self.division, self.wins, self.losses, self.differential, self.used_trades)
            },
            (false, true) => {
                format!("{} Wins | {} Losses ({:+}) | {} Used Trades | {} Awards", self.wins, self.losses, self.differential, self.used_trades, self.awards.len())
            },
            (false, false) => {
                format!("{} Wins | {} Losses ({:+}) | {} Used Trades", self.wins, self.losses, self.differential, self.used_trades)
            }
        }
    }

    fn valid_sorts() -> Vec<SortType> {
        todo!()
    }

    fn default_sort() -> SortType {
        todo!()
    }

    fn sort_str(&self, _mode: SortType) -> String {
        todo!()
    }

    fn sort_num(&self, _mode: SortType) -> isize {
        todo!()
    }

    fn sort_bool(&self, _mode: SortType) -> bool {
        todo!()
    }

    fn sort_date(&self) -> DateTime<Utc> {
        todo!()
    }
}

impl Awardable for Coach {
    fn awards<'draft>(&'draft self) -> &'draft HashMap<String, String> {
        &self.awards
    }
    fn awards_mut<'draft>(&'draft mut self) -> &'draft mut HashMap<String, String> {
        &mut self.awards
    }
}

impl Coach {
    pub async fn display_name(&self, ctx: &Context, server_id: Option<u64>) -> String {
        if self.name.is_some() {
            self.name.as_ref().unwrap().to_owned()
        }
        else {
            if let Ok(u) = ctx.http.get_user(self.discord_id).await {
                if server_id.is_some() {
                    u.nick_in(ctx, server_id.unwrap()).await.unwrap_or(u.name)
                }
                else {
                    u.name
                }
            }
            else {
                self.discord_name.clone()
            }
        }
    }

    pub fn get_mons<'a>(&self, draft: &'a Draft) -> Vec<&'a DraftPokemon> {
        let t: Vec<&'a DraftPokemon> = self.mons.iter().map(|i| &draft.mons[i]).collect();
        t
    }

    pub fn _get_picks<'a>(&self, draft: &'a Draft) -> HashMap<String, u8> {
        let p = draft.tier_picks[&self.division].clone();
        p
    }

    pub fn get_remaining_picks<'a>(&self, draft: &'a Draft, tier: &String) -> u8 {
        if draft.tier_picks.len() != 0 && draft.tier_picks.contains_key(&self.division) {
            let mut p = draft.tier_picks[&self.division].clone();
            p.iter_mut().for_each(|(t, i)| {
                if self.extra_picks.contains_key(t) {
                    *i += self.extra_picks[t];
                }
            });
            for (t, i) in &self.extra_picks {
                if !p.contains_key(t) {
                    p.insert(t.clone(), *i);
                }
            }
            if p.len() != 0 && p.contains_key(tier) {
                let m = self.get_mons(draft).iter().filter(|p| &p.tier == tier).count();
                p[tier] - m as u8
            }
            else if draft.tier_picks.contains_key(&0) {
                let p = draft.tier_picks[&0].clone();
                if p.len() != 0 && p.contains_key(tier) {
                    let m = self.get_mons(draft).iter().filter(|p| &p.tier == tier).count();
                    p[tier] - m as u8
                }
                else {
                    0
                }
            }
            else {
                0
            }
        }
        else if draft.tier_picks.contains_key(&0) {
            let mut p = draft.tier_picks[&0].clone();
            p.iter_mut().for_each(|(t, i)| {
                if self.extra_picks.contains_key(t) {
                    *i += self.extra_picks[t];
                }
            });
            for (t, i) in &self.extra_picks {
                if !p.contains_key(t) {
                    p.insert(t.clone(), *i);
                }
            }
            if p.len() != 0 && p.contains_key(tier) {
                let m = self.get_mons(draft).iter().filter(|p| &p.tier == tier).count();
                p[tier] - m as u8
            }
            else {
                0
            }
        }
        else {
            draft.total_picks - self.mons.len() as u8
        }
    }

    pub fn trade_limit<'a>(&self, draft: &'a Draft) -> u8 {
        draft.trades_max + self.extra_trades
    }

    pub fn _playoff_wins<'a>(&self, draft: &'a Draft) -> Option<u8> {
        if draft.playoff_matches.len() > 0 {
            let won = draft.playoff_matches.iter().filter(|p| {
                p.complete &&
                (p.coach1 == self.discord_id && p.winner_is_1.clone().unwrap_or(false)) || (p.coach2 == self.discord_id && !p.winner_is_1.clone().unwrap_or(false))
            });
            Some(won.count() as u8)
        }
        else {
            None
        }
    }

    pub fn playoff_losses<'a>(&self, draft: &'a Draft) -> Option<u8> {
        if draft.playoff_matches.len() > 0 {
            let lost = draft.playoff_matches.iter().filter(|p| {
                p.complete &&
                (p.coach1 == self.discord_id && !p.winner_is_1.clone().unwrap_or(false)) || (p.coach2 == self.discord_id && p.winner_is_1.clone().unwrap_or(false))
            });
            Some(lost.count() as u8)
        }
        else {
            None
        }
    }

    pub fn playoff_wins_from_season<'a>(&self, result: &'a SeasonResult) -> Option<u8> {
        if result.playoff_matches.len() > 0 {
            let won = result.playoff_matches.iter().filter(|p| {
                p.complete &&
                (p.coach1 == self.discord_id && p.winner_is_1.clone().unwrap()) || (p.coach2 == self.discord_id && !p.winner_is_1.clone().unwrap())
            });
            Some(won.count() as u8)
        }
        else {
            None
        }
    }

    pub fn playoff_losses_from_season<'a>(&self, result: &'a SeasonResult) -> Option<u8> {
        if result.playoff_matches.len() > 0 {
            let lost = result.playoff_matches.iter().filter(|p| {
                p.complete &&
                (p.coach1 == self.discord_id && !p.winner_is_1.clone().unwrap()) || (p.coach2 == self.discord_id && p.winner_is_1.clone().unwrap())
            });
            Some(lost.count() as u8)
        }
        else {
            None
        }
    }
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct SeasonResult {
    pub draft_id: String,
    pub season: u16,
    pub coaches: HashMap<u64, Coach>,
    pub standings: Vec<u64>,
    pub draft_order: Vec<u64>,
    pub matches: Vec<Match>,
    pub playoff_matches: Vec<PlayoffMatch>,
    pub playoff_challonge: Option<String>,
    pub mons: HashMap<i16, DraftPokemon>,
    #[serde(default)]
    pub winning_team: Option<u8>,
    #[serde(default)]
    pub division_names: Option<HashMap<u8, String>>,
    #[serde(default)]
    pub team_divisions: bool,
}

impl SeasonResult {
    pub fn new_from_draft(draft: &Draft) -> Result<SeasonResult, String> {
        if draft.league {
            if vec![DraftStatus::Playing, DraftStatus::SeasonPlayoffs].contains(&draft.status) {
                Ok(
                    SeasonResult {
                        draft_id: draft.draft_id.clone(),
                        season: draft.league_season,
                        coaches: draft.coaches.clone(),
                        standings: Vec::new(),
                        draft_order: draft.draft_order.clone(),
                        matches: draft.matches.clone(),
                        playoff_matches: draft.playoff_matches.clone(),
                        playoff_challonge: draft.challonge.clone(),
                        mons: draft.mons.clone(),
                        winning_team: None,
                        division_names: draft.division_names.clone(),
                        team_divisions: draft.team_divisions,
                    }
                )
            }
            else {
                Err("Can't create season results when drafting is still in progress".into())
            }
        }
        else {
            Err("Can't create season results from a tournament draft".into())
        }
    }

    pub async fn embed(&self, ctx: &Context) -> CreateEmbed {
        let logo = ctx.get_draft_logo(&self.draft_id).await;
        let dname = ctx.get_draft_name(self.draft_id.clone()).await;
        let mvp = self.mons.iter().filter(|(_, m)| m.has_award(format!("{} Season {} MVP", dname, self.season))).next();

        let mut e = CreateEmbed::default();
        e.title(format!("{} Season {} Final Results", dname, self.season));
        if logo.is_some() {
            e.thumbnail(logo.unwrap());
        }
        e.field("Regular Season Matches", self.matches.iter().filter(|m| !m.playoffs).count(), true);
        if self.playoff_matches.len() > 0 {
            e.field("Playoff Matches", self.playoff_matches.len(), true);
        }
        if self.team_divisions {
            let win_team_name = match &self.division_names {
                Some(n) => {
                    n[&self.winning_team.unwrap()].clone()
                },
                None => {
                    format!("Division {}", self.winning_team.unwrap())
                }
            };

            let team_standings = {
                let mut s = String::new();
                let mut i = 1;
                let team_coaches = self.coaches.iter().filter(|(_, c)| c.division == self.winning_team.unwrap())
                    .sorted_by(|(_, a), (_, b)| b.playoff_differential.cmp(&a.playoff_differential));
                
                for (_, c) in team_coaches {
                    let record = if self.playoff_matches.len() > 0 {
                        let win = c.playoff_wins_from_season(self).unwrap();
                        let loss = c.playoff_losses_from_season(self).unwrap();
                        format!("({}-{} {:+})", win, loss, c.playoff_differential)
                    }
                    else {
                        format!("({}-{} {:+})", c.wins, c.losses, c.differential)
                    };
                    let cname = c.display_name(ctx, None).await;
                    let place = if i == 1 {
                        format!("**{}: {} {}**\n", i, cname, record)
                    } else {
                        format!("{}: {} {}\n", i, cname, record)
                    };
                    s.push_str(place.as_str());
    
                    i += 1;
                }

                s
            };

            e.field(format!("{} Won", win_team_name), team_standings, false);
        }
        else {
            let standings = {
                let mut s = String::new();
                let mut i = 1;
                for c in &self.standings {
                    let c = &self.coaches[c];
                    let record = if self.playoff_matches.len() > 0 {
                        let win = c.playoff_wins_from_season(self).unwrap();
                        let loss = c.playoff_losses_from_season(self).unwrap();
                        format!("({}-{} {:+})", win, loss, c.playoff_differential)
                    }
                    else {
                        format!("({}-{} {:+})", c.wins, c.losses, c.differential)
                    };
                    let cname = c.display_name(ctx, None).await;
                    let place = if i == 1 {
                        format!("**{}: {} {}**\n", i, cname, record)
                    } else {
                        format!("{}: {} {}\n", i, cname, record)
                    };
                    s.push_str(place.as_str());
    
                    i += 1;
                }
                s
            };
            e.field("Standings", standings, false);
        }

        if let Some((_, m)) = mvp {
            e.field("Most Valuable Pokemon", format!("{}\n{}", m.field_name(ctx).await, m.field_value(ctx).await), false);
        }

        e.color(EMBED_COLOR);

        e
    }
}

#[async_trait]
impl Listable for SeasonResult {
    fn type_name() -> String {
        "Season".to_string()
    }

    fn type_name_plural() -> String {
        "Seasons".to_string()
    }

    async fn field_name(&self, _: &Context) -> String {
        format!("Season {}", self.season)
    }

    async fn field_value(&self, ctx: &Context) -> String {
        if self.standings.len() >= 3 {
            format!("**1st Place: {}** | 2nd Place: {} | 3rd Place: {}",
                self.coaches[&self.standings[0]].display_name(ctx, None).await,
                self.coaches[&self.standings[1]].display_name(ctx, None).await,
                self.coaches[&self.standings[2]].display_name(ctx, None).await)
        }
        else {
            format!("**Winner: {}** | Runner-up: {}",
                self.coaches[&self.standings[0]].display_name(ctx, None).await,
                self.coaches[&self.standings[1]].display_name(ctx, None).await)
        }
    }

    fn valid_sorts() -> Vec<SortType> {
        todo!()
    }

    fn default_sort() -> SortType {
        todo!()
    }

    fn sort_str(&self, _mode: SortType) -> String {
        todo!()
    }

    fn sort_num(&self, _mode: SortType) -> isize {
        todo!()
    }

    fn sort_bool(&self, _mode: SortType) -> bool {
        todo!()
    }

    fn sort_date(&self) -> DateTime<Utc> {
        todo!()
    }
}

fn default_letter() -> String {
    "A".to_string()
}

pub(crate) static NEXT_LETTER: phf::Map<&'static str, &'static str> = phf_map! {
        "A" => "B",
        "B" => "C",
        "C" => "D",
        "D" => "E",
        "E" => "F",
        "F" => "G",
        "G" => "H",
        "H" => "I",
        "I" => "J",
        "J" => "K",
        "K" => "L",
        "L" => "M",
        "M" => "N",
        "N" => "O",
        "O" => "P",
        "P" => "Q",
        "Q" => "R",
        "R" => "S",
        "S" => "T",
        "T" => "U",
        "U" => "V",
        "V" => "W",
        "W" => "X",
        "X" => "Y",
        "Y" => "Z",
        "Z" => "AA",
        "AA" => "AB",
        "AB" => "AC",
        "AC" => "AD",
        "AD" => "AE",
        "AE" => "AF",
        "AF" => "AG",
        "AG" => "AH",
        "AH" => "AI",
        "AI" => "AJ",
        "AJ" => "AK",
        "AK" => "AL",
        "AL" => "AM",
        "AM" => "AN",
        "AN" => "AO",
        "AO" => "AP",
        "AP" => "AQ",
        "AQ" => "AR",
        "AR" => "AS",
        "AS" => "AT",
        "AT" => "AU",
        "AU" => "AV",
        "AV" => "AW",
        "AW" => "AX",
        "AX" => "AY",
        "AY" => "AZ",
};

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Draft {
    pub draft_id: String,
    pub draft_name: String,
    pub logo: Option<String>,
    pub owner_id: u64,
    pub moderator_ids: Option<Vec<u64>>,
    pub moderator_role_id: Option<u64>,
    pub player_role_id: Option<u64>,
    pub announcements_channel: Option<u64>,
    pub porygon_channel: Option<u64>,
    pub server_id: u64,
    pub challonge: Option<String>,
    pub format: Format,
    pub tier_source: TierSource,
    pub tiers_allowed: Vec<String>,
    pub league: bool,
    pub league_season: u16,
    pub total_picks: u8,
    pub tier_picks: HashMap<u8, HashMap<String, u8>>,
    pub trade_policy: TradePolicy,
    pub trades_max: u8,
    pub coaches: HashMap<u64, Coach>,
    pub status: DraftStatus,
    pub draft_type: DraftType,
    pub draft_order: Vec<u64>,
    pub draft_round: Vec<u64>,
    pub matches: Vec<Match>,
    #[serde(default)]
    pub playoff_matches: Vec<PlayoffMatch>,
    pub past_seasons: Option<HashMap<u16, SeasonResult>>,
    pub mons: HashMap<i16, DraftPokemon>,
    #[serde(default)]
    pub closed: bool,
    #[serde(default)]
    pub last_round_drafted: Vec<(u64, i16)>,
    #[serde(default)]
    pub playoff_best_of: u8,
    #[serde(default)]
    pub division_names: Option<HashMap<u8, String>>,
    #[serde(default)]
    pub team_divisions: bool,
    #[serde(default = "default_letter")]
    pub next_playoff_letter: String,
}

impl Draft {
    /*
    pub fn new_league(owner_id: u64, draft_name: String, draft_id: String, server_id: u64, format: Format, tier_source: TierSource, tiers_allowed: Vec<String>, tiered_mons: Option<HashMap<i16, DraftPokemon>>) -> Draft {
        let past_seasons: HashMap<u16, SeasonResult> = HashMap::new();
        Self::new(owner_id, draft_name, draft_id, server_id, format, tier_source, tiers_allowed, tiered_mons, true, 1, Some(past_seasons))
    }

    pub fn new_tournament(owner_id: u64, draft_name: String, draft_id: String, server_id: u64, format: Format, tier_source: TierSource, tiers_allowed: Vec<String>, tiered_mons: Option<HashMap<i16, DraftPokemon>>) -> Draft {
        Self::new(owner_id, draft_name, draft_id, server_id, format, tier_source, tiers_allowed, tiered_mons, false, 0, None)
    }

    pub fn new(owner_id: u64, draft_name: String, draft_id: String, server_id: u64, format: Format, tier_source: TierSource, tiers_allowed: Vec<String>, tiered_mons: Option<HashMap<i16, DraftPokemon>>, league: bool, league_season: u16, past_seasons: Option<HashMap<u16, SeasonResult>>) -> Draft {
        let mut d = Draft {
            draft_id,
            draft_name,
            logo: None,
            owner_id,
            moderator_ids: None,
            moderator_role_id: None,
            announcements_channel: None,
            porygon_channel: None,
            server_id,
            challonge: None,
            format,
            tier_source,
            tiers_allowed,
            league,
            league_season,
            total_picks: 6,
            tier_picks: HashMap::new(),
            trade_policy: TradePolicy::Disabled,
            trades_max: 0,
            coaches: HashMap::new(),
            status: DraftStatus::Init,
            draft_type: DraftType::Linear,
            draft_order: Vec::new(),
            draft_round: Vec::new(),
            matches: Vec::new(),
            past_seasons,
            mons: HashMap::new(),
        };
        match d.tier_source {
            TierSource::Smogon => {
                d.mons = tiered_mons.unwrap();
            },
            _ => {
                debug!("no tiers yet");
            }
        }
        d
    }
    */

    pub fn _playoff_best_of(&self) -> u16 {
        self.playoff_best_of as u16 * 2 + 1
    }

    pub fn update_playoff_match(&self, pmatch: &PlayoffMatch) -> Result<(), String> {
        let c = match (env::var("CHALLONGE_USER"), env::var("CHALLONGE_TOKEN")) {
            (Ok(u), Ok(k)) => {
                Challonge::new(u, k)
            },
            _ => {
                return Err("Missing Challonge API keys in environment".into());
            }
        };
        let id = match self.challonge.clone() {
            Some(i) => {
                TournamentId::Url("".into(), i)
            },
            None => {
                return Err("Challonge tournament not created automatically or set manually".into());
            }
        };
        let mid = MatchId(pmatch.challonge_id);
        let m = c.get_match(&id, &mid, false).unwrap();
        let (cha_c1, cha_c2) = (m.player1.id.0, m.player2.id.0);
        let (dsc_c1, dsc_c2) = (
            match self.participant_id_to_discord_id(cha_c1) {
                Some(i) => i,
                None => {
                    return Err("Challonge participants don't match coaches!".into());
                }
            },
            match self.participant_id_to_discord_id(cha_c2) {
                Some(i) => i,
                None => {
                    return Err("Challonge participants don't match coaches!".into());
                }
            }
        );
        let (direct, reverse) = (
            dsc_c1 == pmatch.coach1 && dsc_c2 == pmatch.coach2,
            dsc_c1 == pmatch.coach2 && dsc_c2 == pmatch.coach1
        );
        if !direct && !reverse {
            return Err("Challonge match participants are incorrectly mapped to coaches. This may need manual correction.".into());
        }
        let mut mu = MatchUpdate::new();
        mu.player1_votes(0);
        mu.player2_votes(0);
        let mut scores = Vec::new();
        for _match in &pmatch.matches {
            let (match_direct, match_reverse) = (
                dsc_c1 == _match.coach1 && dsc_c2 == _match.coach2,
                dsc_c1 == _match.coach2 && dsc_c2 == _match.coach1
            );
            if match_direct {
                let score = MatchScore(_match.c1_score as u64, _match.c2_score as u64);
                scores.push(score);
            }
            else if match_reverse {
                let score = MatchScore(_match.c2_score as u64, _match.c1_score as u64);
                scores.push(score);
            }
        }
        let scores = MatchScores(scores);
        mu.scores_csv(scores);
        if pmatch.winner_is_1.is_some() {
            let winner1 = *pmatch.winner_is_1.as_ref().unwrap();
            let winner_participant_id = if winner1 {
                self.discord_id_to_participant_id(pmatch.coach1).unwrap()
            }
            else {
                self.discord_id_to_participant_id(pmatch.coach2).unwrap()
            };
            
            mu.winner_id(ParticipantId(winner_participant_id));
        }

        let _ = c.mark_as_underway(&id, &mid);

        if let Err(y) = c.update_match(&id, &mid, &mu) {
            error!("Update Playoff Match: {:?}", y);
            info!("{:?}", mu);
            return Err("There was a problem updating the Challonge bracket.".into());
        }

        let new_index = c.match_index(&id, None, None).unwrap();
        let open = new_index.0.iter().filter(|m| m.state == MatchState::Open && (
            (
                m.player1.id.0 == self.discord_id_to_participant_id(pmatch.coach1).unwrap() &&
                m.player2.id.0 == self.discord_id_to_participant_id(pmatch.coach2).unwrap()
            ) ||
            (
                m.player1.id.0 == self.discord_id_to_participant_id(pmatch.coach2).unwrap() &&
                m.player2.id.0 == self.discord_id_to_participant_id(pmatch.coach1).unwrap()
            )
        )).collect::<Vec<&challonge::Match>>();
        if open.len() == 1 {
            let nm = open.first().unwrap();
            let (ncha_c1, ncha_c2) = (nm.player1.id.0, nm.player2.id.0);
            let (ndsc_c1, ndsc_c2) = (
                match self.participant_id_to_discord_id(ncha_c1) {
                    Some(i) => i,
                    None => {
                        return Err("Challonge participants don't match coaches!".into());
                    }
                },
                match self.participant_id_to_discord_id(ncha_c2) {
                    Some(i) => i,
                    None => {
                        return Err("Challonge participants don't match coaches!".into());
                    }
                }
            );
            let (direct, reverse) = (
                ndsc_c1 == pmatch.coach1 && ndsc_c2 == pmatch.coach2,
                ndsc_c1 == pmatch.coach2 && ndsc_c2 == pmatch.coach1
            );
            if !direct && !reverse {
                warn!("Error in 'grand finals': ndsc_1 {} pc1 {} ndsc_2 {} pc2 {}", ndsc_c1, pmatch.coach1, ndsc_c2, pmatch.coach2);
                return Err("There appears to be an error in the grand finals. Please review the bracket manually.".into());
            }
            let (ndsc_c1_fromlosers, ndsc_c2_fromlosers) = (
                self.coaches[&ndsc_c1].playoff_losses(self).unwrap() >= 1,
                self.coaches[&ndsc_c2].playoff_losses(self).unwrap() >= 1
            );

            if pmatch.winner_is_1.is_some() {
                let winner1 = *pmatch.winner_is_1.as_ref().unwrap();
                let winner_participant_id = if winner1 {
                    self.discord_id_to_participant_id(pmatch.coach1).unwrap()
                }
                else {
                    self.discord_id_to_participant_id(pmatch.coach2).unwrap()
                };
                match ((ndsc_c1 == dsc_c1 && ndsc_c2 == dsc_c2), (ndsc_c1 == dsc_c2 && ndsc_c2 == dsc_c1), direct, reverse, winner1) {
                    // on challonge, coach orders are the same for both matches (1T, 2F)
                    // the new match has the same coach order as the pmatch (3T, 4F)
                    // the first coach in pmatch won that pmatch (5T)
                    (true, false, true, false, true) => {
                        // NOW check if that first coach came from losers bracket
                        if ndsc_c1_fromlosers {
                            Ok(())
                        }
                        // if they are NOT, then the second finals match is skipped. we update it with the same winner and 0-0 scores
                        else {
                            let mut mu = MatchUpdate::new();
                            mu.winner_id(ParticipantId(winner_participant_id));
                            mu.scores_csv(MatchScores(vec![MatchScore(0, 0)]));

                            if let Err(y) = c.update_match(&id, &nm.id, &mu) {
                                error!("Update Playoff Match: {:?}", y);
                                info!("{:?}", mu);
                                return Err("There was a problem updating the Challonge bracket.".into());
                            }
                            else {
                                Ok(())
                            }
                        }
                    },
                    // on challonge, coach orders are the same for both matches (1T, 2F)
                    // the new match has the same coach order as the pmatch (3T, 4F)
                    // the first coach in pmatch did NOT win that pmatch (5F)
                    (true, false, true, false, false) => {
                        // NOW check if that SECOND coach came from losers bracket
                        if ndsc_c2_fromlosers {
                            Ok(())
                        }
                        // if they are NOT, then the second finals match is skipped. we update it with the same winner and 0-0 scores
                        else {
                            let mut mu = MatchUpdate::new();
                            mu.winner_id(ParticipantId(winner_participant_id));
                            mu.scores_csv(MatchScores(vec![MatchScore(0, 0)]));

                            if let Err(y) = c.update_match(&id, &nm.id, &mu) {
                                error!("Update Playoff Match: {:?}", y);
                                info!("{:?}", mu);
                                return Err("There was a problem updating the Challonge bracket.".into());
                            }
                            else {
                                Ok(())
                            }
                        }
                    },
                    // on challonge, coach orders are NOT the same for both matches (1F, 2T)
                    // the new match has the same coach order as the pmatch (3T, 4F)
                    // the first coach in pmatch won that pmatch (5T)
                    (false, true, true, false, true) => {
                        // NOW check if the FIRST coach in the NEW match (SECOND in OLD match, FIRST in PMATCH, WINNER) came from losers bracket
                        if ndsc_c1_fromlosers {
                            Ok(())
                        }
                        // if they are NOT, then the second finals match is skipped. we update it with the same winner and 0-0 scores
                        else {
                            let mut mu = MatchUpdate::new();
                            mu.winner_id(ParticipantId(winner_participant_id));
                            mu.scores_csv(MatchScores(vec![MatchScore(0, 0)]));

                            if let Err(y) = c.update_match(&id, &nm.id, &mu) {
                                error!("Update Playoff Match: {:?}", y);
                                info!("{:?}", mu);
                                return Err("There was a problem updating the Challonge bracket.".into());
                            }
                            else {
                                Ok(())
                            }
                        }
                    },
                    // on challonge, coach orders are NOT the same for both matches (1F, 2T)
                    // the new match has the same coach order as the pmatch (3T, 4F)
                    // the first coach in pmatch did NOT win that pmatch (5F)
                    (false, true, true, false, false) => {
                        // NOW check if the SECOND coach in the NEW match (FIRST in OLD match, SECOND in PMATCH, WINNER) came from losers bracket
                        if ndsc_c2_fromlosers {
                            Ok(())
                        }
                        // if they are NOT, then the second finals match is skipped. we update it with the same winner and 0-0 scores
                        else {
                            let mut mu = MatchUpdate::new();
                            mu.winner_id(ParticipantId(winner_participant_id));
                            mu.scores_csv(MatchScores(vec![MatchScore(0, 0)]));

                            if let Err(y) = c.update_match(&id, &nm.id, &mu) {
                                error!("Update Playoff Match: {:?}", y);
                                info!("{:?}", mu);
                                return Err("There was a problem updating the Challonge bracket.".into());
                            }
                            else {
                                Ok(())
                            }
                        }
                    },
                    // on challonge, coach orders are NOT the same for both matches (1F, 2T)
                    // the new match is in REVERSE COACH ORDER to the pmatch (3F, 4T)
                    // the first coach in pmatch won that pmatch (5T)
                    (false, true, false, true, true) => {
                        // pmatch WINNER is SECOND in new match. check if they're from the losers bracket
                        if ndsc_c2_fromlosers {
                            Ok(())
                        }
                        // if they are NOT, then the second finals match is skipped. we update it with the same winner and 0-0 scores
                        else {
                            let mut mu = MatchUpdate::new();
                            mu.winner_id(ParticipantId(winner_participant_id));
                            mu.scores_csv(MatchScores(vec![MatchScore(0, 0)]));

                            if let Err(y) = c.update_match(&id, &nm.id, &mu) {
                                error!("Update Playoff Match: {:?}", y);
                                info!("{:?}", mu);
                                return Err("There was a problem updating the Challonge bracket.".into());
                            }
                            else {
                                Ok(())
                            }
                        }
                    },
                    // on challonge, coach orders are NOT the same for both matches (1F, 2T)
                    // the new match is in REVERSE COACH ORDER to the pmatch (3F, 4T)
                    // the first coach in pmatch did NOT win that pmatch (5F)
                    (false, true, false, true, false) => {
                        // pmatch WINNER is FIRST in new match. check if they're from the losers bracket
                        if ndsc_c1_fromlosers {
                            Ok(())
                        }
                        // if they are NOT, then the second finals match is skipped. we update it with the same winner and 0-0 scores
                        else {
                            let mut mu = MatchUpdate::new();
                            mu.winner_id(ParticipantId(winner_participant_id));
                            mu.scores_csv(MatchScores(vec![MatchScore(0, 0)]));

                            if let Err(y) = c.update_match(&id, &nm.id, &mu) {
                                error!("Update Playoff Match: {:?}", y);
                                info!("{:?}", mu);
                                return Err("There was a problem updating the Challonge bracket.".into());
                            }
                            else {
                                Ok(())
                            }
                        }
                    },
                    _ => Err("Ran into an unexpected error trying to bypass second grand finals match. Check the bracket manually.".into())
                }
            }
            else {
                Ok(())
            }
        }
        else {
            Ok(())
        }
    }

    pub fn participant_id_to_discord_id(&self, participant: u64) -> Option<u64> {
        let coaches = self.coaches.iter().filter(|(_, c)| c.challonge_participant.is_some() && c.challonge_participant.as_ref().unwrap() == &participant)
        .map(|(i, _)| *i).collect::<Vec<u64>>();
        coaches.first().map(|i| *i)
    }

    pub fn discord_id_to_participant_id(&self, discord: u64) -> Option<u64> {
        let coaches = self.coaches.iter().filter(|(_, c)| c.discord_id == discord && c.challonge_participant.is_some())
        .map(|(_, c)| *c.challonge_participant.as_ref().unwrap()).collect::<Vec<u64>>();
        coaches.first().map(|i| *i)
    }

    pub fn fix_participants(&mut self) -> Result<(), (String, Draft)> {
        let backup = self.clone();

        let c = match (env::var("CHALLONGE_USER"), env::var("CHALLONGE_TOKEN")) {
            (Ok(u), Ok(k)) => {
                Challonge::new(u, k)
            },
            _ => {
                return Err(("Missing Challonge API keys in environment".into(), backup));
            }
        };
        let id = match self.challonge.clone() {
            Some(i) => {
                TournamentId::Url("".into(), i)
            },
            None => {
                return Err(("Challonge tournament not created automatically or set manually".into(), backup));
            }
        };

        let p = c.participant_index(&id).expect("get it").0.iter().map(|p| (p.misc.parse::<u64>().unwrap(), p.clone())).collect::<HashMap<u64, Participant>>();
        let co = self.coaches.iter_mut().filter(|(_, c)| !c.dropout);
        for (i, c) in co {
            c.challonge_participant = Some(p[i].id.0);
            c.seed = p[i].seed as u8
        }

        Ok(())
    }

    pub fn create_playoff_challonge(&mut self, days: Option<u8>) -> Result<(), (String, Draft)> {
        let backup = self.clone();

        let c = match (env::var("CHALLONGE_USER"), env::var("CHALLONGE_TOKEN")) {
            (Ok(u), Ok(k)) => {
                Challonge::new(u, k)
            },
            _ => {
                return Err(("Missing Challonge API keys in environment".into(), backup));
            }
        };
        let mut tc = TournamentCreate::new();
        let mut url = format!("{}{}", self.draft_id, self.league_season);
        let dur = Duration::days(days.unwrap_or(7) as i64) + Duration::minutes(5);
        let start = Utc::now() + dur;
        tc.name(format!("{} Season {} Playoffs", self.draft_name, self.league_season))
            .game_name("Pokémon Showdown")
            .ranked_by(RankedBy::MatchWins)
            .tournament_type(challonge::TournamentType::DoubleElimination)
            .url(&url)
            .check_in_duration(368640);
        tc.start_at = Some(start);
        let mut tries = 0;
        let t = loop {
            match c.create_tournament(&tc) {
                Ok(t) => break t,
                Err(y) => {
                    match tries {
                        0 => {
                            url = format!("{}_s{}", self.draft_id, self.league_season);
                            tc.url(&url);
                            tries += 1;
                        },
                        1 => {
                            url = format!("{}_{}", self.draft_id, self.league_season);
                            tc.url(&url);
                            tries += 1;
                        },
                        2 => {
                            url = format!("{}season{}", self.draft_id, self.league_season);
                            tc.url(&url);
                            tries += 1;
                        },
                        _ => {
                            error!("Can't create tourn: {:?}", y);
                            return Err(("Couldn't create tournament on Challonge. Either all the possible URLs are taken, or the API is experiencing issues.".into(), backup));
                        }
                    }
                } 
            }
        };
        let mut next_seed = 1;
        let divisions = self.coaches.iter().filter(|(_, c)| !c.dropout).map(|(_, c)| c.division).unique().count();
        if divisions > 1 {
            info!("# of divisions: {}", divisions);
            let c_groups = self.coaches.iter()
                .filter(|(_, c)| !c.dropout)
                .sorted_by(|(_, a), (_, b)| {
                    let a_key = (a.division, a.wins as i16 - a.losses as i16, a.differential);
                    let b_key = (b.division, b.wins as i16 - b.losses as i16, b.differential);
                    b_key.cmp(&a_key)
                })
                .group_by(|(_, c)| c.division);
            let mut top_per_div = Vec::new();
            for (i, group) in c_groups.into_iter() {
                let group = group.map(|(_, c)| c).collect::<Vec<&Coach>>();
                info!("{} Coaches in Division {}", group.len(), i);
                let c = group.iter().sorted_by(|a, b| {
                    let a_key = (a.wins as i16 - a.losses as i16, a.differential);
                    let b_key = (b.wins as i16 - b.losses as i16, b.differential);
                    b_key.cmp(&a_key)
                }).map(|c| c).next().unwrap();
                top_per_div.push(c.discord_id);
            }
            let top_coaches_div = self.coaches.iter_mut()
                .filter(|(_, c)| !c.dropout)
                .filter(|(_, c)| top_per_div.contains(&c.discord_id))
                .sorted_by(|(_, a), (_, b)| {
                    let a_key = (a.wins as i16 - a.losses as i16, a.differential);
                    let b_key = (b.wins as i16 - b.losses as i16, b.differential);
                    b_key.cmp(&a_key)
                })
                .map(|(_, c)| c);
            for c in top_coaches_div {
                info!("Seeding Top Coach {} in Division {}", c.discord_name, c.division);
                c.seed = next_seed;
                next_seed += 1;
            }
        }
        for (_, c) in self.coaches.iter_mut().filter(|(_, c)| !c.dropout && c.seed == 0).sorted_by(|(_, a), (_, b)| {
            let a_key = (a.wins as i16 - a.losses as i16, a.differential);
            let b_key = (b.wins as i16 - b.losses as i16, b.differential);
            b_key.cmp(&a_key)
        }) {
            info!("Coach {} gets seed {}", c.discord_name, next_seed);
            c.seed = next_seed;
            next_seed += 1;
        }
        let pc = self.coaches.iter().filter(|(_, c)| !c.dropout).sorted_by(|(_, a), (_, b)| a.seed.cmp(&b.seed))
            .map(|(_, c)| {
                let mut pc = ParticipantCreate::new();
                pc.name(c.name.as_ref().unwrap())
                    .seed(c.seed as u64)
                    .misc(c.discord_id.to_string());
                (c.discord_id, pc)
            })
            .collect::<HashMap<u64, ParticipantCreate>>();
        let p = pc.iter().sorted_by(|(_, a), (_, b)| a.seed.cmp(&b.seed))
            .map(|(i, p)| {
                (*i, c.create_participant(&t.id, p))
            }).collect::<HashMap<u64, Result<Participant, challonge::error::Error>>>();
        if p.iter().any(|(_, r)| r.is_err()) {
            let y = p.iter().filter(|(_, p)| p.is_err()).next().unwrap().1.as_ref().unwrap_err();
            error!("Failed adding participants: {:?}", y);
            return Err(("Failed adding participants to Challonge".into(), backup));
        }
        let ph = p.iter().map(|(i, r)| (*i, r.as_ref().unwrap())).collect::<HashMap<u64, &Participant>>();
        info!("pc length: {}, p length: {}, ph length: {}, co length: {}", pc.len(), p.len(), ph.len(), self.coaches.len());
        let co = self.coaches.iter_mut().filter(|(_, c)| !c.dropout);
        for (i, c) in co {
            c.challonge_participant = Some(ph[i].id.0);
        }
        /*
        task::spawn(async move {
            let dur = tc.start_at.unwrap() - Utc::now();
            let _ = delay_for(dur.to_std().unwrap()).await;
            let _ = c.tournament_start(&t.id, &TournamentIncludes::All);
        });*/
        self.status = DraftStatus::SeasonPlayoffs;
        self.challonge = Some(url);
        if let Err(y) = c.fix_participants_count(&t.id) {
            warn!("participant fix failed: {:?}", y)
        }
        if let Err(y) = c.tournament_abort_checkins(&t.id, &TournamentIncludes::All) {
            warn!("cancel checkin failed: {:?}", y)
        }
        if let Err(y) = c.open_for_predictions(&t.id, &TournamentIncludes::All) {
            warn!("predicc open failed: {:?}", y)
        }
        
        Ok(())
    }

    pub fn get_challonge_tournament(&self) -> Result<Tournament, String> {
        let c = match (env::var("CHALLONGE_USER"), env::var("CHALLONGE_TOKEN")) {
            (Ok(u), Ok(k)) => {
                Challonge::new(u, k)
            },
            _ => {
                return Err("Missing Challonge API keys in environment".into());
            }
        };
        let id = match self.challonge.clone() {
            Some(i) => {
                TournamentId::Url("".into(), i)
            },
            None => {
                return Err("Challonge tournament not created automatically or set manually".into());
            }
        };
        c.get_tournament(&id, &TournamentIncludes::All).map_err(|_| {
            "There was an error retrieving the tournament from Challonge. Perhaps it was not created successfully.".into()
        })
    }

    pub fn _start_tournament(&self) {
        let c = match (env::var("CHALLONGE_USER"), env::var("CHALLONGE_TOKEN")) {
            (Ok(u), Ok(k)) => {
                Challonge::new(u, k)
            },
            _ => {
                return;
            }
        };
        let id = match self.challonge.clone() {
            Some(i) => {
                TournamentId::Url("".into(), i)
            },
            None => {
                return;
            }
        };
        let _ = c.tournament_start(&id, &TournamentIncludes::All);
    }

    pub fn get_match_id(&self, coach1_challonge: u64, coach2_challonge: u64, repairing: bool) -> Result<(u64, String), String> {
        let c = match (env::var("CHALLONGE_USER"), env::var("CHALLONGE_TOKEN")) {
            (Ok(u), Ok(k)) => {
                Challonge::new(u, k)
            },
            _ => {
                return Err("Missing Challonge API keys in environment".into());
            }
        };
        let id = match self.challonge.clone() {
            Some(i) => {
                TournamentId::Url("".into(), i)
            },
            None => {
                return Err("Challonge tournament not created automatically or set manually".into());
            }
        };
        let m = match c.match_index(&id, None, None) {
            Ok(m) => m,
            Err(_) => {
                return Err("Match index couldn't be retrieved. The playoff bracket may need to be started manually.".into());
            }
        };
        let filter = m.0.iter().filter(|m| 
            (repairing || (m.winner_id.is_none() && m.state == MatchState::Open)) && (
                (m.player1.id.0 == coach1_challonge && m.player2.id.0 == coach2_challonge) ||
                (m.player2.id.0 == coach1_challonge && m.player1.id.0 == coach2_challonge)
            )
        ).collect::<Vec<&challonge::Match>>();
        if filter.len() == 1 {
            let m = filter.first().unwrap();
            Ok((m.id.0, m.identifier.clone()))
        }
        else {
            error!("No matchy-match: {} repairing {} c1c {} c2c {}", filter.len(), repairing, coach1_challonge, coach2_challonge);
            info!("matches: {:?}", &filter);
            Err("No Challonge match found which involves these coaches.".into())
        }
    }

    pub fn challonge_url(&self) -> Result<String, String> {
        let t = match self.get_challonge_tournament() { Ok(t) => t,
            Err(w) => return Err(w),
        };
        Ok(t.full_challonge_url)
    }

    pub fn finalize_season(&mut self) -> Result<(), (String, Draft)> {
        let backup = self.clone();

        let standings = {
            if self.status == DraftStatus::Playing {
                let records = self.coaches.iter().sorted_by(|(_, a), (_, b)| b.wins.cmp(&a.wins)).map(|(i, _)| *i).collect::<Vec<u64>>();
                records
            } else if self.status == DraftStatus::SeasonPlayoffs {
                if self.team_divisions {
                    Vec::new()
                }
                else {
                    let c = match (env::var("CHALLONGE_USER"), env::var("CHALLONGE_TOKEN")) {
                        (Ok(u), Ok(k)) => {
                            Challonge::new(u, k)
                        },
                        _ => {
                            return Err(("Missing Challonge API keys in environment".into(), backup));
                        }
                    };
                    let id = match self.challonge.clone() {
                        Some(i) => {
                            TournamentId::Url("".into(), i)
                        },
                        None => {
                            return Err(("Challonge tournament not created automatically or set manually".into(), backup));
                        }
                    };
                    let t = c.get_tournament(&id, &TournamentIncludes::All).unwrap();
                    if t.progress_meter != 100 {
                        return Err(("The Challonge bracket does not seem to be complete! Check again if any more matches are waiting to be played.".into(), backup));
                    }
                    let _ = c.tournament_finalize(&id, &TournamentIncludes::All);
                    let p = c.participant_index(&id).unwrap();
                    let q = p.0.iter().map(|p| (p.final_rank.unwrap(), p.seed, p.misc.parse::<u64>().unwrap()));
                    let placed = q.sorted_by(|(a, sa, _), (b, sb, _)| (a, sa).cmp(&(b, sb))).map(|(_, _, i)| i).collect::<Vec<u64>>();
                    let mut dropouts = self.coaches.iter().filter(|(_, c)| c.dropout).collect::<Vec<(&u64, &Coach)>>();
                    if dropouts.len() > 0 {
                        dropouts.sort_by(|(_, a), (_, b)| b.wins.cmp(&a.wins));
                        let mut placed = placed.clone();
                        for (i, _) in dropouts {
                            placed.push(*i);
                        }
                        placed
                    }
                    else {
                        placed
                    }
                }
            }
            else {
                Vec::new()
            }
        };

        let winning_team = if self.team_divisions && self.status == DraftStatus::SeasonPlayoffs {
            let mut team_scores = HashMap::new();
            let finished_pm = self.playoff_matches.iter().filter(|p| p.winner_is_1.is_some());
            for pm in finished_pm {
                let winner_id = if pm.winner_is_1.unwrap() {
                    pm.coach1
                }
                else {
                    pm.coach2
                };
                let team_id = self.coaches[&winner_id].division;
                *team_scores.entry(team_id).or_insert(0) += 1;
            }

            let order = team_scores.iter().sorted_by(|(_, a), (_, b)| b.cmp(a)).map(|(x, _)| *x).collect::<Vec<u8>>();
            Some(order[0])
        }
        else {
            None
        };

        let awardname = if self.league { if self.team_divisions { format!("{} Season {} Team Winner", self.draft_name, self.league_season) } else { format!("{} Season {} Winner", self.draft_name, self.league_season) } } else { format!("{} Winner", self.draft_name) };
        if self.team_divisions {
            let win = winning_team.clone().unwrap();
            let team = self.coaches.iter_mut().filter(|(_, c)| c.division == win);
            for (_, c) in team {
                c.grant_award(awardname.clone());
            }
        }
        else {
            self.coaches.entry(standings[0]).and_modify(|c| {
                c.grant_award(awardname);
            });
        }

        let mvpaward = if self.league { format!("{} Season {} MVP", self.draft_name, self.league_season) } else { format!("{} MVP", self.draft_name) };
        let mons_by_performance = self.mons.iter().filter(|(_, m)| m.coach.is_some())
            .sorted_by(|(_, a), (_, b)| (b.active_kills as i32 + b.passive_kills as i32 - b.deaths as i32).cmp(&(a.active_kills as i32 + a.passive_kills as i32 - a.deaths as i32)))
            .collect::<Vec<(&i16, &DraftPokemon)>>();
        if let Some((i, _)) = mons_by_performance.first() {
            let i = **i;
            self.mons.entry(i).and_modify(|m| {
                m.grant_award(mvpaward);
            });
        }

        let mut season_result = match SeasonResult::new_from_draft(&self) {
            Ok(r) => r,
            Err(y) => {
                return Err((y, backup));
            }
        };

        season_result.standings = standings;
        season_result.winning_team = winning_team;

        // insert past season
        if self.past_seasons.is_none() {
            let mut ps = HashMap::new();
            ps.insert(self.league_season, season_result);
            self.past_seasons = Some(ps);
        }
        else {
            let mut ps = self.past_seasons.clone().unwrap();
            ps.insert(self.league_season, season_result);
            self.past_seasons = Some(ps);
        }

        // update status
        self.status = DraftStatus::SeasonComplete;
        self.closed = true;
        self.draft_order.clear();
        self.draft_round.clear();
        self.last_round_drafted.clear();
        self.trade_policy = TradePolicy::Disabled;

        Ok(())
    }

    pub async fn new_season(&mut self, ctx: &Context) -> Result<(), (String, Draft)> {
        let backup = self.clone();

        if self.status != DraftStatus::SeasonComplete {
            return Err(("A season is currently in progress!".into(), backup));
        }

        let guild = match ctx.http.get_guild(self.server_id).await {
            Ok(g) => g,
            Err(_) => return Err(("Couldn't access draft server".into(), backup))
        };
        if self.player_role_id.is_some() {
            let mut m = match guild.members(ctx, None, None).await {
                Ok(m) => m,
                Err(_) => return Err(("Couldn't access server members list".into(), backup))
            };
            m.retain(|m| self.coaches.contains_key(&m.user.id.0));
    
            for mut m in m {
                let _ = m.remove_role(ctx, self.player_role_id.unwrap()).await;
            }
        }

        self.coaches.clear();
        self.challonge = None;
        self.league_season += 1;
        self.matches.clear();
        self.playoff_matches.clear();
        self.tier_picks.clear();
        for (_, m) in &mut self.mons {
            m.reset(true, false);
        }
        self.status = DraftStatus::Waiting;

        Ok(())
    }

    pub fn save(&self) {
        let string = ron::ser::to_string_pretty(self, ron::ser::PrettyConfig::new()).unwrap();

        let pathn = format!("drafts/{}.ron", self.draft_id);

        let path = Path::new(&pathn);
        let display = path.display();

        let mut file = match File::create(&path) {
            Err(why) => panic!("couldn't create {}: {}", display, why),
            Ok(file) => file,
        };

        match file.write_all(string.as_bytes()) {
            Err(why) => panic!("couldn't write to {}: {}", display, why),
            Ok(_) => debug!("successfully wrote to {}", display),
        }
    }

    pub fn league_tournament_descriptor(&self) -> String {
        if self.league {
            "League".to_owned()
        }
        else {
            "Tournament".to_owned()
        }
    }

    pub async fn embed(&self, ctx: &Context) -> Result<CreateEmbed, Error> {
        let owner = ctx.http.get_user(self.owner_id).await?;
        let owner_name = owner.name;
        let server = ctx.http.get_guild(self.server_id).await?;
        let server_name = server.name.clone();

        let bank = self.mons.iter().filter(|(_, y)| {
            self.tiers_allowed.contains(&y.tier) && !y.drafted
        });
        let bank_size = bank.count();

        let tiers_perm = if self.tiers_allowed.len() != 0 {
            match self.tiers_allowed.len() {
                1 => {
                    Some(self.tiers_allowed.first().unwrap().clone())
                },
                _ => {
                    let mut o = String::new();
                    let mut i = 1;
                    let mut tiers = self.tiers_allowed.clone();
                    tiers.sort_by(|a, b| a.cmp(&b));
                    for s in tiers {
                        if i == self.tiers_allowed.len() {
                            o.push_str(s.as_str());
                        }
                        else {
                            o.push_str(format!("{}, ", s).as_str());

                            i += 1;
                        }
                    }

                    Some(o)
                }
            }
        }
        else { None };

        let c_name = if self.announcements_channel.is_some() {
            let cid = self.announcements_channel.unwrap();
            match server.channels(ctx).await {
                Ok(z) => {
                    let zc: Vec<&ChannelId> = z.iter().filter(|(x, _)| {
                        x.as_u64() == &cid
                    }).map(|(x, _)| x).collect();
                    if zc.len() != 0 {
                        let c = zc.first().unwrap();
                        let n = c.name(ctx).await.unwrap();
                        Some(n)
                    }
                    else {
                        None
                    }
                },
                Err(_) => None,
            }
        }
        else { None };

        let p_name = if self.porygon_channel.is_some() {
            let cid = self.porygon_channel.unwrap();
            match server.channels(ctx).await {
                Ok(z) => {
                    let zc: Vec<&ChannelId> = z.iter().filter(|(x, _)| {
                        x.as_u64() == &cid
                    }).map(|(x, _)| x).collect();
                    if zc.len() != 0 {
                        let c = zc.first().unwrap();
                        let n = c.name(ctx).await.unwrap();
                        Some(n)
                    }
                    else {
                        None
                    }
                },
                Err(_) => None,
            }
        }
        else { None };

        let ttype = if self.league {
            "League"
        } else { 
            "Tournament"
        };

        let dname = if self.draft_name == "Unset".to_owned() { "__Unset__" } else { self.draft_name.as_str() };
        let did = if self.draft_id == "???".to_owned() { "__???__" } else { self.draft_id.as_str() };

        let e = {
            let mut e = CreateEmbed::default();
            if self.logo.is_some() {
                e.thumbnail(self.logo.as_ref().unwrap());
            }
            e.fields(vec![
                ("Draft Name (ID)", format!("{} ({})", dname, did), false),
                ("Commissioner", format!("{}", owner_name), true),
                ("Server", server_name, true),
            ]);
            if c_name.is_some() {
                e.field("Announcements Channel", format!("#{}", c_name.unwrap()), true);
            }
            if p_name.is_some() {
                e.field("Porygon Battle Results Channel (output, not live-links)", format!("#{}", p_name.unwrap()), true);
            }
            e.fields(vec![
                ("Battle Format", format!("{:?}", self.format), true),
                ("Drafting Order", format!("{:?}", self.draft_type), true),
                ("Tier Source", format!("{:?}", self.tier_source), true),
            ]);
            if tiers_perm.is_some() {
                e.field("Tiers Permitted", tiers_perm.unwrap(), false);
            }
            e.fields(vec![
                ("Total Picks", format!("{}", self.total_picks), true),
                ("Format", ttype.to_owned(), true),
            ]);
            if self.league {
                e.field("League Season", self.league_season, true);
            }
            e.fields(vec![
                ("Trade Policy", format!("{}", self.trade_policy), true),
            ]);
            if self.trade_policy != TradePolicy::Disabled {
                e.field("Max Trades", self.trades_max, true);
            }
            if self.coaches.len() != 0 {
                e.field("Enrolled Coaches", self.coaches.len(), true);
            }
            if self.challonge.is_some() {
                let link = self.challonge_url().ok();
                if link.is_some() {
                    e.field("Playoffs Challonge Link", link.unwrap(), false);
                }
            }
            e.fields(vec![
                ("Total Pokémon", format!("{}", self.mons.iter().filter(|(_, m)| self.tiers_allowed.contains(&m.tier)).count()), true),
                ("Remaining in Bank", format!("{}", bank_size), true),
                ("Status", format!("{}", self.status), true),
            ]);
            e.field("Open to Join", !self.closed, true);

            e.color(EMBED_COLOR);

            e
        };

        Ok(e)
    }

    pub fn add_coach(&mut self, coach: Coach) -> Result<(), String> {
        if [DraftStatus::Waiting, DraftStatus::Ready].contains(&self.status) {
            self.coaches.insert(coach.discord_id.clone(), coach);
            if self.status == DraftStatus::Waiting && self.coaches.len() >= 2 {
                self.status = DraftStatus::Ready;
            }
            self.save();

            Ok(())
        }
        else {
            Err("The draft is not open to new coaches.".to_owned())
        }
    }
}

#[async_trait]
impl Listable for Draft {
    fn type_name() -> String {
        "Draft".to_string()
    }

    fn type_name_plural() -> String {
        "Drafts".to_string()
    }

    async fn field_name(&self, _: &Context) -> String {
        format!("[{}] {} ({})", self.draft_id, self.draft_name, self.league_tournament_descriptor())
    }

    async fn field_value(&self, _: &Context) -> String {
        format!("Status: {} | Format: {:?} | Tiers: {:?} | Trades: {} | Players: {}", self.status, self.format, self.tier_source, self.trade_policy, self.coaches.len())
    }

    fn valid_sorts() -> Vec<SortType> {
        vec![SortType::StrID]
    }

    fn default_sort() -> SortType {
        SortType::StrID
    }

    fn sort_str(&self, _mode: SortType) -> String {
        self.draft_id.clone()
    }

    fn sort_num(&self, _mode: SortType) -> isize {
        todo!()
    }

    fn sort_bool(&self, _mode: SortType) -> bool {
        todo!()
    }

    fn sort_date(&self) -> DateTime<Utc> {
        todo!()
    }
}
