use std::{
    collections::{
        HashMap, 
        HashSet
    }, 
    fs::File, 
    io::Write, 
    path::Path, 
    time::Duration, str::Split
};

use chrono::Utc;
use flaken::Flaken;
use log::*;

use pokerust::{PokemonForm, FromId};
use rand::prelude::SliceRandom;
use rayon::prelude::*;
use regex::Regex;
use serenity::{
    Error, 
    async_trait, 
    builder::CreateEmbed, 
    client::Context, 
    framework::standard::{
        Args, 
        Delimiter
    }, 
    model::channel::Message, 
    utils::MessageBuilder
};

use itertools::Itertools;

use crate::{DefaultDrafts, DoublesTierList, DraftRegistry, EMBED_COLOR, PPATierList, SmogonTierList, Learnsets, data::{Awardable, AwardableKey, Coach, Draft, DraftPokemon, DraftStatus, DraftType, Format, Listable, Match, MatchPokemon, PlayoffMatch, PokemonFormExt, SortType, TierSource, TradePerformance, TradePolicy, NEXT_LETTER}};

#[derive(Debug, PartialEq, Clone)]
pub enum DraftParseError {
    Invalid,
    Missing
}

#[async_trait]
pub trait GenesectContext {
    async fn add_award(&self, draft_id: &String, target: AwardableKey, award: String) -> Result<String, String>;
    async fn add_coach_to_draft(&self, draft_id: &String, coach_id: u64, coach_discord_name: String, division: u8) -> Result<(String, usize, Option<u64>), String>;
    async fn add_coach_trade(&self, draft_id: &String, coach_id: &u64, extra: i8) -> u8;
    async fn add_coach_pick(&self, draft_id: &String, coach_id: &u64, tier: &String, extra: i8) -> u8;
    async fn add_coach_mega(&self, draft_id: &String, coach_id: &u64, extra: i8) -> u8;
    async fn add_draft(&self, draft: Draft) -> Result<(), String>;
    async fn add_mon_to_draft(&self, draft_id: String, mon: DraftPokemon) -> Result<(), String>;
    async fn check_coach_has_mon(&self, draft_id: String, discord_id: &u64, mon: &i16, mega_id: Option<i16>) -> bool;
    async fn check_draft_id<S: Into<String> + Send>(&self, id: S) -> bool;
    async fn check_draft_open(&self, draft_id: &String) -> bool;
    async fn check_draft_joined(&self, draft_id: String, discord_id: &u64) -> bool;
    async fn check_draft_showdown(&self, draft_id: &String, showdown_name: &String) -> (bool, u64);
    async fn check_tier(&self, draft_id: &String, tier: &String) -> bool;
    async fn create_match(&self, draft_id: String, mut content: String) -> Result<Match, String>;
    async fn coach_remaining_picks_embed(&self, draft_id: &String, discord_id: &u64) -> CreateEmbed;
    async fn commit_match(&self, draft_id: &String, the_match: Match, repairing: bool) -> Result<Option<PlayoffMatch>, String>;
    async fn commit_picks(&self, draft_id: &String, tier_picks: HashMap<u8, HashMap<String, u8>>) -> CreateEmbed;
    async fn end_season(&self, draft_id: &String) -> Result<(String, Option<CreateEmbed>), String>;
    async fn filter_draft_id(&self, arg: Option<&str>) -> Result<String, DraftParseError>;
    async fn get_announce_channel_id(&self, draft_id: String) -> u64;
    async fn get_bank(&self, draft_id: String) -> Result<Vec<DraftPokemon>, String>;
    async fn get_coach_logo(&self, draft_id: &String, discord_id: &u64) -> Option<String>;
    async fn get_coach_name(&self, draft_id: &String, discord_id: &u64, server_id: Option<u64>) -> String;
    async fn get_coaches(&self, draft_id: &String, division: Option<u8>) -> Result<Vec<Coach>, String>;
    async fn get_default_draft(&self, server_id: &u64) -> Option<String>;
    async fn get_draft_embed(&self, draft_id: String) -> Result<CreateEmbed, Error>;
    async fn get_draft_logo(&self, draft_id: &String) -> Option<String>;
    async fn get_draft_name(&self, draft_id: String) -> String;
    async fn get_draft_owner_and_moderators(&self, draft_id: &String) -> Result<(u64, Option<u64>), String>;
    async fn get_draft_status(&self, draft_id: &String) -> DraftStatus;
    async fn get_drafts_in_server(&self, server_id: &u64) -> Result<Vec<Draft>, String>;
    async fn get_drafts(&self) -> Result<Vec<Draft>, String>;
    async fn get_last_round_picks(&self, draft_id: &String, server_id: Option<u64>) -> Option<CreateEmbed>;
    async fn get_matches(&self, draft_id: String, discord_id: Option<u64>) -> Result<Vec<Match>, String>;
    async fn get_mod_role(&self, draft_id: String) -> Option<u64>;
    async fn get_player_role(&self, draft_id: &String) -> Option<u64>;
    async fn get_mon_embed(&self, draft_id: String, mon: i16, server_id: Option<u64>) -> Result<CreateEmbed, Error>;
    async fn get_mon_info(&self, draft_id: &String, mon: &i16) -> String;
    async fn get_next_match_id(&self, draft_id: &String) -> usize;
    async fn get_picks(&self, draft_id: &String) -> HashMap<u8, HashMap<String, u8>>;
    async fn get_playoffs(&self, draft_id: String, discord_id: Option<u64>) -> Result<Vec<PlayoffMatch>, String>;
    async fn get_season_embed(&self, draft_id: &String, season: u16) -> Result<CreateEmbed, String>;
    async fn get_server_id(&self, draft_id: String) -> u64;
    async fn get_team(&self, draft_id: String, discord_id: &u64) -> Result<Vec<DraftPokemon>, String>;
    async fn get_tiers(&self, format: &Format, source: &TierSource) -> Vec<String>;
    async fn get_moves(&self, version_group: u8) -> HashMap<i16, Vec<i16>>;
    async fn is_owner(&self, draft_id: String, discord_id: &u64) -> bool;
    async fn is_porygon_channel(&self, server_id: &u64, channel_id: &u64) -> (bool, String);
    async fn make_picks_embed(&self, draft_id: &String, tier_picks: &HashMap<u8, HashMap<String, u8>>) -> CreateEmbed;
    async fn match_embed_from_id(&self, draft_id: &String, match_id: u16, server_id: Option<u64>) -> Result<CreateEmbed, String>;
    async fn match_embed(&self, draft_id: &String, the_match: &Match, server_id: Option<u64>) -> CreateEmbed;
    async fn modify_draft(&self, msg: &Message, draft_id: String, mut args: Args) -> Result<(), Error>;
    async fn new_season(&self, draft_id: &String) -> Result<String, String>;
    async fn pester_coach(&self, draft_name: String, draft_id: String, discord_id: u64, round: u16) -> Result<(), String>;
    async fn pick_mon(&self, draft_id: String, discord_id: &u64, mon: i16, moderator: bool, server_id: Option<u64>, mega_id: Option<i16>) -> Result<(String, String, CreateEmbed, u16, bool), String>;
    async fn playoff_match_embed(&self, pmatch: &PlayoffMatch, draft_id: &String, server_id: Option<u64>) -> CreateEmbed;
    async fn save_default_drafts(&self);
    async fn set_coach_value(&self, draft_id: String, discord_id: &u64, key: String, value: String) -> Result<String, String>;
    async fn set_coach_dropout(&self, draft_id: &String, coach_id: &u64) -> (bool, String);
    async fn set_default_draft(&self, server_id: &u64, draft_id: &String) -> Option<String>;
    async fn start_draft(&self, draft_id: String) -> Result<String, String>;
    async fn start_playoffs(&self, draft_id: &String, days: Option<u8>) -> Result<String, String>;
    async fn team_embed(&self, draft_id: &String, coach: &u64, server_id: Option<u64>) -> Result<(CreateEmbed, Vec<i16>), String>;
    async fn teams_embed(&self, draft_id: &String, division: Option<u8>, server_id: Option<u64>) -> Result<CreateEmbed, String>;
    async fn trade(&self, draft_id: String, discord_id: &u64, o_mon: i16, n_mon: i16, moderator: bool, target_user: Option<u64>, server_id: Option<u64>) -> Result<(String, String, String, CreateEmbed, u16, TradePerformance), String>;
}

#[async_trait]
impl GenesectContext for Context {

    async fn check_draft_id<S: Into<String> + Send>(&self, id: S) -> bool {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().unwrap();
        reg.contains_key(&id.into())
    }

    async fn check_draft_open(&self, draft_id: &String) -> bool {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().unwrap();
        !reg[draft_id].closed
    }

    async fn add_draft(&self, draft: Draft) -> Result<(), String> {
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("Bot didn't start right");
        if !reg.contains_key(&draft.draft_id) {
            draft.save();
            reg.insert(draft.draft_id.clone(), draft);
            Ok(())
        }
        else {
            Err("Draft ID already exists.".to_owned())
        }
    }

    async fn add_mon_to_draft(&self, draft_id: String, mon: DraftPokemon) -> Result<(), String> {
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("Bot didn't start right");
        if reg.contains_key(&draft_id) {
            let mut result: Result<(), String> = Err("Something has gone terribly wrong.".to_owned());
            reg.entry(draft_id).and_modify(|draft| {
                if !draft.mons.contains_key(&mon.id) {
                    draft.mons.insert(mon.id, mon);
                    draft.save();
        
                    result = Ok(());
                }
                else if draft.mons[&mon.id].tier.as_str() == "Banned" {
                    draft.mons.entry(mon.id).and_modify(|m| {
                        m.tier = mon.tier.clone();
                    });
                    draft.save();

                    result = Ok(());
                }
                else {
                    result = Err("Pokémon already exists in draft bank.".to_owned());
                }
            });
            result
        }
        else {
            Err("Draft ID does not exist.".to_owned())
        }
    }

    /*async fn add_coach_to_draft(&self, draft_id: String, coach: Coach) -> Result<(String, usize, Option<u64>), String> {
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("Bot didn't start right");
        if reg.contains_key(&draft_id) {
            let mut result: Result<(String, usize, Option<u64>), String> = Err("Something has gone terribly wrong.".to_owned());
            reg.entry(draft_id).and_modify(|draft| {
                if !draft.coaches.contains_key(&coach.discord_id) {
                    draft.draft_order.push(coach.discord_id);
                    result = match draft.add_coach(coach) {
                        Ok(_) => {
                            Ok((draft.draft_name.clone(), draft.draft_order.len(), draft.player_role_id.clone()))
                        },
                        Err(why) => {
                            Err(why)
                        }
                    };
                    draft.save();
                }
                else {
                    result = Err("Coach already exists in draft.".to_owned());
                }
            });
            result
        }
        else {
            Err("Draft ID does not exist.".to_owned())
        }
    }*/

    async fn add_coach_to_draft(&self, draft_id: &String, coach_id: u64, coach_discord_name: String, division: u8) -> Result<(String, usize, Option<u64>), String> {
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("Bot didn't start right");
        if reg.contains_key(draft_id) {
            let (season_has_coach, season_with_coach) = {
                let draft = &reg[draft_id];
                if draft.past_seasons.is_some() {
                    let ps = draft.past_seasons.as_ref().unwrap();
                    let mut i = draft.league_season - 1;
                    let out = loop {
                        if ps.contains_key(&i) && ps[&i].coaches.contains_key(&coach_id) {
                            break true;
                        }
                        else if i == 0 {
                            break false;
                        }
                        else {
                            i -= 1;
                            continue;
                        }
                    };

                    (out, i)
                }
                else {
                    (false, 0)
                }
            };
            if season_has_coach {
                let mut coach = {
                    let draft = &reg[draft_id];
                    draft.past_seasons.as_ref().unwrap()[&season_with_coach].coaches[&coach_id].clone()
                };

                if division != 0 {
                    coach.division = division;
                }
                coach.seed = 0;
                coach.mons.clear();
                coach.wins = 0;
                coach.losses = 0;
                coach.differential = 0;
                coach.used_trades = 0;
                coach.extra_trades = 0;
                coach.playoff_differential = 0;
                coach.challonge_participant = None;
                coach.dropout = false;

                let mut result: Result<(String, usize, Option<u64>), String> = Err("Something has gone terribly wrong.".to_owned());
                reg.entry(draft_id.into()).and_modify(|draft| {
                    if !draft.coaches.contains_key(&coach.discord_id) {
                        draft.draft_order.push(coach.discord_id);
                        result = match draft.add_coach(coach) {
                            Ok(_) => {
                                Ok((draft.draft_name.clone(), draft.draft_order.len(), draft.player_role_id.clone()))
                            },
                            Err(why) => {
                                Err(why)
                            }
                        };
                        draft.save();
                    }
                    else {
                        result = Err("Coach already exists in draft.".to_owned());
                    }
                });
                result
            }
            else {
                let coach = Coach {
                    discord_id: coach_id,
                    discord_name: coach_discord_name,
                    showdown_name: None,
                    name: None,
                    logo: None,
                    division,
                    seed: 0,
                    mons: Vec::new(),
                    wins: 0,
                    losses: 0,
                    differential: 0,
                    awards: HashMap::new(),
                    next_pick: None,
                    used_trades: 0,
                    extra_trades: 0,
                    playoff_differential: 0,
                    challonge_participant: None,
                    dropout: false,
                    extra_picks: HashMap::new(),
                    megas: HashSet::new(),
                    megas_allowed: 0,
                };
                let mut result: Result<(String, usize, Option<u64>), String> = Err("Something has gone terribly wrong.".to_owned());
                reg.entry(draft_id.into()).and_modify(|draft| {
                    if !draft.coaches.contains_key(&coach.discord_id) {
                        draft.draft_order.push(coach.discord_id);
                        result = match draft.add_coach(coach) {
                            Ok(_) => {
                                Ok((draft.draft_name.clone(), draft.draft_order.len(), draft.player_role_id.clone()))
                            },
                            Err(why) => {
                                Err(why)
                            }
                        };
                        draft.save();
                    }
                    else {
                        result = Err("Coach already exists in draft.".to_owned());
                    }
                });
                result
            }
        }
        else {
            Err("Draft ID does not exist.".to_owned())
        }
    }

    async fn get_tiers(&self, format: &Format, source: &TierSource) -> Vec<String> {
        let data = self.data.read().await;
        let reg = match (format, source) {
            (Format::Singles, TierSource::Smogon) => {
                data.get::<SmogonTierList>().expect("bot bork").clone()
            },
            (Format::Doubles, TierSource::Smogon) => {
                data.get::<DoublesTierList>().expect("bot bork").clone()
            },
            (_, TierSource::PPA) => {
                data.get::<PPATierList>().expect("bot borked").iter().map(|(x, (y, _))| (x.to_owned(), y.to_owned())).collect()
            }
            _ => {
                data.get::<SmogonTierList>().expect("bot bork").clone()
            }
        };
        let t: Vec<String> = reg.iter().map(|(_, y)| { y.to_owned() }).unique().collect();

        t
    }

    async fn get_moves(&self, version_group: u8) -> HashMap<i16, Vec<i16>> {
        let data = self.data.read().await;
        let reg = data.get::<Learnsets>().expect("bot breoken");
        reg[&version_group].clone()
    }

    async fn check_tier(&self, draft_id: &String, tier: &String) -> bool {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        let tier_mons = draft.mons.iter().filter(|(_, m)| &m.tier == tier);
        tier_mons.count() > 0 && draft.tiers_allowed.contains(tier)
    }

    async fn make_picks_embed(&self, draft_id: &String, tier_picks: &HashMap<u8, HashMap<String, u8>>) -> CreateEmbed {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        let dname = draft.draft_name.clone();
        let team_size = draft.total_picks;
        let mut fields: Vec<(String, String)> = Vec::new();

        for (div, tp) in tier_picks {
            let k = if *div == 0 {
                format!("Division {} (Default)", div)
            }
            else {
                format!("Division {}", div)
            };
            let mut v = MessageBuilder::new();

            let mut tp: Vec<(String, u8)> = tp.iter().map(|(t, p)| (t.to_owned(), p.to_owned())).collect();
            tp.sort_by(|(a, _), (b, _)| a.cmp(b));
            
            let mut total = 0;
            for (t, p) in tp {
                v.push_bold(format!("Tier {}: ", t));
                v.push_line(format!("{} Picks", p));
                total += p;
            }
            if total > team_size {
                v.push_bold("Total Picks:");
                v.push_line(format!(" {}", total));
                v.push_bold_line("Higher than team size! This won't block drafting, but make sure this is correct.");
            }
            else if total == team_size {
                v.push_bold("Total Picks:");
                v.push_line(format!(" {}", total));
            }
            else if total < team_size {
                v.push_bold("Total Picks:");
                v.push_line(format!(" {}", total));
                v.push_bold_line("Less than team size! This will block drafting!");
            }

            let v = v.build();

            fields.push((k, v));
        }
        fields.sort_by(|(a, _), (b, _)| a.cmp(b));

        let e = {
            let mut e = CreateEmbed::default();

            e.title(format!("Picks per Tier/Division for {}", dname));
            
            for (n, v) in fields {
                e.field(n, v, true);
            }
            e.footer(|f| f.text(format!("Team Size: {}", team_size)));
            
            e.color(EMBED_COLOR);

            e
        };
        e
    }

    async fn coach_remaining_picks_embed(&self, draft_id: &String, discord_id: &u64) -> CreateEmbed {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        let dname = draft.draft_name.clone();
        let team_size = draft.total_picks;
        let coach = &draft.coaches[discord_id];
        let tier_picks = draft.tier_picks.clone();
        let mut fields: Vec<(String, String)> = Vec::new();

        for (div, tp) in tier_picks.clone() {
            if div == coach.division || tier_picks.len() == 1 {
                let k = if div == 0 {
                    format!("Division {} (Default)", div)
                }
                else {
                    format!("Division {}", div)
                };
                let mut v = MessageBuilder::new();
                
                //let mut tp: Vec<(String, u8)> = tp.iter().map(|(t, p)| (t.to_owned(), p.to_owned())).collect();
                //tp.sort_by(|(a, _), (b, _)| a.cmp(b));
                //let tp: HashMap<String, u8> = tp.iter().map(|(t, p)| (t.to_owned(), p.to_owned())).collect();
                
                /*for (t, p) in tp {
                    let p_remaining = coach.get_remaining_picks(draft, &t);
                    v.push_bold(format!("Tier {}: ", t));
                    v.push_line(format!("{} Picks Remaining (out of {})", p_remaining, p));
                }*/
                let mut t_a = draft.tiers_allowed.clone();
                t_a.sort_by(|a, b| a.cmp(b));

                for t in t_a {
                    if tp.contains_key(&t) {
                        let p = tp[&t];
                        let p_remaining = coach.get_remaining_picks(draft, &t);
                        v.push_bold(format!("Tier {}: ", t));
                        v.push_line(format!("{} Picks Remaining (out of {})", p_remaining, p));
                    }
                    else if tier_picks.contains_key(&0) {
                        let tp = &tier_picks[&0];
                        if tp.contains_key(&t) {
                            let p = tp[&t];
                            let p_remaining = coach.get_remaining_picks(draft, &t);
                            v.push_bold(format!("Tier {}: ", t));
                            v.push_line(format!("{} Picks Remaining (out of {})", p_remaining, p));
                        }
                        else {
                            let p = draft.total_picks;
                            let p_remaining = coach.get_remaining_picks(draft, &t);
                            v.push_bold(format!("Tier {}: ", t));
                            v.push_line(format!("{} Picks Remaining (out of {})", p_remaining, p));
                        }
                    }
                    else {
                        let p = draft.total_picks;
                        let p_remaining = coach.get_remaining_picks(draft, &t);
                        v.push_bold(format!("Tier {}: ", t));
                        v.push_line(format!("{} Picks Remaining (out of {})", p_remaining, p));
                    }
                }
                v.push_bold("Total Remaining Picks:");
                v.push_line(format!(" {}", draft.total_picks as usize - coach.mons.len()));
        
                let v = v.build();
        
                fields.push((k, v));
            }
        }
        fields.sort_by(|(a, _), (b, _)| a.cmp(b));
        if fields.len() == 0 {
            let k = "Total Remaining Picks".to_owned();
            let v = format!("{}", draft.total_picks as usize - coach.mons.len());
            fields.push((k, v));
        }

        let e = {
            let mut e = CreateEmbed::default();

            e.title(format!("Your Remaining Picks in {}", dname));
            
            for (n, v) in fields {
                e.field(n, v, true);
            }
            e.footer(|f| f.text(format!("Team Size: {}/{}", coach.mons.len(), team_size)));
            
            e.color(EMBED_COLOR);

            e
        };
        e
    }

    async fn commit_picks(&self, draft_id: &String, tier_picks: HashMap<u8, HashMap<String, u8>>) -> CreateEmbed {
        info!("Committing Tier-Picks to draft {}", draft_id);
        let embed = self.make_picks_embed(draft_id, &tier_picks).await;
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("Bot didn't start right");
        debug!("Tier-Picks: Entering registry");
        reg.entry(draft_id.to_owned()).and_modify(|draft| {
            draft.tier_picks = tier_picks;
            draft.save();
        });

        debug!("Tier-Picks: returning embed");
        embed
    }

    async fn get_draft_name(&self, draft_id: String) -> String {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[&draft_id];
        draft.draft_name.clone()
    }

    async fn get_draft_status(&self, draft_id: &String) -> DraftStatus {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        draft.status.clone()
    }

    async fn get_coach_name(&self, draft_id: &String, discord_id: &u64, server_id: Option<u64>) -> String {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        let c = &draft.coaches[discord_id];
        c.display_name(self, server_id).await
    }

    async fn get_coach_logo(&self, draft_id: &String, discord_id: &u64) -> Option<String> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        let c = &draft.coaches[discord_id];
        c.logo.clone()
    }

    async fn get_draft_logo(&self, draft_id: &String) -> Option<String> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        draft.logo.clone()
    }

    async fn get_mon_info(&self, draft_id: &String, mon: &i16) -> String {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        let m = &draft.mons[mon];
        format!("**{}**: {}", m.field_name(self).await, m.field_value(self).await)
    }

    async fn get_draft_embed(&self, draft_id: String) -> Result<CreateEmbed, Error> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[&draft_id];
        draft.embed(self).await
    }

    async fn get_bank(&self, draft_id: String) -> Result<Vec<DraftPokemon>, String> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[&draft_id];
        if draft.mons.len() > 0 {
            let mut bank: Vec<DraftPokemon> = draft.mons.iter().filter(|(_, m)| {
                !m.drafted && draft.tiers_allowed.contains(&m.tier)
            }).map(|(_, m)| m.to_owned()).collect();
            if bank.len() > 0 {
                bank.sort_by(|a, b| a.type1.cmp(&b.type1));
                /*let mut csv = String::from("pokemon,tier\n");
                for m in &bank {
                    csv.push_str(format!("{},{}\n", &m.name_id, &m.tier).as_str());
                }
                let mut file = match File::create(Path::new("test.csv")) {
                    Err(why) => unreachable!(),
                    Ok(file) => file,
                };
        
                match file.write_all(csv.as_bytes()) {
                    Err(why) => unreachable!(),
                    Ok(_) => debug!("successfully wrote to test.csv"),
                }*/
                Ok(bank)
            }
            else {
                Err("All available Pokémon have been drafted!".to_owned())
            }
        }
        else {
            Err("The bank is empty!".to_owned())
        }
    }

    async fn get_team(&self, draft_id: String, discord_id: &u64) -> Result<Vec<DraftPokemon>, String> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[&draft_id];
        if draft.mons.len() > 0 {
            let mut bank: Vec<DraftPokemon> = draft.mons.iter().filter(|(_, m)| {
                m.drafted && m.coach.unwrap() == discord_id.to_owned()
            }).map(|(_, m)| m.to_owned()).collect();
            if bank.len() > 0 {
                bank.sort_by(|a, b| a.type1.cmp(&b.type1));
                Ok(bank)
            }
            else {
                Err("You have not picked any Pokémon yet!".to_owned())
            }
        }
        else {
            Err("The draft is empty!".to_owned())
        }
    }

    async fn get_drafts(&self) -> Result<Vec<Draft>, String> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let mut drafts: Vec<Draft> = reg.iter().map(|(_, m)| m.to_owned()).collect();
        if drafts.len() > 0 {
            drafts.sort_by(|a, b| a.draft_id.cmp(&b.draft_id));
            Ok(drafts)
        }
        else {
            Err("There are no drafts right now!".to_owned())
        }
    }

    async fn get_drafts_in_server(&self, server_id: &u64) -> Result<Vec<Draft>, String> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let mut drafts: Vec<Draft> = reg.iter().filter(|(_, d)| {
            &d.server_id == server_id
        }).map(|(_, m)| m.to_owned()).collect();
        if drafts.len() > 0 {
            drafts.sort_by(|a, b| a.draft_id.cmp(&b.draft_id));
            Ok(drafts)
        }
        else {
            Err("There are no drafts right now!".to_owned())
        }
    }

    async fn get_matches(&self, draft_id: String, discord_id: Option<u64>) -> Result<Vec<Match>, String> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[&draft_id];
        if draft.matches.len() > 0 {
            if discord_id.is_some() {
                let did = discord_id.unwrap();
                let mut f: Vec<Match> = draft.matches.iter().filter(|m| m.coach1 == did || m.coach2 == did).map(|m| m.to_owned()).collect();
                f.sort_by(|a, b| a.id.cmp(&b.id));
                Ok(f)
            }
            else {
                Ok(draft.matches.clone())
            }
        }
        else {
            Err("The draft hasn't played any matches yet!".to_owned())
        }
    }

    async fn get_playoffs(&self, draft_id: String, discord_id: Option<u64>) -> Result<Vec<PlayoffMatch>, String> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[&draft_id];
        if draft.playoff_matches.len() > 0 {
            let pma = draft.playoff_matches.iter().sorted_by(|a, b| {
                let al = &a.letter;
                let bl = &b.letter;
                al.cmp(&bl)
            });
            if discord_id.is_some() {
                let did = discord_id.unwrap();
                let f: Vec<PlayoffMatch> = pma.filter(|m| m.coach1 == did || m.coach2 == did).map(|m| m.to_owned()).collect();
                Ok(f)
            }
            else {
                Ok(pma.map(|m| m.to_owned()).collect())
            }
        }
        else {
            Err("The draft hasn't played any playoff matches yet!".to_owned())
        }
    }

    async fn get_coaches(&self, draft_id: &String, division: Option<u8>) -> Result<Vec<Coach>, String> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        if draft.coaches.len() > 0 {
            let c: Vec<Coach> = draft.coaches.clone().iter().map(|(_, c)| c.to_owned()).collect();
            let mut c = if division.is_some() {
                let d: Vec<Coach> = c.iter().filter(|c| c.division == division.unwrap()).map(|c| c.to_owned()).collect();
                d
            }
            else {
                c
            };
            c.sort_by(|a, b| b.differential.cmp(&a.differential));
            c.sort_by(|a, b| a.division.cmp(&b.division));
            Ok(c)
        }
        else {
            Err("The draft doesn't have any coaches yet!".to_owned())
        }
    }

    async fn get_picks(&self, draft_id: &String) -> HashMap<u8, HashMap<String, u8>> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        draft.tier_picks.clone()
    }

    async fn teams_embed(&self, draft_id: &String, division: Option<u8>, server_id: Option<u64>) -> Result<CreateEmbed, String> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        if draft.coaches.len() > 0 {
            let c: Vec<Coach> = draft.coaches.clone().iter().map(|(_, c)| c.to_owned()).collect();
            let c = if division.is_some() {
                let d: Vec<Coach> = c.iter().filter(|c| c.division == division.unwrap()).map(|c| c.to_owned()).collect();
                d
            }
            else {
                c
            };
            let mut teams: Vec<(String, String)> = Vec::new();
            let len = c.len();
            for co in c {
                let mut n = co.display_name(self, server_id).await;
                n.push(' ');
                let record = format!("({}-{} {:+})", co.wins, co.losses, co.differential);
                n.push_str(record.as_str());
                let mut v = String::new();
                if co.mons.len() > 0 {
                    let mons = co.get_mons(draft);
                    for m in mons {
                        let name = if len < 6 {
                            m.field_name(self).await
                        }
                        else {
                            m.name_with_text_type()
                        };
                        v.push_str(name.as_str());
                        v.push('\n');
                    }
                }
                else {
                    v.push_str("Empty");
                }
                teams.push((n, v));
            }
            let e = {
                let mut e = CreateEmbed::default();
                
                for (n, v) in teams {
                    e.field(n, v, true);
                }
                
                e.color(EMBED_COLOR);
        
                e
            };
            Ok(e)
        }
        else {
            Err("The draft doesn't have any coaches yet!".to_owned())
        }
    }

    async fn team_embed(&self, draft_id: &String, coach: &u64, server_id: Option<u64>) -> Result<(CreateEmbed, Vec<i16>), String> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        if draft.coaches.len() > 0 {
            let co = &draft.coaches[coach];

            let mut fields: Vec<(String, String)> = Vec::new();
            let mut ids = Vec::new();
            if co.mons.len() > 0 {
                let mons = co.get_mons(draft);
                let mut index = 1;
                for m in mons {
                    let n = format!("{}: {}", index, m.field_name(self).await);
                    let v = m.field_value(self).await;
                    fields.push((n, v));
                    ids.push(m.id);
                    index += 1;
                }
            }

            let e = {
                let mut e = CreateEmbed::default();

                let mut n = co.display_name(self, server_id).await;
                n.push(' ');
                let record = format!("({}-{} {:+})", co.wins, co.losses, co.differential);
                n.push_str(record.as_str());
                
                e.title(n);

                for (n, v) in fields {
                    e.field(n, v, false);
                }

                e.color(EMBED_COLOR);
        
                e
            };
            Ok((e, ids))
        }
        else {
            Err("The draft doesn't have any coaches yet!".to_owned())
        }
    }

    async fn is_porygon_channel(&self, server_id: &u64, channel_id: &u64) -> (bool, String) {
        match self.get_drafts_in_server(server_id).await {
            Ok(d) => {
                let mut pd = d.iter().filter(|d| d.porygon_channel.is_some() && d.porygon_channel.unwrap() == *channel_id);
                let one = pd.clone().count() == 1;
                let two = if one {
                    pd.next().unwrap().draft_id.clone()
                }
                else {
                    "".to_owned()
                };
                (one, two)
            },
            Err(_) => (false, "".to_owned()),
        }
    }

    async fn check_draft_joined(&self, draft_id: String, discord_id: &u64) -> bool {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[&draft_id];
        draft.coaches.contains_key(discord_id)
    }

    async fn check_draft_showdown(&self, draft_id: &String, showdown_name: &String) -> (bool, u64) {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        let mut c = draft.coaches.iter().filter(|(_, c)| !c.dropout && c.showdown_name.is_some() && &c.showdown_name.clone().unwrap() == showdown_name);
        let exists = c.clone().count() == 1;
        let id = if exists {
            c.next().unwrap().1.discord_id
        }
        else { 0 };
        (exists, id)
    }

    async fn check_coach_has_mon(&self, draft_id: String, discord_id: &u64, mon: &i16, mega_id: Option<i16>) -> bool {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[&draft_id];
        if draft.coaches.contains_key(discord_id) {
            let c = &draft.coaches[discord_id];

            if mega_id.is_some() {
                let mega = mega_id.unwrap();
                c.megas.contains(&mega) && c.mons.contains(mon)
            }
            else { c.mons.contains(mon) }
        }
        else { false }
    }

    async fn get_next_match_id(&self, draft_id: &String) -> usize {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        draft.matches.len() + 1
    }

    async fn commit_match(&self, draft_id: &String, the_match: Match, repairing: bool) -> Result<Option<PlayoffMatch>, String> {
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("Bot didn't start right");
        let mut res = Err("something went wrong".into());
        let mut the_match = the_match.clone();
        reg.entry(draft_id.to_owned()).and_modify(|draft| {
            if vec![DraftStatus::Playing, DraftStatus::SeasonPlayoffs].contains(&draft.status) {
                for m in the_match.team1.clone() {
                    draft.mons.entry(m.id).and_modify(|p| {
                        p.active_kills += m.active_kills as u16;
                        p.passive_kills += m.passive_kills as u16;
                        p.deaths += m.deaths as u16;
                        p.lifetime_active += m.active_kills as u16;
                        p.lifetime_passive += m.passive_kills as u16;
                        p.lifetime_deaths += m.deaths as u16;
                    });
                }
                for m in the_match.team2.clone() {
                    draft.mons.entry(m.id).and_modify(|p| {
                        p.active_kills += m.active_kills as u16;
                        p.passive_kills += m.passive_kills as u16;
                        p.deaths += m.deaths as u16;
                        p.lifetime_active += m.active_kills as u16;
                        p.lifetime_passive += m.passive_kills as u16;
                        p.lifetime_deaths += m.deaths as u16;
                    });
                }
                let pmatch = if draft.status == DraftStatus::SeasonPlayoffs {
                    let mut pmatches = draft.playoff_matches.iter_mut().filter(|p| {
                        !p.complete && (
                            (p.coach1 == the_match.coach1 && p.coach2 == the_match.coach2) ||
                            (p.coach1 == the_match.coach2 && p.coach2 == the_match.coach1)
                        )
                    }).collect::<Vec<&mut PlayoffMatch>>();
                    let exist = pmatches.len();
                    if exist == 1 {
                        let pm = pmatches.pop().unwrap();
                        let r = pm.add_match(the_match.clone()).map_err(|(y, _)| y).err();
                        if r.is_none() {
                            if pm.score1 >= draft.playoff_best_of * 2 || pm.score2 >= draft.playoff_best_of * 2 {
                                let r = pm.finalize_match().map_err(|(y, _)| y).err();
                                if r.is_some() {
                                    res = Err(r.unwrap());
                                    return;
                                }
                            }
                            Some(pm.clone())
                        }
                        else {
                            res = Err(r.unwrap());
                            return;
                        }
                    }
                    else if exist == 0 {
                        let c_seeded = draft.coaches.iter().filter(|(_, c)| (draft.team_divisions || (c.seed != 0 && c.challonge_participant.is_some())) && (draft.team_divisions != (c.seed != 0 && c.challonge_participant.is_some())));
                        if repairing && c_seeded.count() == 0 {
                            let _ = draft.fix_participants();
                        }
                        let (cid, l) = if draft.team_divisions {
                            let l = draft.next_playoff_letter.clone();
                            draft.next_playoff_letter = NEXT_LETTER[draft.next_playoff_letter.as_str()].to_string();
                            let mut flake = Flaken::default().node(1).epoch(1613940319944);
                            let i = flake.next();

                            (i, l)
                        }
                        else {
                            let (c1c, c2c) = (&draft.coaches[&the_match.coach1].challonge_participant.unwrap(), &draft.coaches[&the_match.coach2].challonge_participant.unwrap());
                            let (cid, l) = match draft.get_match_id(*c1c, *c2c, repairing) {
                                Ok(i) => i,
                                Err(y) => {
                                    res = Err(y);
                                    return;
                                }
                            };
                            (cid, l)
                        };
                        let mut pm = PlayoffMatch::new_from_match(cid, l, the_match.clone());
                        if pm.score1 >= draft.playoff_best_of * 2 || pm.score2 >= draft.playoff_best_of * 2 {
                            let r = pm.finalize_match().map_err(|(y, _)| y).err();
                            if r.is_some() {
                                res = Err(r.unwrap());
                                return;
                            }
                        }
                        draft.playoff_matches.push(pm.clone());
                        Some(pm)
                    }
                    else {
                        None
                    }
                } else { None };
                if pmatch.is_some() {
                    let pmatch = pmatch.as_ref().unwrap();
                    if !repairing && !draft.team_divisions {
                        if let Err(y) = draft.update_playoff_match(pmatch) {
                            res = Err(y);
                            return;
                        }
                    }
                    the_match.playoffs = true;
                }
                draft.coaches.entry(the_match.coach1).and_modify(|c| {
                    match (the_match.winner_is_1, pmatch.is_some()) {
                        (true, false) => {
                            c.wins += 1;
                            c.differential += the_match.absolute_differential() as i16;
                        },
                        (false, false) => {
                            c.losses += 1;
                            c.differential -= the_match.absolute_differential() as i16;
                        },
                        (true, true) => {
                            c.playoff_differential += the_match.absolute_differential() as i16;
                        },
                        (false, true) => {
                            c.playoff_differential -= the_match.absolute_differential() as i16;
                        }
                    }
                });
                draft.coaches.entry(the_match.coach2).and_modify(|c| {
                    match (!the_match.winner_is_1, pmatch.is_some()) {
                        (true, false) => {
                            c.wins += 1;
                            c.differential += the_match.absolute_differential() as i16;
                        },
                        (false, false) => {
                            c.losses += 1;
                            c.differential -= the_match.absolute_differential() as i16;
                        },
                        (true, true) => {
                            c.playoff_differential += the_match.absolute_differential() as i16;
                        },
                        (false, true) => {
                            c.playoff_differential -= the_match.absolute_differential() as i16;
                        }
                    }
                });
                draft.matches.push(the_match.clone());
                draft.save();
                res = Ok(pmatch);
            }
        });
        res
    }

    async fn match_embed(&self, draft_id: &String, the_match: &Match, server_id: Option<u64>) -> CreateEmbed {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        let embed = the_match.embed(self, draft, server_id).await;
        embed
    }

    async fn match_embed_from_id(&self, draft_id: &String, match_id: u16, server_id: Option<u64>) -> Result<CreateEmbed, String> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        if draft.matches.len() > 0 {
            let mut filt = draft.matches.iter().filter(|m| m.id == match_id);
            if filt.clone().count() == 1 {
                let m = filt.next().unwrap();
                Ok(m.embed(self, draft, server_id).await)
            }
            else {
                Err("That match hasn't been played or been submitted yet!".to_owned())
            }
        }
        else {
            Err("This draft hasn't played any matches yet!".to_owned())
        }
    }

    async fn playoff_match_embed(&self, pmatch: &PlayoffMatch, draft_id: &String, server_id: Option<u64>) -> CreateEmbed {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        let embed = pmatch.embed(self, draft, server_id).await;
        embed
    }

    async fn is_owner(&self, draft_id: String, discord_id: &u64) -> bool {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[&draft_id];
        &draft.owner_id == discord_id
    }

    async fn get_mod_role(&self, draft_id: String) -> Option<u64> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[&draft_id];
        draft.moderator_role_id.clone()
    }
    
    async fn get_player_role(&self, draft_id: &String) -> Option<u64> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        draft.player_role_id.clone()
    }

    async fn get_server_id(&self, draft_id: String) -> u64 {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[&draft_id];
        draft.server_id.clone()
    }

    async fn get_announce_channel_id(&self, draft_id: String) -> u64 {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[&draft_id];
        draft.announcements_channel.unwrap().clone()
    }

    async fn start_draft(&self, draft_id: String) -> Result<String, String> {
        if self.check_draft_id(&draft_id).await {
            let mut result = {
                let mut data = self.data.write().await;
                let reg = data.get_mut::<DraftRegistry>().expect("Bot didn't start right");
                
                let mut result: Result<String, String> = Err("Something has gone terribly wrong.".to_owned());
                reg.entry(draft_id.clone()).and_modify(|draft| {
                    if draft.status == DraftStatus::Ready {
                        if draft.mons.len() >= (draft.total_picks as usize * draft.coaches.len()) {
                            draft.draft_round = draft.draft_order.clone();

                            draft.status = DraftStatus::Drafting;
                            draft.save();

                            result = Ok(draft.draft_name.clone());
                        }
                        else {
                            result = Err("There are not enough Pokémon in the bank for all coaches to complete the draft!".to_owned());
                        }
                    }
                    else {
                        result = Err("The draft must be in the Ready state! After draft creation, at least 2 coaches must have joined the draft.".to_owned());
                    }
                });
                result
            };
            let data = self.data.read().await;
            let reg = data.get::<DraftRegistry>().expect("bot bork");
            let draft = &reg[&draft_id];
            result = match result {
                Ok(_) => {
                    match self.pester_coach(draft.draft_name.clone(), draft.draft_id.clone(), draft.draft_round.first().unwrap().clone(), 1).await {
                        Ok(_) => Ok(draft.draft_name.clone()),
                        Err(why) => Err(why)
                    }
                },
                Err(why) => Err(why)
            };

            result
        }
        else {
            Err("Draft ID does not exist.".to_owned())
        }
    }

    async fn pester_coach(&self, draft_name: String, draft_id: String, discord_id: u64, round: u16) -> Result<(), String> {
        match self.http.get_user(discord_id).await {
            Ok(user) => {
                let lrp = self.get_last_round_picks(&draft_id, None).await;
                match user.dm(self, |m| {
                    m.content(format!("**Round {}:** It's your turn to pick in {} (Draft ID: {})!\nUse the `g~bank {}` command either here or in a mutual server to see available Pokémon and the `g~pick {} <pokemon>` command to make your pick.", round, draft_name, draft_id, draft_id, draft_id));
                    if lrp.is_some() {
                        m.set_embed(lrp.unwrap());
                    }
                    m
                }).await {
                    Ok(_) => Ok(()),
                    Err(why) => Err(why.to_string())
                }
            },
            Err(why) => {
                Err(why.to_string())
            }
        }
    }

    async fn get_last_round_picks(&self, draft_id: &String, server_id: Option<u64>) -> Option<CreateEmbed> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        if draft.last_round_drafted.len() > 0 {
            let picks = draft.last_round_drafted.iter().map(|(u, m)| {
                (
                    self.get_coach_name(draft_id, u, server_id),
                    draft.mons[m].field_name(self)
                )
            });
            let mut embed = CreateEmbed::default();
            embed.title("Last Round's Picks");
            embed.description("Other coaches made these picks since the last time you submitted a pick:");
            let mut i = 1;
            for (u, m) in picks {
                let (u, m) = (u.await, m.await);
                embed.field(format!("{}: {}", i, m), format!("Picked by **{}**", u), false);
                i += 1;
            }
            embed.color(EMBED_COLOR);

            Some(embed)
        }
        else {
            None
        }
    }

    async fn pick_mon(&self, draft_id: String, discord_id: &u64, mon: i16, moderator: bool, server_id: Option<u64>, mega_id: Option<i16>) -> Result<(String, String, CreateEmbed, u16, bool), String> {
        if self.check_draft_id(&draft_id).await {
            let result = {
                let mut data = self.data.write().await;
                let reg = data.get_mut::<DraftRegistry>().expect("Bot didn't start right");
                let mut result: Result<(String, String, u16, bool), String> = Err("Something has gone terribly wrong.".to_owned());
                reg.entry(draft_id.clone()).and_modify(|draft| {
                    if draft.status == DraftStatus::Drafting {
                        if draft.draft_round.first().unwrap() == discord_id {
                            if draft.mons.contains_key(&mon) {
                                let m = &draft.mons[&mon];
                                let coach_id = Some(discord_id.clone());
                                let drafted_same_sp = draft.mons.iter().filter(|(_, p)| {
                                    p.coach == coach_id && p.spid == m.spid
                                });
                                if draft.tiers_allowed.contains(&m.tier) || (draft.tiers_allowed.is_empty() && !vec!["Illegal", "Banned"].contains(&m.tier.as_str())) {
                                    if drafted_same_sp.count() == 0 || m.tier == "Garbage" {
                                        let (picks_remaining, mega_allowed) = {
                                            let c = &draft.coaches[discord_id];
                                            (c.get_remaining_picks(draft, &m.tier), c.megas.len() < c.megas_allowed.into())
                                        };
                                        if picks_remaining > 0 {
                                            if mega_id.is_none() || mega_allowed {
                                                let already_drafted = {
                                                    let m = &draft.mons[&mon];
                                                    m.drafted
                                                };
                                                if !already_drafted {
                                                    draft.mons.entry(mon.clone()).and_modify(|m| {
                                                        m.coach = Some(discord_id.clone());
                                                        m.drafted = true;
                                                        m.mega_id = mega_id.clone();
                                                        if mega_id.is_some() {
                                                            let mega = mega_id.clone().unwrap();
                                                            if let Ok(mega_f) = PokemonForm::from_id(mega) {
                                                                m.form_name = match &mega_f.form_names.iter().filter(|x| x.language.name == "en").next() {
                                                                    Some(n) => Some(n.name.clone()),
                                                                    None => None,
                                                                };
                                                            }
                                                        }
                                                    });
                                                    let m = &draft.mons[&mon];
                                                    draft.coaches.entry(discord_id.clone()).and_modify(|c| {
                                                        c.mons.push(m.id);
                                                        if mega_id.is_some() {
                                                            c.megas.insert(mega_id.unwrap());
                                                        }
                                                    });
                                                    if draft.last_round_drafted.len() == draft.coaches.len() - 1 {
                                                        draft.last_round_drafted.remove(0);
                                                    }
                                                    draft.last_round_drafted.push((*discord_id, m.id));
                                                    let c = &draft.coaches[discord_id];
                                                    draft.draft_round.remove(0);
                                                    if draft.draft_round.len() == 0 && c.mons.len() < draft.total_picks as usize {
                                                        draft.draft_round = draft.draft_order.clone();
                                                        if draft.draft_type == DraftType::Snake && c.mons.len() % 2 != 0 {
                                                            draft.draft_round.reverse();
                                                        }
                                                    }
                                                    if draft.draft_round.len() == 0 && c.mons.len() >= draft.total_picks as usize {
                                                        draft.status = DraftStatus::Playing;
                                                    }
                                                    draft.save();
            
                                                    let round = if draft.draft_round.len() == draft.draft_order.len() || draft.draft_round.len() == 0 {
                                                        c.mons.len() + 1
                                                    }
                                                    else {
                                                        c.mons.len()
                                                    };
                                
                                                    result = Ok((draft.draft_name.clone(), m.mon_name(), round as u16, draft.draft_round.len() > 0));
                                                }
                                                else {
                                                    result = Err("That Pokémon has already been drafted by someone else!".to_owned());
                                                }
                                            }
                                            else {
                                                if moderator {
                                                    result = Err("They can't pick Mega Pokémon!".to_owned());
                                                }
                                                else {
                                                    result = Err("You can't pick Mega Pokémon!".to_owned());
                                                }
                                            }
                                        }
                                        else {
                                            if moderator {
                                                result = Err("They don't have any picks left in that tier!".to_owned());
                                            }
                                            else {
                                                result = Err("You don't have any picks left in that tier!".to_owned());
                                            }
                                        }
                                    }
                                    else {
                                        if moderator {
                                            result = Err("They've already drafted a Pokémon of that species!".to_owned());
                                        }
                                        else {
                                            result = Err("You've already drafted a Pokémon of that species!".to_owned());
                                        }
                                    }
                                }
                                else {
                                    result = Err("That Pokémon is in a banned tier!".to_owned());
                                }
                            }
                            else {
                                result = Err("That Pokémon is not in this draft's bank!".to_owned());
                            }
                        }
                        else {
                            if moderator {
                                result = Err("It is not their turn to pick!".to_owned());
                            }
                            else {
                                result = Err("It is not your turn to pick!".to_owned());
                            }
                        }
                    }
                    else {
                        result = Err("This draft is not in progress.".to_owned());
                    }
                });
                result
            };
            let data = self.data.read().await;
            let reg = data.get::<DraftRegistry>().expect("bot bork");
            let draft = &reg[&draft_id];
            let retult = match result {
                Ok((d, m, p, e)) => {
                    let embed = draft.mons[&mon].embed(self, draft, server_id).await.unwrap();
                    if e {
                        let round = if draft.draft_round.len() == draft.draft_order.len() || draft.draft_round.len() == 0 {
                            p - 1
                        }
                        else {
                            p
                        };
                        match self.pester_coach(draft.draft_name.clone(), draft.draft_id.clone(), draft.draft_round.first().unwrap().clone(), p).await {
                            Ok(_) => Ok((d, m, embed, round, !e)),
                            Err(why) => Err(why)
                        }
                    }
                    else {
                        let round = if draft.draft_round.len() == draft.draft_order.len() || draft.draft_round.len() == 0 {
                            p - 1
                        }
                        else {
                            p
                        };
                        Ok((d, m, embed, round, !e))
                    }
                },
                Err(why) => Err(why)
            };

            retult
        }
        else {
            Err("Draft ID does not exist.".to_owned())
        }
    }

    async fn trade(&self, draft_id: String, discord_id: &u64, o_mon: i16, n_mon: i16, moderator: bool, target_user: Option<u64>, server_id: Option<u64>) -> Result<(String, String, String, CreateEmbed, u16, TradePerformance), String> {
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("Bot didn't start right");
        if reg.contains_key(&draft_id) {
            let mut result: Result<(String, String, u16, bool, TradePerformance), String> = Err("Something has gone terribly wrong.".to_owned());
            reg.entry(draft_id.clone()).and_modify(|draft| {
                if draft.status == DraftStatus::Drafting {
                    if moderator {
                        if draft.mons.contains_key(&o_mon) && draft.mons.contains_key(&n_mon) {
                            let coach_id = Some(discord_id.clone());
                            let o_drafted_same_sp = {
                                let o = &draft.mons[&o_mon];
                                draft.mons.iter().filter(|(_, p)| {
                                    p.coach == coach_id && p.spid == o.spid
                                }).count()
                            };
                            let n_drafted_same_sp = {
                                let o = &draft.mons[&o_mon];
                                let n = &draft.mons[&n_mon];
                                draft.mons.iter().filter(|(_, p)| {
                                    p.coach == coach_id && (p.spid == n.spid && p.spid != o.spid)
                                }).count()
                            };
                            let same_tier = {
                                let o = &draft.mons[&o_mon];
                                let n = &draft.mons[&n_mon];
                                o.tier == n.tier
                            };
                            let mega_involved = {
                                let o = &draft.mons[&o_mon];
                                let n = &draft.mons[&n_mon];
                                o.mega_id.is_some() || n.mega_id.is_some()
                            };
                            let n_tier = {
                                let n = &draft.mons[&n_mon];
                                n.tier.clone()
                            };
                            
                            if o_drafted_same_sp == 1 {
                                if n_drafted_same_sp == 0 || n_tier == "Garbage" {
                                    let n_tier_picks_remaining = {
                                        let n = &draft.mons[&n_mon];
                                        let c = &draft.coaches[discord_id];
                                        c.get_remaining_picks(draft, &n.tier)
                                    };
                                    if n_tier_picks_remaining > 0 || same_tier || draft.tier_picks.is_empty() {
                                        if !mega_involved {
                                            let already_drafted = {
                                                let n = &draft.mons[&n_mon];
                                                n.drafted
                                            };
                                            if !already_drafted {
                                                draft.mons.entry(o_mon.clone()).and_modify(|m| {
                                                    m.reset(true, false);
                                                });
                                                draft.mons.entry(n_mon.clone()).and_modify(|m| {
                                                    m.coach = Some(discord_id.clone());
                                                    m.drafted = true;
                                                });
                                                draft.coaches.entry(discord_id.clone()).and_modify(|c| {
                                                    c.mons.retain(|m| m != &o_mon);
                                                    c.mons.push(n_mon.clone());
                                                });
                                                let c = &draft.coaches[discord_id];
                                                /*draft.draft_round.remove(0);
                                                if draft.draft_round.len() == 0 && c.mons.len() < draft.total_picks as usize {
                                                    draft.draft_round = draft.draft_order.clone();
                                                    if draft.draft_type == DraftType::Snake && c.mons.len() % 2 != 0 {
                                                        draft.draft_round.reverse();
                                                    }
                                                }*/
                                                if draft.draft_round.len() == 0 && c.mons.len() >= draft.total_picks as usize {
                                                    draft.status = DraftStatus::Playing;
                                                }
                                                draft.save();
                                                let n = &draft.mons[&n_mon];
    
                                                result = Ok((draft.draft_name.clone(), n.mon_name(), c.mons.len() as u16, draft.draft_round.len() > 0, TradePerformance::Fix));
                                            }
                                            else {
                                                result = Err("That Pokémon has already been drafted by someone else!".to_owned());
                                            }
                                        }
                                        else {
                                            result = Err("You can't trade Mega Pokémon!".to_owned());
                                        }
                                    }
                                    else {
                                        if moderator {
                                            result = Err("They don't have any picks left in that tier!".to_owned());
                                        }
                                        else {
                                            result = Err("You don't have any picks left in that tier!".to_owned());
                                        }
                                    }
                                }
                                else {
                                    if moderator {
                                        result = Err("They've already drafted a Pokémon of that species!".to_owned());
                                    }
                                    else {
                                        result = Err("You've already drafted a Pokémon of that species!".to_owned());
                                    }
                                }
                            }
                            else {
                                result = Err("They have not already drafted a Pokémon of that species! You can only fix a pick they have already made.".to_owned());
                            }
                        }
                        else {
                            result = Err("That Pokémon is not in this draft's bank!".to_owned());
                        }
                    }
                    else {
                        result = Err("You can't make trades during the drafting phase!".to_owned());
                    }
                }
                else if draft.status == DraftStatus::Playing {
                    let pn = if moderator { "They" } else { "You" };
                    match (draft.trade_policy != TradePolicy::Disabled, moderator) {
                        (true, _) => {
                            match (draft.trade_policy, target_user) {
                                //player trade
                                (TradePolicy::Free, Some(u)) | (TradePolicy::Player, Some(u)) => {
                                    if !moderator {
                                        if draft.mons.contains_key(&o_mon) && draft.mons.contains_key(&n_mon) {
                                            let local_id = Some(discord_id.clone());
                                            let remote_id = Some(u);
                                            let local_o_drafted_same_sp = {
                                                let o = &draft.mons[&o_mon];
                                                draft.mons.iter().filter(|(_, p)| {
                                                    //drafted by local
                                                    p.coach == local_id &&
                                                    //AND species is the same as what's being traded away
                                                    p.spid == o.spid
                                                }).count()
                                            };
                                            let local_n_drafted_same_sp = {
                                                let o = &draft.mons[&o_mon];
                                                let n = &draft.mons[&n_mon];
                                                draft.mons.iter().filter(|(_, p)| {
                                                    //drafted by local
                                                    p.coach == local_id && 
                                                    //AND species is the same as what's being received
                                                    (p.spid == n.spid && 
                                                        //AND species is NOT the same as what's being traded away
                                                        p.spid != o.spid)
                                                    //thus trading for another form of the same species IS allowed
                                                }).count()
                                            };
    
                                            let remote_n_drafted_same_sp = {
                                                let o = &draft.mons[&o_mon];
                                                let n = &draft.mons[&n_mon];
                                                draft.mons.iter().filter(|(_, p)| {
                                                    //drafted by remote
                                                    p.coach == remote_id && 
                                                    //AND species is the same as what's being received (ie local's old)
                                                    (p.spid == o.spid && 
                                                        // AND species is NOT the same as what's being traded away (local's new)
                                                        p.spid != n.spid)
                                                }).count()
                                            };
                                            let remote_o_drafted_same_sp = {
                                                let n = &draft.mons[&n_mon];
                                                draft.mons.iter().filter(|(_, p)| {
                                                    p.coach == remote_id && p.spid == n.spid
                                                }).count()
                                            };
    
                                            let same_tier = {
                                                let o = &draft.mons[&o_mon];
                                                let n = &draft.mons[&n_mon];
                                                o.tier == n.tier
                                            };

                                            let mega_involved = {
                                                let o = &draft.mons[&o_mon];
                                                let n = &draft.mons[&n_mon];
                                                o.mega_id.is_some() || n.mega_id.is_some()
                                            };

                                            debug!("{}, {}, {}, {}", local_o_drafted_same_sp, local_n_drafted_same_sp, remote_o_drafted_same_sp, remote_n_drafted_same_sp);
                                            
                                            if local_o_drafted_same_sp == 1 && remote_o_drafted_same_sp == 1 {
                                                if local_n_drafted_same_sp == 0 && remote_n_drafted_same_sp == 0 {
                                                    if !mega_involved {
                                                        let local_trades_remaining = {
                                                            let c = &draft.coaches[discord_id];
                                                            c.trade_limit(draft) - c.used_trades
                                                        };
                                                        let remote_trades_remaining = {
                                                            let c = &draft.coaches[&u];
                                                            c.trade_limit(draft) - c.used_trades
                                                        };
                                                        if local_trades_remaining > 0 && remote_trades_remaining > 0 {
                                                            let local_n_tier_picks_remaining = {
                                                                let n = &draft.mons[&n_mon];
                                                                let c = &draft.coaches[discord_id];
                                                                c.get_remaining_picks(draft, &n.tier)
                                                            };
            
                                                            let remote_o_tier_picks_remaining = {
                                                                let o = &draft.mons[&o_mon];
                                                                let c = &draft.coaches[&u];
                                                                c.get_remaining_picks(draft, &o.tier)
                                                            };
            
                                                            if (local_n_tier_picks_remaining > 0 && remote_o_tier_picks_remaining > 0) || same_tier || draft.tier_picks.is_empty() {
                                                                draft.mons.entry(o_mon.clone()).and_modify(|m| {
                                                                    m.reset(true, false);
                                                                    m.coach = Some(u);
                                                                    m.drafted = true;
                                                                });
                                                                draft.mons.entry(n_mon.clone()).and_modify(|m| {
                                                                    m.reset(true, false);
                                                                    m.coach = Some(discord_id.clone());
                                                                    m.drafted = true;
                                                                });
            
                                                                draft.coaches.entry(discord_id.clone()).and_modify(|c| {
                                                                    c.mons.retain(|m| m != &o_mon);
                                                                    c.mons.push(n_mon.clone());
                                                                    if !moderator { c.used_trades += 1; }
                                                                });
                                                                draft.coaches.entry(u).and_modify(|c| {
                                                                    c.mons.retain(|m| m != &n_mon);
                                                                    c.mons.push(o_mon.clone());
                                                                    if !moderator { c.used_trades += 1; }
                                                                });
                                                                
                                                                let c = &draft.coaches[discord_id];
            
                                                                draft.save();
                                                                let n = &draft.mons[&n_mon];
                                            
                                                                result = Ok((draft.draft_name.clone(), n.mon_name(), c.mons.len() as u16, draft.draft_round.len() > 0, TradePerformance::CoachPlayer));
                                                            }
                                                            else {
                                                                result = Err(format!("{} can't make player trades across different tiers!", pn));
                                                            }
                                                        }
                                                        else {
                                                            result = Err(format!("{} don't have any trades left!", pn));
                                                        }
                                                    }
                                                    else {
                                                        result = Err("You can't trade Mega Pokémon!".to_owned());
                                                    }
                                                }
                                                else {
                                                    result = Err(format!("{}'ve already drafted a Pokémon of that species!", pn));
                                                }
                                            }
                                            else {
                                                result = Err(format!("{} have not drafted a Pokémon of that species! {} can only trade away a Pokémon {} already have.", pn, pn, pn.to_lowercase()));
                                            }
                                        }
                                        else {
                                            result = Err("That Pokémon is not in this draft's bank!".to_owned());
                                        }
                                    }
                                    else {
                                        result = Err("Moderator-forced player trades are currently unimplemented.".into());
                                    }
                                },
                                //bank trade
                                (TradePolicy::Free, None) | (TradePolicy::Bank, None) => {
                                    if draft.mons.contains_key(&o_mon) && draft.mons.contains_key(&n_mon) {
                                        let coach_id = Some(discord_id.clone());
                                        let o_drafted_same_sp = {
                                            let o = &draft.mons[&o_mon];
                                            draft.mons.iter().filter(|(_, p)| {
                                                p.coach == coach_id && p.spid == o.spid
                                            }).count()
                                        };
                                        let n_drafted_same_sp = {
                                            let o = &draft.mons[&o_mon];
                                            let n = &draft.mons[&n_mon];
                                            draft.mons.iter().filter(|(_, p)| {
                                                p.coach == coach_id && (p.spid == n.spid && p.spid != o.spid)
                                            }).count()
                                        };
                                        let same_tier = {
                                            let o = &draft.mons[&o_mon];
                                            let n = &draft.mons[&n_mon];
                                            o.tier == n.tier
                                        };
                                        let n_tier = {
                                            let n = &draft.mons[&n_mon];
                                            n.tier.clone()
                                        };
                                        
                                        if o_drafted_same_sp == 1 {
                                            if n_drafted_same_sp == 0 || n_tier == "Garbage" {
                                                let trades_remaining = {
                                                    let c = &draft.coaches[discord_id];
                                                    c.trade_limit(draft) - c.used_trades
                                                };

                                                if trades_remaining > 0 {

                                                    let n_tier_picks_remaining = {
                                                        let n = &draft.mons[&n_mon];
                                                        let c = &draft.coaches[discord_id];
                                                        c.get_remaining_picks(draft, &n.tier)
                                                    };
                                                    if n_tier_picks_remaining > 0 || same_tier || draft.tier_picks.is_empty() {
                                                        let already_drafted = {
                                                            let n = &draft.mons[&n_mon];
                                                            n.drafted
                                                        };
                                                        if !already_drafted {
                                                            draft.mons.entry(o_mon.clone()).and_modify(|m| {
                                                                m.reset(true, true);
                                                            });
                                                            draft.mons.entry(n_mon.clone()).and_modify(|m| {
                                                                m.reset(true, false);
                                                                m.coach = Some(discord_id.clone());
                                                                m.drafted = true;
                                                            });
                                                            draft.coaches.entry(discord_id.clone()).and_modify(|c| {
                                                                c.mons.retain(|m| m != &o_mon);
                                                                c.mons.push(n_mon.clone());
                                                                if !moderator { c.used_trades += 1; }
                                                            });
                                                            let c = &draft.coaches[discord_id];
                                                            /*draft.draft_round.remove(0);
                                                            if draft.draft_round.len() == 0 && c.mons.len() < draft.total_picks as usize {
                                                                draft.draft_round = draft.draft_order.clone();
                                                                if draft.draft_type == DraftType::Snake && c.mons.len() % 2 != 0 {
                                                                    draft.draft_round.reverse();
                                                                }
                                                            }*/
                                                            if draft.draft_round.len() == 0 && c.mons.len() >= draft.total_picks as usize {
                                                                draft.status = DraftStatus::Playing;
                                                            }
                                                            draft.save();
                                                            let n = &draft.mons[&n_mon];
                                        
                                                            result = Ok((draft.draft_name.clone(), n.mon_name(), c.mons.len() as u16, draft.draft_round.len() > 0, TradePerformance::CoachBank));
                                                        }
                                                        else {
                                                            result = Err("That Pokémon has already been drafted by someone else!".to_owned());
                                                        }
                                                    }
                                                    else {
                                                        result = Err(format!("{} don't have any picks left in that tier!", pn));
                                                    }
                                                }
                                                else {
                                                    result = Err(format!("{} don't have any trades left!", pn));
                                                }
                                            }
                                            else {
                                                result = Err(format!("{}'ve already drafted a Pokémon of that species!", pn));
                                            }
                                        }
                                        else {
                                            result = Err(format!("{} have not drafted a Pokémon of that species! {} can only trade away a Pokémon {} already have.", pn, pn, pn.to_lowercase()));
                                        }
                                    }
                                    else {
                                        result = Err("That Pokémon is not in this draft's bank!".to_owned());
                                    }
                                },
                                //mod-force player trade (dubious)
                                (TradePolicy::Bank, Some(_)) => {
                                    result = Err("Moderator-forced player trades are currently unimplemented.".into());
                                },
                                //mod-force bank trade
                                (TradePolicy::Player, None) => {
                                    if moderator {
                                        if draft.mons.contains_key(&o_mon) && draft.mons.contains_key(&n_mon) {
                                            let coach_id = Some(discord_id.clone());
                                            let o_drafted_same_sp = {
                                                let o = &draft.mons[&o_mon];
                                                draft.mons.iter().filter(|(_, p)| {
                                                    p.coach == coach_id && p.spid == o.spid
                                                }).count()
                                            };
                                            let n_drafted_same_sp = {
                                                let o = &draft.mons[&o_mon];
                                                let n = &draft.mons[&n_mon];
                                                draft.mons.iter().filter(|(_, p)| {
                                                    p.coach == coach_id && (p.spid == n.spid && p.spid != o.spid)
                                                }).count()
                                            };
                                            let same_tier = {
                                                let o = &draft.mons[&o_mon];
                                                let n = &draft.mons[&n_mon];
                                                o.tier == n.tier
                                            };
                                            let n_tier = {
                                                let n = &draft.mons[&n_mon];
                                                n.tier.clone()
                                            };
                                            
                                            if o_drafted_same_sp == 1 {
                                                if n_drafted_same_sp == 0 || n_tier == "Garbage" {
                                                    let n_tier_picks_remaining = {
                                                        let n = &draft.mons[&n_mon];
                                                        let c = &draft.coaches[discord_id];
                                                        c.get_remaining_picks(draft, &n.tier)
                                                    };
                                                    if n_tier_picks_remaining > 0 || same_tier || draft.tier_picks.is_empty() {
                                                        let already_drafted = {
                                                            let n = &draft.mons[&n_mon];
                                                            n.drafted
                                                        };
                                                        if !already_drafted {
                                                            draft.mons.entry(o_mon.clone()).and_modify(|m| {
                                                                m.reset(true, true);
                                                            });
                                                            draft.mons.entry(n_mon.clone()).and_modify(|m| {
                                                                m.reset(true, false);
                                                                m.coach = Some(discord_id.clone());
                                                                m.drafted = true;
                                                            });
                                                            draft.coaches.entry(discord_id.clone()).and_modify(|c| {
                                                                c.mons.retain(|m| m != &o_mon);
                                                                c.mons.push(n_mon.clone());
                                                            });
                                                            let c = &draft.coaches[discord_id];
                                                            /*draft.draft_round.remove(0);
                                                            if draft.draft_round.len() == 0 && c.mons.len() < draft.total_picks as usize {
                                                                draft.draft_round = draft.draft_order.clone();
                                                                if draft.draft_type == DraftType::Snake && c.mons.len() % 2 != 0 {
                                                                    draft.draft_round.reverse();
                                                                }
                                                            }*/
                                                            if draft.draft_round.len() == 0 && c.mons.len() >= draft.total_picks as usize {
                                                                draft.status = DraftStatus::Playing;
                                                            }
                                                            draft.save();
                                                            let n = &draft.mons[&n_mon];
                                        
                                                            result = Ok((draft.draft_name.clone(), n.mon_name(), c.mons.len() as u16, draft.draft_round.len() > 0, TradePerformance::ModBank));
                                                        }
                                                        else {
                                                            result = Err("That Pokémon has already been drafted by someone else!".to_owned());
                                                        }
                                                    }
                                                    else {
                                                        result = Err(format!("{} don't have any picks left in that tier!", pn));
                                                    }
                                                }
                                                else {
                                                    result = Err(format!("{}'ve already drafted a Pokémon of that species!", pn));
                                                }
                                            }
                                            else {
                                                result = Err(format!("{} have not drafted a Pokémon of that species! {} can only trade away a Pokémon {} already have.", pn, pn, pn.to_lowercase()));
                                            }
                                        }
                                        else {
                                            result = Err("That Pokémon is not in this draft's bank!".to_owned());
                                        }
                                    }
                                    else {
                                        result = Err("This draft does not allow bank trades!".to_owned());
                                    }
                                },
                                _ => {
                                    result = Err("An unexpected error occurred when attempting to trade.".to_owned());
                                }
                            }
                        },
                        (false, true) => {
                            match target_user {
                                Some(_) => {
                                    result = Err("Moderator-forced player trades are currently unimplemented.".into());
                                },
                                None => {
                                    if draft.mons.contains_key(&o_mon) && draft.mons.contains_key(&n_mon) {
                                        let coach_id = Some(discord_id.clone());
                                        let o_drafted_same_sp = {
                                            let o = &draft.mons[&o_mon];
                                            draft.mons.iter().filter(|(_, p)| {
                                                p.coach == coach_id && p.spid == o.spid
                                            }).count()
                                        };
                                        let n_drafted_same_sp = {
                                            let o = &draft.mons[&o_mon];
                                            let n = &draft.mons[&n_mon];
                                            draft.mons.iter().filter(|(_, p)| {
                                                p.coach == coach_id && (p.spid == n.spid && p.spid != o.spid)
                                            }).count()
                                        };
                                        let same_tier = {
                                            let o = &draft.mons[&o_mon];
                                            let n = &draft.mons[&n_mon];
                                            o.tier == n.tier
                                        };
                                        let n_tier = {
                                            let n = &draft.mons[&n_mon];
                                            n.tier.clone()
                                        };
                                        
                                        if o_drafted_same_sp == 1 {
                                            if n_drafted_same_sp == 0 || n_tier == "Garbage" {
                                                let n_tier_picks_remaining = {
                                                    let n = &draft.mons[&n_mon];
                                                    let c = &draft.coaches[discord_id];
                                                    c.get_remaining_picks(draft, &n.tier)
                                                };
                                                if n_tier_picks_remaining > 0 || same_tier || draft.tier_picks.is_empty() {
                                                    let already_drafted = {
                                                        let n = &draft.mons[&n_mon];
                                                        n.drafted
                                                    };
                                                    if !already_drafted {
                                                        draft.mons.entry(o_mon.clone()).and_modify(|m| {
                                                            m.reset(true, true);
                                                        });
                                                        draft.mons.entry(n_mon.clone()).and_modify(|m| {
                                                            m.reset(true, false);
                                                            m.coach = Some(discord_id.clone());
                                                            m.drafted = true;
                                                        });
                                                        draft.coaches.entry(discord_id.clone()).and_modify(|c| {
                                                            c.mons.retain(|m| m != &o_mon);
                                                            c.mons.push(n_mon.clone());
                                                        });
                                                        let c = &draft.coaches[discord_id];
                                                        /*draft.draft_round.remove(0);
                                                        if draft.draft_round.len() == 0 && c.mons.len() < draft.total_picks as usize {
                                                            draft.draft_round = draft.draft_order.clone();
                                                            if draft.draft_type == DraftType::Snake && c.mons.len() % 2 != 0 {
                                                                draft.draft_round.reverse();
                                                            }
                                                        }*/
                                                        if draft.draft_round.len() == 0 && c.mons.len() >= draft.total_picks as usize {
                                                            draft.status = DraftStatus::Playing;
                                                        }
                                                        draft.save();
                                                        let n = &draft.mons[&n_mon];
                                    
                                                        result = Ok((draft.draft_name.clone(), n.mon_name(), c.mons.len() as u16, draft.draft_round.len() > 0, TradePerformance::ModBank));
                                                    }
                                                    else {
                                                        result = Err("That Pokémon has already been drafted by someone else!".to_owned());
                                                    }
                                                }
                                                else {
                                                    result = Err(format!("{} don't have any picks left in that tier!", pn));
                                                }
                                            }
                                            else {
                                                result = Err(format!("{}'ve already drafted a Pokémon of that species!", pn));
                                            }
                                        }
                                        else {
                                            result = Err(format!("{} have not drafted a Pokémon of that species! {} can only trade away a Pokémon {} already have.", pn, pn, pn.to_lowercase()));
                                        }
                                    }
                                    else {
                                        result = Err("That Pokémon is not in this draft's bank!".to_owned());
                                    }
                                }
                            }
                        },
                        (false, false) => {
                            result = Err("This draft does not allow trades!".to_owned());
                        }
                    }
                }
                else {
                    result = Err("This draft is not in a state ready for trading.".to_owned());
                }
            });
            let draft = &reg[&draft_id];
            let retult = match result {
                Ok((d, m, p, _, r)) => {
                    let n = &draft.mons[&n_mon];
                    let o = &draft.mons[&o_mon];
                    let e = match r {
                        TradePerformance::Fix => {
                            let m = &draft.mons[&n_mon];
                            m.embed(self, draft, server_id).await.expect("embed failure")
                        },
                        TradePerformance::CoachBank => {
                            let m = &draft.mons[&n_mon];
                            m.embed(self, draft, server_id).await.expect("embed failure")
                        },
                        TradePerformance::CoachPlayer => {
                            let l_c = &draft.coaches[discord_id];
                            let l_name = l_c.display_name(self, server_id).await;
                            let r_c = &draft.coaches[&target_user.unwrap()];
                            let r_name = r_c.display_name(self, server_id).await;
                            let mut e = CreateEmbed::default();
                            e.title(format!("{}'s Trade with {}", &l_name, &r_name));
                            e.field(format!("{} receives:", &l_name), n.field_name(self).await, false);
                            e.field(format!("{} receives:", &r_name), o.field_name(self).await, false);
                            e.color(EMBED_COLOR);

                            e
                        },
                        TradePerformance::ModBank => {
                            let m = &draft.mons[&n_mon];
                            m.embed(self, draft, server_id).await.expect("embed failure")
                        },
                        /*TradePerformance::ModPlayer => {
                            let m = &draft.mons[&n_mon];
                            m.embed(self, draft, server_id).await.expect("embed failure")
                        }*/
                    };
                    Ok((d, m, o.mon_name(), e, p, r))
                },
                Err(why) => Err(why)
            };

            retult
        }
        else {
            Err("Draft ID does not exist.".to_owned())
        }
    }

    async fn get_mon_embed(&self, draft_id: String, mon: i16, server_id: Option<u64>) -> Result<CreateEmbed, Error> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[&draft_id];
        if draft.mons.contains_key(&mon) {
            let m = &draft.mons[&mon];
            if draft.tiers_allowed.contains(&m.tier) {
                m.embed(self, draft, server_id).await
            }
            else {
                Err(Error::Other("That Pokémon is not legal in this draft."))
            }
        }
        else {
            Err(Error::Other("That Pokémon is not legal in this draft."))
        }
    }

    async fn get_draft_owner_and_moderators(&self, draft_id: &String) -> Result<(u64, Option<u64>), String> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("Bot didn't start right");
        if reg.contains_key(draft_id) {
            let draft = &reg[draft_id];
            let o = draft.owner_id;
            let m = draft.moderator_role_id.clone();
            Ok((o, m))
        }
        else {
            Err("Draft ID does not exist.".to_owned())
        }
    }

    async fn set_coach_value(&self, draft_id: String, discord_id: &u64, key: String, value: String) -> Result<String, String> {
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("Bot didn't start right");
        if reg.contains_key(&draft_id) {
            let mut result: Result<String, String> = Err("Something has gone terribly wrong.".to_owned());
            reg.entry(draft_id.clone()).and_modify(|draft| {
                if draft.coaches.contains_key(discord_id) {
                    draft.coaches.entry(*discord_id).and_modify(|coach| {
                        match key.as_str() {
                            "showdown" => {
                                result = Ok(coach.showdown_name.clone().unwrap_or("".to_owned()));
                                let value = value.to_lowercase();
                                coach.showdown_name = Some(value);
                            },
                            "team" | "name" => {
                                result = Ok(coach.name.clone().unwrap_or("".to_owned()));
                                coach.name = Some(value);
                            },
                            "logo" => {
                                result = Ok(coach.logo.clone().unwrap_or("".to_owned()));
                                coach.logo = Some(value);
                            },
                            _ => (),
                        }
                    });
                    draft.save();
                }
                else {
                    result = Err("You're not a member of this draft!".to_owned())
                }
            });
            result
        }
        else {
            Err("No such draft!".to_owned())
        }
    }

    async fn modify_draft(&self, msg: &Message, draft_id: String, mut args: Args) -> Result<(), Error> {
        let mut draft = {
            let mut data = self.data.write().await;
            let reg = data.get_mut::<DraftRegistry>().expect("Bot didn't start right");
            reg.remove(&draft_id).unwrap()
        };

        //let old_tiers = draft.tiers_allowed.clone();

        let mut modify: HashSet<String> = HashSet::new();
        async fn process_kv(ctx: &Context, msg: &Message, key: &str, value: &str, draft: &mut Draft, modify: &mut HashSet<String>) {
            match key {
                "name" | "title" => {
                    modify.insert("name".to_owned());
                    draft.draft_name = value.to_owned();
                },
                "logo" => {
                    if value.starts_with("https:") {
                        modify.insert("logo".to_owned());
                        draft.logo = Some(value.to_owned());
                    }
                },
                "announcements" | "channel" => {
                    let v = value.replace("#", "")
                        .replace("<", "")
                        .replace(">", "");
                    match v.parse::<u64>() {
                        Ok(c) => {
                            modify.insert("channel".to_owned());
                            draft.announcements_channel = Some(c);
                        },
                        Err(_) => ()
                    }
                },
                "porygon" => {
                    let v = value.replace("#", "")
                        .replace("<", "")
                        .replace(">", "");
                    match v.parse::<u64>() {
                        Ok(c) => {
                            let (isc, _) = ctx.is_porygon_channel(msg.guild_id.unwrap().as_u64(), &c).await;
                            if !isc {
                                modify.insert("porygon".to_owned());
                                draft.porygon_channel = Some(c);
                            }
                        },
                        Err(_) => ()
                    }
                },
                "mods" | "moderators" | "mod" | "moderator" | "modrole" => {
                    let v = value.replace("@", "")
                        .replace("&", "")
                        .replace("<", "")
                        .replace(">", "");
                    match v.parse::<u64>() {
                        Ok(c) => {
                            modify.insert("modrole".to_owned());
                            draft.moderator_role_id = Some(c);
                        },
                        Err(_) => ()
                    }
                },
                "format" => {
                    let l = value.to_lowercase();
                    match l.as_str() {
                        "single" | "singles" => {
                            modify.insert("format".to_owned());
                            draft.format = Format::Singles;
                        },
                        "double" | "doubles" => {
                            modify.insert("format".to_owned());
                            draft.format = Format::Doubles;
                        },
                        "triple" | "triples" => (),
                        "rotation" => (),
                        _ => (),
                    }
                },
                "tiers" | "tier" | "tiersource" => {
                    let l = value.to_lowercase();
                    match l.as_str() {
                        "smogon" => {
                            modify.insert("tiersource".to_owned());
                            draft.tier_source = TierSource::Smogon;
                        },
                        "custom" => {
                            modify.insert("tiersource".to_owned());
                            draft.tier_source = TierSource::Custom;
                        },
                        "wbe" => {
                            modify.insert("tiersource".to_owned());
                            draft.tier_source = TierSource::WBE;
                        },
                        "ppa" => {
                            modify.insert("tiersource".to_owned());
                            draft.tier_source = TierSource::PPA;
                        },
                        _ => (),
                    }
                },
                "legaltiers" | "legal" | "allowed" | "allowedtiers" => {
                    if draft.tier_source == TierSource::Smogon || draft.tier_source == TierSource::PPA {
                        let despaced = value.replace(" ", "");
                        let split: Vec<String> = despaced.split(',').map(|s| s.to_owned()).collect();
                        let tiers = ctx.get_tiers(&draft.format, &draft.tier_source).await;
                        let allowed: Vec<String> = tiers.iter().filter(|x| {
                            split.contains(x)
                        }).map(|x| x.to_owned()).collect();
                        modify.insert("tiers".to_owned());
                        draft.tiers_allowed = allowed;
                    }
                },
                "bannedtiers" | "banned" => {
                    if draft.tier_source == TierSource::Smogon || draft.tier_source == TierSource::PPA {
                        let despaced = value.replace(" ", "");
                        let banned: Vec<String> = despaced.split(',').map(|s| s.to_owned()).collect();
                        let tiers = ctx.get_tiers(&draft.format, &draft.tier_source).await;
                        let allowed: Vec<String> = tiers.iter().filter(|x| {
                            !banned.contains(x)
                        }).map(|x| x.to_owned()).collect();
                        modify.insert("tiers".to_owned());
                        draft.tiers_allowed = allowed;
                    }
                },
                "league" => {
                    let l = value.to_lowercase();
                    match l.as_str() {
                        "true" | "t" | "yes" | "y" => {
                            modify.insert("league".to_owned());
                            draft.league = true;
                        },
                        "false" | "f" | "no" | "n" => {
                            modify.insert("league".to_owned());
                            draft.league = false;
                        },
                        _ => (),
                    }
                },
                "tournament" => {
                    let l = value.to_lowercase();
                    match l.as_str() {
                        "true" | "t" | "yes" | "y" => {
                            modify.insert("league".to_owned());
                            draft.league = false;
                        },
                        "false" | "f" | "no" | "n" => {
                            modify.insert("league".to_owned());
                            draft.league = true;
                        },
                        _ => (),
                    }
                },
                "season" => {
                    if let Ok(s) = value.parse::<u16>() {
                        modify.insert("season".to_owned());
                        draft.league_season = s;
                    }
                },
                "picks" => {
                    if let Ok(s) = value.parse::<u8>() {
                        modify.insert("picks".to_owned());
                        draft.total_picks = s;
                    }
                },
                "trading" | "tradepolicy" => {
                    let l = value.to_lowercase();
                    match l.as_str() {
                        "true" | "t" | "yes" | "y" | "free" => {
                            modify.insert("trading".to_owned());
                            draft.trade_policy = TradePolicy::Free;
                        },
                        "bank" => {
                            modify.insert("trading".to_owned());
                            draft.trade_policy = TradePolicy::Bank;
                        },
                        "player" => {
                            modify.insert("trading".to_owned());
                            draft.trade_policy = TradePolicy::Player;
                        },
                        "false" | "f" | "no" | "n" => {
                            modify.insert("trading".to_owned());
                            draft.trade_policy = TradePolicy::Disabled;
                        },
                        _ => (),
                    }
                },
                "trades" | "trademax" | "maxtrades" => {
                    if let Ok(s) = value.parse::<u8>() {
                        modify.insert("trades".to_owned());
                        draft.trades_max = s;
                    }
                },
                "direction" | "order" => {
                    let l = value.to_lowercase();
                    match l.as_str() {
                        "linear" | "l" => {
                            modify.insert("direction".to_owned());
                            draft.draft_type = DraftType::Linear;
                        },
                        "snake" | "s" => {
                            modify.insert("direction".to_owned());
                            draft.draft_type = DraftType::Snake;
                        },
                        _ => (),
                    }
                },
                "closed" => {
                    draft.closed = !draft.closed;
                },
                _ => (),
            }
        }
        while let Ok(arg) = args.single_quoted::<String>() {
            let split: Vec<String> = arg.split(':').map(|s| s.to_owned()).collect();
            if split.len() == 2 && !vec!["", " "].contains(&split[1].as_str()) {
                let l = split[0].to_lowercase();
                let key = l.as_str();
                let value = split[1].as_str();
                process_kv(self, msg, key, value, &mut draft, &mut modify).await;
            }
        }
        let mut done = false;
        
        while !done {
            let ready = {
                draft.tier_source == TierSource::Custom ||
                draft.tier_source == TierSource::WBE || 
                (
                    draft.tier_source == TierSource::Smogon && !draft.tiers_allowed.is_empty()
                ) || 
                (
                    draft.tier_source == TierSource::PPA && !draft.tiers_allowed.is_empty()
                )
            };
        
            let (desc, foot) = if ready {
                (
                    "Your draft is **ready** to be resumed, but **review the information below *carefully*** before confirming.",
                    "Respond to this message with more draft settings in key:value format, or respond \"done\"."
                )
            }
            else {
                (
                    "Your draft is still missing some essential information. **Review the information below *carefully***. __Underlined__ fields *must* be filled before the draft is ready.",
                    "Respond to this message with more draft settings in key:value format."
                )
            };

            let mut embed = draft.embed(&self).await.unwrap();

            let message = msg.channel_id.send_message(self, |m| {
                embed.title("Draft Modification Process");
                embed.description(desc);
                embed.footer(|f| {
                    f.text(foot);
                    f
                });
                m.set_embed(embed);
                m
            }).await?;

            // wait for replies
            let r = match msg.author.await_reply(&self)
                .channel_id(msg.channel_id)
                .timeout(Duration::from_secs(6000))
                .filter(|m| {
                    let c_u = m.content.to_lowercase();
                    m.content.contains(":") || c_u == "done".to_owned()
                })
                .await {
                Some(answer) => {
                    let resp = answer.content.clone();
                    let _ = answer.delete(self).await;
                    resp
                },
                None => {
                    "timeout".to_owned()
                }
            };

            let mut args = Args::new(r.as_str(), &[Delimiter::Single(' ')]);
            
            while let Ok(arg) = args.single_quoted::<String>() {
                let split: Vec<String> = arg.split(':').map(|s| s.to_owned()).collect();
                if split.len() == 2 && !vec!["", " "].contains(&split[1].as_str()) {
                    let l = split[0].to_lowercase();
                    let key = l.as_str();
                    let value = split[1].as_str();
                    process_kv(self, msg, key, value, &mut draft, &mut modify).await;
                }
                else if split.len() == 1 && split[0].to_lowercase().as_str() == "done" && ready {
                    done = true;
                }
            }

            let _ = message.delete(self).await?;
        }

        if draft.tier_source == TierSource::Smogon {
            if modify.contains("tiersource") {
                let old_mons = draft.mons.clone();
                let mut prog = msg.channel_id.send_message(self, |m| {
                    m.embed(|e| {
                        e.title("Loading Tiered Mons");
                        e.description("Auto-populating draft bank from Smogon tier lists. This may take a while.");
                        e.color(EMBED_COLOR);
        
                        e
                    });
        
                    m
                }).await?;
                let sm_filter = {
                    let data = self.data.read().await;
                    let reg = if draft.format == Format::Singles {
                        data.get::<SmogonTierList>().unwrap()
                    }
                    else {
                        data.get::<DoublesTierList>().unwrap()
                    };
                    let f: HashMap<i16, String> = reg.iter().filter(|(_, y)| {
                        draft.tiers_allowed.contains(y)
                    }).map(|(x, y)| (x.clone(), y.clone())).collect();
                    f
                };
                let mut tiered: HashMap<i16, DraftPokemon> = HashMap::new();
                let mut i: u64 = 1;
                let len = sm_filter.len();
                debug!("Draft Modify Tiers:");
                for (x, y) in sm_filter {
                    debug!("    {}, {}", x, &y);
                    let d = DraftPokemon::new_from_id(x.to_owned(), y.to_owned(), None).unwrap();
                    let _ = prog.edit(self, |m| {
                        m.embed(|e| {
                            e.title("Loading Tiered Mons");
                            e.description("Auto-populating draft bank from Smogon tier lists. This may take a while.");
        
                            e.field("Progress", format!("{} / {}", i, len), false);
                            e.color(EMBED_COLOR);
        
                            e
                        });
        
                        m
                    }).await;
                    tiered.insert(d.id.clone(), d);
        
                    i += 1;
                }
        
                draft.mons = tiered;

                for (i, m) in &mut draft.mons {
                    if old_mons.contains_key(i) {
                        let o = &old_mons[i];
                        m.active_kills = o.active_kills;
                        m.awards = o.awards.clone();
                        m.coach = o.coach;
                        m.deaths = o.deaths;
                        m.drafted = o.drafted;
                        m.lifetime_active = o.lifetime_active;
                        m.lifetime_deaths = o.lifetime_deaths;
                        m.lifetime_passive = o.lifetime_passive;
                        m.passive_kills = o.passive_kills;
                        m.returned = o.returned;
                    }
                }
            }
        }
        else if draft.tier_source == TierSource::PPA {
            if modify.contains("tiersource") {
                let old_mons = draft.mons.clone();
                let mut prog = msg.channel_id.send_message(self, |m| {
                    m.embed(|e| {
                        e.title("Loading Tiered Mons");
                        e.description("Auto-populating draft bank from PPA tier lists. This may take a while.");
                        e.color(EMBED_COLOR);
        
                        e
                    });
        
                    m
                }).await?;
                let sm_filter = {
                    let data = self.data.read().await;
                    let reg = data.get::<PPATierList>().unwrap();
                    let f: HashMap<i16, (String, Option<u8>)> = reg.iter().filter(|(_, (y, _))| {
                        draft.tiers_allowed.contains(y)
                    }).map(|(x, y)| (x.clone(), y.clone())).collect();
                    f
                };
                let mut tiered: HashMap<i16, DraftPokemon> = HashMap::new();
                let mut i: u64 = 1;
                let len = sm_filter.len();
                debug!("Draft Modify Tiers:");
                for (x, (y, b)) in sm_filter {
                    debug!("    {}, {}", x, &y);
                    let d = DraftPokemon::new_from_id(x.to_owned(), y.to_owned(), b).unwrap();
                    let _ = prog.edit(self, |m| {
                        m.embed(|e| {
                            e.title("Loading Tiered Mons");
                            e.description("Auto-populating draft bank from PPA tier lists. This may take a while.");
        
                            e.field("Progress", format!("{} / {}", i, len), false);
                            e.color(EMBED_COLOR);
        
                            e
                        });
        
                        m
                    }).await;
                    tiered.insert(d.id.clone(), d);
        
                    i += 1;
                }
        
                draft.mons = tiered;

                for (i, m) in &mut draft.mons {
                    if old_mons.contains_key(i) {
                        let o = &old_mons[i];
                        m.active_kills = o.active_kills;
                        m.awards = o.awards.clone();
                        m.coach = o.coach;
                        m.deaths = o.deaths;
                        m.drafted = o.drafted;
                        m.lifetime_active = o.lifetime_active;
                        m.lifetime_deaths = o.lifetime_deaths;
                        m.lifetime_passive = o.lifetime_passive;
                        m.passive_kills = o.passive_kills;
                        m.returned = o.returned;
                    }
                }
            }
        }
        //draft.status = DraftStatus::Waiting;
        let mut embed = draft.embed(&self).await?;

        let _ = msg.channel_id.send_message(self, |m| {
            embed.title("Draft Modification Finalized");
            embed.description("Your draft has been modified. Don't forget to **add draftable Pokémon** if you're not using premade tiers.");
            embed.footer(|f| {
                f.text("Share the draft ID to your coaches so they can join, or add them manually.");
                f
            });
            m.set_embed(embed);
            m
        }).await?;

        draft.save();
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("Bot didn't start right");
        reg.insert(draft_id, draft);

        Ok(())
    }

    async fn filter_draft_id(&self, arg: Option<&str>) -> Result<String, DraftParseError> {
        if arg.is_some() {
            let a = arg.unwrap();
            let v = a.replace(" ", "-");
            let r = v.to_uppercase();
            if self.check_draft_id(&r).await {
                Ok(r)
            }
            else {
                Err(DraftParseError::Invalid)
            }
        }
        else {
            Err(DraftParseError::Missing)
        }
    }

    async fn set_default_draft(&self, server_id: &u64, draft_id: &String) -> Option<String> {
        let old = {
            let mut data = self.data.write().await;
            let reg = data.get_mut::<DefaultDrafts>().expect("bot bork");
            reg.insert(*server_id, draft_id.to_string())
        };
        self.save_default_drafts().await;
        old
    }

    async fn get_default_draft(&self, server_id: &u64) -> Option<String> {
        let data = self.data.read().await;
        let reg = data.get::<DefaultDrafts>().expect("bot bork");
        if reg.contains_key(server_id) {
            Some(reg[server_id].clone())
        }
        else {
            None
        }
    }

    async fn save_default_drafts(&self) {
        let data = self.data.read().await;
        let reg = data.get::<DefaultDrafts>().expect("bot brok");
        let string = ron::ser::to_string_pretty(reg, ron::ser::PrettyConfig::new()).unwrap();

        let path = Path::new("data/default_drafts.ron");
        let display = path.display();

        let mut file = match File::create(&path) {
            Err(why) => panic!("couldn't create {}: {}", display, why),
            Ok(file) => file,
        };

        match file.write_all(string.as_bytes()) {
            Err(why) => panic!("couldn't write to {}: {}", display, why),
            Ok(_) => debug!("successfully wrote to {}", display),
        }
    }

    async fn add_coach_trade(&self, draft_id: &String, coach_id: &u64, extra: i8) -> u8 {
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("bot brok");
        let mut new_extra = 0;
        let mut new_limit = 0;
        reg.entry(draft_id.clone()).and_modify(|draft| {
            draft.coaches.entry(coach_id.clone()).and_modify(|c| {
                if extra.is_negative() && c.extra_trades > 0 {
                    let sub = (extra * -1) as u8;
                    if (c.extra_trades as i8 + extra).is_negative() {
                        c.extra_trades = 0;
                    }
                    else {
                        c.extra_trades -= sub;
                    }
                }
                else if extra.is_positive() {
                    c.extra_trades += extra as u8;
                }
                new_extra = c.extra_trades;
            });
            new_limit = draft.trades_max + new_extra;
            draft.save();
        });
        new_limit
    }

    async fn add_coach_pick(&self, draft_id: &String, coach_id: &u64, tier: &String, extra: i8) -> u8 {
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("bot brok");
        //let mut new_extra = 0;
        //let mut new_limit = 0;
        reg.entry(draft_id.clone()).and_modify(|draft| {
            draft.coaches.entry(coach_id.clone()).and_modify(|c| {
                if !c.extra_picks.contains_key(tier) {
                    c.extra_picks.insert(tier.clone(), 0);
                }
                c.extra_picks.entry(tier.clone()).and_modify(|i| {
                    if extra.is_negative() && *i > 0 {
                        let sub = (extra * -1) as u8;
                        if (*i as i8 + extra).is_negative() {
                            *i = 0;
                        }
                        else {
                            *i -= sub;
                        }
                    }
                    else if extra.is_positive() {
                        *i += extra as u8;
                    }
                });
            });
            draft.save();
        });
        let draft = &reg[draft_id];
        let c = &draft.coaches[coach_id];
        c.get_remaining_picks(draft, tier)
    }

    async fn add_coach_mega(&self, draft_id: &String, coach_id: &u64, extra: i8) -> u8 {
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("bot brok");
        let mut new_limit = 0;
        reg.entry(draft_id.clone()).and_modify(|draft| {
            draft.coaches.entry(coach_id.clone()).and_modify(|c| {
                if extra.is_negative() && c.megas_allowed > 0 {
                    let sub = (extra * -1) as u8;
                    if (c.megas_allowed as i8 + extra).is_negative() {
                        c.megas_allowed = 0;
                    }
                    else {
                        c.megas_allowed -= sub;
                    }
                }
                else if extra.is_positive() {
                    c.megas_allowed += extra as u8;
                }

                new_limit = c.megas_allowed;
            });
            draft.save();
        });
        new_limit
    }

    async fn set_coach_dropout(&self, draft_id: &String, coach_id: &u64) -> (bool, String) {
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("bot brok");
        let mut newstate = false;
        reg.entry(draft_id.clone()).and_modify(|draft| {
            draft.coaches.entry(coach_id.clone()).and_modify(|c| {
                c.dropout = !c.dropout;

                newstate = c.dropout;
            });
            draft.save();
        });
        let name = reg[draft_id].coaches[coach_id].display_name(self, None).await;
        (newstate, name)
    }

    async fn add_award(&self, draft_id: &String, target: AwardableKey, award: String) -> Result<String, String> {
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("bot brok");
        let mut result = Err("Something went wrong".into());
        let mut coach = (false, 0);
        reg.entry(draft_id.clone()).and_modify(|draft| {
            match target {
                AwardableKey::Coach(u) => {
                    if draft.coaches.contains_key(&u) {
                        coach = (true, u);
                        draft.coaches.entry(u).and_modify(|c| {
                            c.grant_award(award);
                        });
                        draft.save();
                    }
                    else {
                        result = Err("That user is not a coach in this draft".into());
                    }
                },
                AwardableKey::Mon(m) => {
                    if draft.mons.contains_key(&m) {
                        draft.mons.entry(m).and_modify(|m| {
                            m.grant_award(award);
                            result = Ok(m.mon_name());
                        });
                        draft.save();
                    }
                    else {
                        result = Err("That Pokémon is not part of this draft".into());
                    }
                }
            }
        });
        if coach.0 {
            let draft = &reg[draft_id];
            let cname = draft.coaches[&coach.1].display_name(self, None).await;
            result = Ok(cname);
        }
        result
    }

    async fn start_playoffs(&self, draft_id: &String, days: Option<u8>) -> Result<String, String> {
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("bot brok");
        let mut draft = reg.remove(draft_id).unwrap();
        
        let result = if draft.status == DraftStatus::Playing && draft.league {
            if draft.team_divisions {
                draft.status = DraftStatus::SeasonPlayoffs;
                draft.challonge = None;

                Ok(format!("**Playoffs for Season {} are now ready to begin.**", draft.league_season))
            }
            else {
                match draft.create_playoff_challonge(days) {
                    Ok(_) => {
                        let days = days.unwrap_or(7);
                        Ok(format!("**Playoffs for Season {} are open for predictions!**\nVisit the Challonge bracket: {}\nMatches begin in **{} days**", draft.league_season, draft.challonge_url().unwrap(), days))
                    },
                    Err((y, d)) => {
                        draft = d;
                        Err(y)
                    }
                }
            }
        }
        else {
            Err("Playoffs must occur within a league that is currently playing a season.".into())
        };
        draft.save();
        reg.insert(draft.draft_id.clone(), draft);
        result
    }

    async fn end_season(&self, draft_id: &String) -> Result<(String, Option<CreateEmbed>), String> {
        let mut draft = {
            let mut data = self.data.write().await;
            let reg = data.get_mut::<DraftRegistry>().expect("bot brok");
            reg.remove(draft_id).unwrap()
        };
        let result = if draft.league {
            match draft.finalize_season() {
                Ok(_) => {
                    Ok(format!("Season {} has concluded", draft.league_season))
                },
                Err((y, d)) => {
                    draft = d;
                    Err(y)
                }
            }
        }
        else {
            draft.closed = true;
            draft.status = DraftStatus::TournComplete;
            Ok(format!("{} has concluded", draft.draft_name))
        };
        draft.save();
        let (league, season) = (draft.league, draft.league_season);
        {
            let mut data = self.data.write().await;
            let reg = data.get_mut::<DraftRegistry>().expect("bot brok");
            reg.insert(draft.draft_id.clone(), draft);
        }
        if league {
            let e = self.get_season_embed(draft_id, season).await.unwrap();
            result.map(|r| (r, Some(e)))
        }
        else {
            result.map(|r| (r, None))
        }
    }

    async fn new_season(&self, draft_id: &String) -> Result<String, String> {
        let ts = {
            let data = self.data.read().await;
            let reg = data.get::<DraftRegistry>().expect("bot not start correcly");
            reg[draft_id].tier_source.clone()
        };
        let ref_tiers = {
            let data = self.data.read().await;
            match ts {
                TierSource::Smogon => {
                    data.get::<SmogonTierList>().cloned().map(|e| e.iter().map(|(i, t)| (*i, (t.clone(), None))).collect::<HashMap<i16, (String, Option<u8>)>>())
                },
                TierSource::PPA => {
                    data.get::<PPATierList>().cloned()
                }
                _ => None,
            }
        };
        let mut data = self.data.write().await;
        let reg = data.get_mut::<DraftRegistry>().expect("bot brok");
        let mut draft = reg.remove(draft_id).unwrap();
        let result = if draft.league {
            match draft.new_season(self).await {
                Ok(_) => {
                    if ref_tiers.is_some() {
                        let ref_tiers = ref_tiers.unwrap();
                        draft.mons.par_iter_mut().for_each(|(i, m)| {
                            if ref_tiers.contains_key(i) {
                                if &ref_tiers[i].0 != &m.tier {
                                    m.change_tier(ref_tiers[i].0.clone(), ref_tiers[i].1);
                                }
                            }
                            else {
                                m.change_tier("Banned".to_owned(), None);
                            }
                        });
                        let new_tiereds = ref_tiers.par_iter().filter(|(i, _)| !draft.mons.contains_key(*i))
                            .filter_map(|(i, (t, a))| DraftPokemon::new_from_id(*i, t.clone(), a.clone()).ok())
                            .collect::<Vec<DraftPokemon>>();
                        for m in new_tiereds {
                            draft.mons.insert(m.id, m);
                        }
                    }

                    Ok(format!("Season {} is initialized.", draft.league_season))
                },
                Err((y, d)) => {
                    draft = d;
                    Err(y)
                }
            }
        }
        else {
            Err("Tournaments do not have multiple seasons!".into())
        };
        draft.save();
        reg.insert(draft.draft_id.clone(), draft);
        result
    }

    async fn get_season_embed(&self, draft_id: &String, season: u16) -> Result<CreateEmbed, String> {
        let data = self.data.read().await;
        let reg = data.get::<DraftRegistry>().expect("bot bork");
        let draft = &reg[draft_id];
        if draft.league {
            if draft.past_seasons.is_some() {
                let ps = draft.past_seasons.as_ref().unwrap();
                if ps.contains_key(&season) {
                    Ok(ps[&season].embed(self).await)
                }
                else {
                    Err("That is not a valid season".into())
                }
            }
            else {
                Err("This league has not completed any seasons yet".into())
            }
        }
        else {
            Err("Tournaments don't have seasons".into())
        }
    }

    async fn create_match(&self, draft_id: String, mut content: String) -> Result<Match, String> {
        info!("Processing Porygon Match Report");
        let date = Utc::now();
        content = Regex::new(r"\n+").unwrap().replace_all(&content, "\n").into_owned();
        //debug!("{}", content);
        let winner_scores = if content.contains("||") && content.contains("**") {
            Regex::new(r"Result:.*\|\| ?(?P<winner>.+) won (?P<winnerscore>[0-9])\-(?P<loserscore>[0-9])").unwrap()
        }
        else if content.contains("**") {
            Regex::new(r"Result:\*\* (?P<winner>.+) won (?P<winnerscore>[0-9])\-(?P<loserscore>[0-9])").unwrap()
        }
        else {
            Regex::new(r"Result: (?P<winner>.+) won (?P<winnerscore>[0-9])\-(?P<loserscore>[0-9])").unwrap()
        };
        let players_mons = if content.contains("**") {
            Regex::new(r"\*\*(?P<player1>.+)\*\*:.*\n(?P<p1mons>(.+,[0-9]+,[0-9]+,[0-9]+\n){1,6})(\|\|)? \n(\|\|)?\*\*(?P<player2>.+)\*\*:.*\n(?P<p2mons>(.+,[0-9]+,[0-9]+,[0-9]+\n){1,6})").unwrap()
        }
        else {
            Regex::new(r"\-[0-9]\n(?P<player1>.+): \n(?P<p1mons>(.+,[0-9]+,[0-9]+,[0-9]+\n){1,6}) \n(?P<player2>.+): \n(?P<p2mons>(.+,[0-9]+,[0-9]+,[0-9]+\n){1,6})").unwrap()
        };
        let replay = Regex::new(r"Replay.*(?P<replay>http.+[^>])>?\n").unwrap();
        let replay_cap = replay.captures(content.as_str()).unwrap();
        let win_scores_cap = winner_scores.captures(content.as_str()).unwrap();
        let p_mons_cap = players_mons.captures(content.as_str()).unwrap();
    
        let winner_string = win_scores_cap.name("winner").unwrap().as_str().to_lowercase();
        let winner = winner_string.as_str();
        let winscore = win_scores_cap.name("winnerscore").unwrap().as_str();
        let losescore = win_scores_cap.name("loserscore").unwrap().as_str();
    
        let p1user = p_mons_cap.name("player1").unwrap().as_str().to_lowercase();
        let p1mons = p_mons_cap.name("p1mons").unwrap().as_str();
        let p2user = p_mons_cap.name("player2").unwrap().as_str().to_lowercase();
        let p2mons = p_mons_cap.name("p2mons").unwrap().as_str();
    
        let replay = replay_cap.name("replay").unwrap().as_str().to_owned();
        
        let (p1e, coach1) = self.check_draft_showdown(&draft_id, &p1user).await;
        let (p2e, coach2) = self.check_draft_showdown(&draft_id, &p2user).await;
        if p1e && p2e {
            async fn process_team(ctx: &Context, mons: Split<'_, char>, team: &mut Vec<MatchPokemon>, coach: &u64, user: &String, draft_id: &String) -> Result<(), String> {
                for m in mons {
                    if m != "" {
                        let csv: Vec<String> = m.split(',').map(|s| s.to_owned()).collect();
                        let (mon, mega) = {
                            let mut m = csv[0].to_lowercase();
                            m = m.replace(" ", "-");
                            m = Regex::new(r"[\p{Punct}--[\-]]+").unwrap().replace_all(m.as_str(), "").into_owned();
                            m = Regex::new(r"-$").unwrap().replace_all(m.as_str(), "").into_owned();
    
                            if m.contains("-mega") {
                                let s: Vec<String> = m.split('-').map(|s| s.to_owned()).collect();
                                (s[0].clone(), Some(m))
                            }
                            else {
                                (m, None)
                            }
                        };
                        debug!("Porygon: Attempting to include Pokemon '{}' in match result", mon);
                        let ak = csv[1].parse::<u8>().unwrap();
                        let pk = csv[2].parse::<u8>().unwrap();
                        let d = csv[3].parse::<u8>().unwrap();
                        match PokemonForm::from_name_exhaustive(mon.as_str()) {
                            Ok(f) => {
                                let mega = mega.map(|m| PokemonForm::from_name_exhaustive(m.as_str()).ok()).flatten().map(|f| f.id);
                                if ctx.check_coach_has_mon(draft_id.clone(), coach, &f.id, mega).await {
                                    let p = MatchPokemon {
                                        id: f.id,
                                        active_kills: ak,
                                        passive_kills: pk,
                                        deaths: d,
                                    };
                                    team.push(p);
                                }
                                else {
                                    let mut mb = MessageBuilder::new();
                                    mb.push("You attempted to submit a match with an incorrect Pokémon on ");
                                    mb.push_safe(user);
                                    mb.push_line("'s team. This could just be an oversight on my part, or that coach brought a Pokémon they didn't draft.");
                                    mb.push("Problematic Pokémon: ");
                                    mb.push_mono_line(mon);
                                    return Err(mb.build());
                                }
                            },
                            Err(_) => {
                                let mut mb = MessageBuilder::new();
                                mb.push_line("You attempted to submit a match with an incorrect or malformed Pokémon name. (Or my author has forgotten an edge case. oops)");
                                mb.push("Problematic name: ");
                                mb.push_mono_line(mon);
                                return Err(mb.build());
                            },
                        }
                    }
                }
                Ok(())
            }

            info!("Porygon: Both Showdown names match, proceeding");
            let p1mons = p1mons.split('\n');
            let winscore = winscore.parse::<u8>().unwrap();
            debug!("Porygon: Player 1 team size: {}; Score: {}", p1mons.clone().count(), winscore);
            let deaths1 = (p1mons.clone().count() as u8 - 1) - winscore;
            let mut team1: Vec<MatchPokemon> = Vec::new();
            if let Err(y) = process_team(self, p1mons, &mut team1, &coach1, &p1user, &draft_id).await {
                return Err(y);
            };
            let p2mons = p2mons.split('\n');
            let losescore = losescore.parse::<u8>().unwrap();
            debug!("Porygon: Player 2 team size: {}; Score: {}", p2mons.clone().count(), losescore);
            let deaths2 = (p2mons.clone().count() as u8 - 1) - losescore;
            let mut team2: Vec<MatchPokemon> = Vec::new();
            if let Err(y) = process_team(self, p2mons, &mut team2, &coach2, &p2user, &draft_id).await {
                return Err(y);
            };
            let winner_is_1 = winner.to_owned() == p1user.clone();
            info!("Porygon: Match parse success. Winner: `{}`; Player 1: `{}`; Player 1 won: {}", winner, p1user, winner_is_1);
            let id = self.get_next_match_id(&draft_id).await as u16;
            let mut flake = Flaken::default().node(1).epoch(1613940319944);
            let uuid = flake.next();
            let (c1_score, c2_score) = if winner_is_1 { (winscore, losescore) } else { (losescore, winscore) };
    
            let m = Match {
                uuid,
                id,
                coach1,
                coach2,
                team1,
                team2,
                winner_is_1,
                c1_score,
                c2_score,
                deaths1,
                deaths2,
                replay,
                date,
                playoffs: false,
            };
    
            Ok(m)
        }
        else {
            let mut mb = MessageBuilder::new();
            mb.push_line("You attempted to submit a match with Showdown accounts that aren't registered in this draft.");
            mb.push_line("The full Porygon output will be repeated here in the next message; please fix it and then re-submit the match.");
            mb.push_bold_line("In the future, ensure your players are only using one Showdown account for draft matches.");
            return Err(mb.build());
        }
    }
}

pub async fn page<T: Clone + Listable>(ctx: &Context, msg: &Message, mut list: Vec<T>, pager_title: String, image: Option<String>) -> Result<(), String> {
    let count = list.len();
    let page_size = T::page_size() as usize;
    let len = match count > page_size {
        true => page_size,
        false => count
    };
    if count > len {
        let mut cur_sort = T::default_sort();
        let mut next_sort = T::default_sort();
        let mut cur_asc = true;
        let mut next_asc = true;
        let mut pagenum = 1;
        let mut max_pages = count / page_size;
        if count % page_size > 0 {
            max_pages += 1;
        }
        let mut page_msg = msg.channel_id.send_message(&ctx, |m| {
            m.embed(|e| {
                e.title("List Loading");
                e.color(EMBED_COLOR);
    
                e
            });
            m
        }).await.unwrap();
        loop {
            if cur_sort != next_sort {
                cur_sort = next_sort;
                match cur_sort {
                    SortType::Random => list.shuffle(&mut rand::thread_rng()),
                    SortType::Nickname | SortType::SpeciesName | SortType::StrID | SortType::Tier => {
                        list.sort_by(|a, b| a.sort_str(cur_sort).cmp(&b.sort_str(cur_sort)))
                    },
                    SortType::NumID | SortType::Type | SortType::CurrentRecord | SortType::LifetimeRecord | SortType::BST => {
                        list.sort_by(|a, b| a.sort_num(cur_sort).cmp(&b.sort_num(cur_sort)))
                    },
                    SortType::Date => {
                        list.sort_by(|a, b| a.sort_date().cmp(&b.sort_date()))
                    },
                    _ => panic!("we shouldn't get here"),
                }
                pagenum = 1;
                cur_asc = true;
                next_asc = true;
            }
            if cur_asc != next_asc {
                cur_asc = next_asc;
                list.reverse();
                pagenum = 1;
            }
            let s_index = (pagenum - 1) * page_size + 1;
            let e_index = match pagenum < max_pages {
                true => pagenum * page_size,
                false => count
            };
            let (page, _left) = list.split_at(e_index);
            let (_first, page) = page.split_at(s_index - 1);
            let page = page.to_vec();

            debug!("paging. page {} of max {}", pagenum, max_pages);

            let sort_msg = {
                let mut s = String::new();
                let asc = if cur_asc { "ascending" } else { "descending" };
                let abc = if cur_asc { "" } else { "reverse " };
                match cur_sort {
                    SortType::Random => s.push_str(format!("{} are currently sorted **randomly**.", T::type_name_plural()).as_str()),
                    SortType::NumID => s.push_str(format!("{} are currently sorted by **{} Pokédex number**.", T::type_name_plural(), asc).as_str()),
                    SortType::StrID => s.push_str(format!("{} are currently sorted by **{} draft ID**.", T::type_name_plural(), abc).as_str()),
                    SortType::Tier => s.push_str(format!("{} are currently sorted by **their drafting Tier in {}alphabetical order**.", T::type_name_plural(), abc).as_str()),
                    SortType::Type => s.push_str(format!("{} are currently sorted by **{} primary Type**.", T::type_name_plural(), asc).as_str()),
                    SortType::CurrentRecord => s.push_str(format!("{} are currently sorted by **{} current season performance**.", T::type_name_plural(), asc).as_str()),
                    SortType::LifetimeRecord => s.push_str(format!("{} are currently sorted by **{} lifetime performance**.", T::type_name_plural(), asc).as_str()),
                    SortType::BST => s.push_str(format!("{} are currently sorted by **{} base stat total**.", T::type_name_plural(), asc).as_str()),
                    SortType::SpeciesName => s.push_str(format!("{} are currently sorted by **their species names in {}alphabetical order**.", T::type_name_plural(), abc).as_str()),
                    SortType::Date => s.push_str(format!("{} are currently sorted by **{} match dates**.", T::type_name_plural(), asc).as_str()),
                    SortType::PlayoffIdentifier => s.push_str(format!("{} are currently sorted by **the playoff match letter code in {}alphabetical order**.", T::type_name_plural(), abc).as_str()),
                    _ => todo!()
                }
                s
            };
            let reax_msg = {
                let mut s = String::new();
                s.push_str("React with 🔀 to (re)sort randomly.\n");
                if T::valid_sorts().contains(&SortType::NumID) && cur_sort != SortType::NumID {
                    s.push_str("React with 🔢 to sort by Pokédex number.\n");
                }
                if T::valid_sorts().contains(&SortType::StrID) && cur_sort != SortType::StrID {
                    s.push_str("React with 🔠 to sort by Pokédex number.\n");
                }
                if T::valid_sorts().contains(&SortType::Type) && cur_sort != SortType::Type {
                    s.push_str("React with 🔥 to sort by primary Type.\n");
                }
                if T::valid_sorts().contains(&SortType::Tier) && cur_sort != SortType::Tier {
                    s.push_str("React with 🎟 to sort by drafting Tier.\n");
                }
                if T::valid_sorts().contains(&SortType::CurrentRecord) && cur_sort != SortType::CurrentRecord {
                    s.push_str("React with ⚔️ to sort by current season performance.\n");
                }
                if T::valid_sorts().contains(&SortType::CurrentRecord) && cur_sort != SortType::CurrentRecord {
                    s.push_str("React with 🏅 to sort by lifetime performance.\n");
                }
                if T::valid_sorts().contains(&SortType::BST) && cur_sort != SortType::BST {
                    s.push_str("React with 💪 to sort by Base Stat Total.\n");
                }
                if T::valid_sorts().contains(&SortType::SpeciesName) && cur_sort != SortType::SpeciesName {
                    s.push_str("React with 🔠 to sort by species name.\n");
                }
                if T::valid_sorts().contains(&SortType::Date) && cur_sort != SortType::Date {
                    s.push_str("React with 📆 to sort by match date.\n");
                }
                if T::valid_sorts().contains(&SortType::PlayoffIdentifier) && cur_sort != SortType::PlayoffIdentifier {
                    s.push_str("React with 🔠 to sort by playoff match letter code.\n");
                }
                if cur_sort != SortType::Random {
                    if cur_sort == SortType::SpeciesName {
                        if cur_asc { 
                            s.push_str("React with 🔽 to sort in reverse alphabetical order.");
                        } else {
                            s.push_str("React with 🔼 to sort in alphabetical order.");
                        };
                    } else {
                        if cur_asc { 
                            s.push_str("React with 🔽 to sort in descending order.");
                        } else {
                            s.push_str("React with 🔼 to sort in ascending order.");
                        };
                    }
                }
                s
            };
            
            let fields = {
                let mut f: Vec<(String, String)> = Vec::new();
                for v in page.clone() {
                    let t1 = Utc::now();
                    f.push((v.field_name(ctx).await, v.field_value(ctx).await));
                    let t2 = Utc::now();
                    debug!("push: {} ms", (t2 - t1).num_milliseconds());
                }
                f
            };
            
            page_msg.edit(&ctx, |m| {
                m.embed(|e| {
                    e.title(&pager_title);
                    e.description(format!("Showing {}-{} of {} {}\n{}", s_index, e_index, count, T::type_name_plural(), sort_msg));

                    if image.is_some() {
                        e.thumbnail(image.clone().unwrap());
                    }
                    
                    for (n, v) in fields {
                        e.field(n, v, false);
                    }
                    
                    e.color(EMBED_COLOR);
                    e.footer(|f| {
                        f.text(format!("React with ⬅️ or ➡️ below to move back and forth through pages.\n{}", reax_msg));

                        f
                    });
        
                    e
                });
                m
            }).await.unwrap();
            if pagenum > 1 {
                let _ = page_msg.react(&ctx, '⬅').await;
            }
            if pagenum < max_pages {
                let _ = page_msg.react(&ctx, '➡').await;
            }
            let _ = page_msg.react(&ctx, '🔀').await;
            if T::valid_sorts().contains(&SortType::NumID) && cur_sort != SortType::NumID {
                let _ = page_msg.react(&ctx, '🔢').await;
            }
            if T::valid_sorts().contains(&SortType::StrID) && cur_sort != SortType::StrID {
                let _ = page_msg.react(&ctx, '🔠').await;
            }
            if T::valid_sorts().contains(&SortType::Type) && cur_sort != SortType::Type {
                let _ = page_msg.react(&ctx, '🔥').await;
            }
            if T::valid_sorts().contains(&SortType::Tier) && cur_sort != SortType::Tier {
                let _ = page_msg.react(&ctx, '🎟').await;
            }
            if T::valid_sorts().contains(&SortType::CurrentRecord) && cur_sort != SortType::CurrentRecord {
                let _ = page_msg.react(&ctx, '⚔').await;
            }
            if T::valid_sorts().contains(&SortType::LifetimeRecord) && cur_sort != SortType::LifetimeRecord {
                let _ = page_msg.react(&ctx, '🏅').await;
            }
            if T::valid_sorts().contains(&SortType::BST) && cur_sort != SortType::BST {
                let _ = page_msg.react(&ctx, '💪').await;
            }
            if T::valid_sorts().contains(&SortType::SpeciesName) && cur_sort != SortType::SpeciesName {
                let _ = page_msg.react(&ctx, '🔠').await;
            }
            if T::valid_sorts().contains(&SortType::Date) && cur_sort != SortType::Date {
                let _ = page_msg.react(&ctx, '📆').await;
            }
            if T::valid_sorts().contains(&SortType::PlayoffIdentifier) && cur_sort != SortType::PlayoffIdentifier {
                let _ = page_msg.react(&ctx, '🔠').await;
            }
            if cur_sort != SortType::Random {
                let ascmoji = if cur_asc { '🔽' } else { '🔼' };
                let _ = page_msg.react(&ctx, ascmoji).await;
            }
        
            if let Some(reaction) = &page_msg.await_reaction(&ctx).timeout(Duration::from_secs(30)).author_id(msg.author.id).await {

                let emoji = &reaction.as_inner_ref().emoji;
        
                match emoji.to_string().as_str() {
                    "⬅" => { 
                        if pagenum > 1 {
                            pagenum -= 1;
                            let _ = page_msg.delete_reactions(&ctx).await;
                        }
                    },
                    "➡" => {
                        if pagenum < max_pages {
                            pagenum += 1;
                            let _ = page_msg.delete_reactions(&ctx).await;
                        }
                    },
                    "🔀" => {
                        cur_sort = SortType::Default;
                        next_sort = SortType::Random;
                        let _ = page_msg.delete_reactions(&ctx).await;
                    },
                    "🔢" => {
                        next_sort = SortType::NumID;
                        let _ = page_msg.delete_reactions(&ctx).await;
                    },
                    "🔥" => {
                        next_sort = SortType::Type;
                        let _ = page_msg.delete_reactions(&ctx).await;
                    },
                    "🎟" => {
                        next_sort = SortType::Tier;
                        let _ = page_msg.delete_reactions(&ctx).await;
                    },
                    "⚔" => {
                        next_sort = SortType::CurrentRecord;
                        let _ = page_msg.delete_reactions(&ctx).await;
                    },
                    "🏅" => {
                        next_sort = SortType::LifetimeRecord;
                        let _ = page_msg.delete_reactions(&ctx).await;
                    },
                    "💪" => {
                        next_sort = SortType::BST;
                        let _ = page_msg.delete_reactions(&ctx).await;
                    },
                    "🔠" => {
                        if T::valid_sorts().contains(&SortType::StrID) {
                            next_sort = SortType::StrID;
                        }
                        else if T::valid_sorts().contains(&SortType::PlayoffIdentifier) {
                            next_sort = SortType::PlayoffIdentifier;
                        }
                        else {
                            next_sort = SortType::SpeciesName;
                        }
                        let _ = page_msg.delete_reactions(&ctx).await;
                    },
                    "📛" => {
                        next_sort = SortType::Nickname;
                        let _ = page_msg.delete_reactions(&ctx).await;
                    },
                    "📆" => {
                        next_sort = SortType::Date;
                        let _ = page_msg.delete_reactions(&ctx).await;
                    },
                    "🔽" => {
                        next_asc = false;
                        let _ = page_msg.delete_reactions(&ctx).await;
                    },
                    "🔼" => {
                        next_asc = true;
                        let _ = page_msg.delete_reactions(&ctx).await;
                    },
                    _ => {
                        let _ = page_msg.delete_reactions(&ctx).await;
                        let _ = page_msg.react(&ctx, '✅').await;
                        let _ = msg.reply(ctx, "You added an incorrect reaction emoji.").await;
                        //warn!("bad emoji: {}, tostr: {}", emoji.as_data(), emoji.to_string());
                        return Ok(())
                    },
                };
            } else {
                let _ = page_msg.delete_reactions(&ctx).await;
                let _ = page_msg.react(&ctx, '✅').await;
                return Ok(())
            };
        }
    }
    else {
        let fields = {
            let mut f: Vec<(String, String)> = Vec::new();
            for v in list.clone() {
                f.push((v.field_name(ctx).await, v.field_value(ctx).await));
            }
            f
        };

        let msg = msg.channel_id.send_message(&ctx, |m| {
            m.embed(|e| {
                e.title(pager_title);
                e.description(format!("Showing 1-{} of {} {}", len, count, T::type_name_plural()));

                if image.is_some() {
                    e.thumbnail(image.unwrap());
                }
                
                for (n, v) in fields {
                    e.field(n, v, false);
                }
                
                e.color(EMBED_COLOR);
    
                e
            });
            m
        }).await;
    
        if let Err(why) = msg {
            error!("Error sending message: {:?}", why);
        }
    }
    
    Ok(())
}
